/*
 * intra.cc
 *
 * Description of this file:
 *    Intra prediction functions definition of the xrddec library
 *
 * --------------------------------------------------------------------------
 *
 *    xrddec - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"
#include "block_info.h"
#include "intra.h"

// ---------------------------------------------------------------------------
// disable warning
#if defined(_MSC_VER) || defined(__ICL)
#pragma warning(disable: 4100)  // unreferenced formal parameter
#endif

/*
 * ===========================================================================
 * global & local variable defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
static const int8_t g_aucXYflg[NUM_INTRA_MODE] = {
    0, 0, 0, 0, 0,
    0, 0, 0, 0, 0,
    0, 0, 0, 0, 0,
    0, 0, 0, 0, 0,
    0, 0, 0, 0, 0,
    1, 1, 1, 1, 1,
    1, 1, 1
};

/* ---------------------------------------------------------------------------
 */
static const int8_t tab_auc_dir_dx[NUM_INTRA_MODE] = {
     0,  0,  0, 11,  2,
    11,  1,  8,  1,  4,
     1,  1,  0,  1,  1,
     4,  1,  8,  1, 11,
     2, 11,  4,  8,  0,
     8,  4, 11,  2, 11,
     1,  8,  1
};

/* ---------------------------------------------------------------------------
 */
static const int8_t tab_auc_dir_dy[NUM_INTRA_MODE] = {
     0,   0,   0, -4,  -1,
    -8,  -1, -11, -2, -11,
    -4,  -8,   0,  8,   4,
    11,   2,  11,  1,   8,
     1,   4,   1,  1,   0,
    -1,  -1,  -4, -1,  -8,
    -1, -11,  -2
};

/* ---------------------------------------------------------------------------
 */
static const int8_t g_aucSign[NUM_INTRA_MODE] = {
     0,  0,  0, -1, -1,
    -1, -1, -1, -1, -1,
    -1, -1,  0,  1,  1,
     1,  1,  1,  1,  1,
     1,  1,  1,  1,  0,
    -1, -1, -1, -1, -1,
    -1, -1, -1
};

/* ---------------------------------------------------------------------------
 */
static const int8_t tab_auc_dir_dxdy[2][NUM_INTRA_MODE][2] = {
    {
        // dx/dy
        {  0, 0 }, {  0, 0 }, {  0, 0 }, { 11, 2 }, {  2, 0 },
        { 11, 3 }, {  1, 0 }, { 93, 7 }, {  1, 1 }, { 93, 8 },
        {  1, 2 }, {  1, 3 }, {  0, 0 }, {  1, 3 }, {  1, 2 },
        { 93, 8 }, {  1, 1 }, { 93, 7 }, {  1, 0 }, { 11, 3 },
        {  2, 0 }, { 11, 2 }, {  4, 0 }, {  8, 0 }, {  0, 0 },
        {  8, 0 }, {  4, 0 }, { 11, 2 }, {  2, 0 }, { 11, 3 },
        {  1, 0 }, { 93, 7 }, {  1, 1 },
    }, {
        // dy/dx
        {  0, 0 }, {  0, 0 }, {  0, 0 }, { 93, 8 }, {  1, 1 },
        { 93, 7 }, {  1, 0 }, { 11, 3 }, {  2, 0 }, { 11, 2 },
        {  4, 0 }, {  8, 0 }, {  0, 0 }, {  8, 0 }, {  4, 0 },
        { 11, 2 }, {  2, 0 }, { 11, 3 }, {  1, 0 }, { 93, 7 },
        {  1, 1 }, { 93, 8 }, {  1, 2 }, {  1, 3 }, {  0, 0 },
        {  1, 3 }, {  1, 2 }, { 93, 8 }, {  1, 1 }, { 93, 7 },
        {  1, 0 }, { 11, 3 }, {  2, 0 }
    }
};

/* ---------------------------------------------------------------------------
 */
static const int8_t tab_log2size[MAX_CU_SIZE + 1] = {
    -1, -1, -1, -1,  2, -1, -1, -1,
     3, -1, -1, -1, -1, -1, -1, -1,
     4, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
     5, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    6
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_DL_Avail64[16 * 16] = {
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_DL_Avail32[8 * 8] = {
    1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 1, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 0, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0,
    1, 0, 0, 0, 1, 0, 0, 0,
    1, 0, 1, 0, 1, 0, 1, 0,
    0, 0, 0, 0, 0, 0, 0, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_DL_Avail16[4 * 4] = {
    1, 0, 1, 0,
    1, 0, 0, 0,
    1, 0, 1, 0,
    0, 0, 0, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_DL_Avail8[2 * 2] = {
    1, 0,
    0, 0
};


/* ---------------------------------------------------------------------------
 */
const int8_t tab_TR_Avail64[16 * 16] = {
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_TR_Avail32[8 * 8] = {
    // 0: 8 1:16 2: 32  pu size
    1, 1, 1, 1, 1, 1, 1, 1,
    1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 0, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 1, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0,
    1, 1, 1, 0, 1, 1, 1, 0,
    1, 0, 1, 0, 1, 0, 1, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_TR_Avail16[4 * 4] = {
    1, 1, 1, 1,
    1, 0, 1, 0,
    1, 1, 1, 0,
    1, 0, 1, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t tab_TR_Avail8[2 * 2] = {
    1, 1,
    1, 0
};

/* ---------------------------------------------------------------------------
 */
const int8_t *tab_DL_Avails[MAX_CU_SIZE_IN_BIT + 1] = {
    NULL, NULL, NULL, tab_DL_Avail8, tab_DL_Avail16, tab_DL_Avail32, tab_DL_Avail64
};


/* ---------------------------------------------------------------------------
 */
const int8_t *tab_TR_Avails[MAX_CU_SIZE_IN_BIT + 1] = {
    NULL, NULL, NULL, tab_TR_Avail8, tab_TR_Avail16, tab_TR_Avail32, tab_TR_Avail64
};


/* records the sample bit depth for intra predeiction
 */

/**
 * ===========================================================================
 * local function definition
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE int is_block_available(xrddec_t *h, int x_4x4, int y_4x4, int dx_4x4, int dy_4x4, int cur_slice_idx)
{
    int x2_4x4 = x_4x4 + dx_4x4;
    int y2_4x4 = y_4x4 + dy_4x4;

    if (x2_4x4 < 0 || y2_4x4 < 0 || x2_4x4 >= h->i_width_in_spu || y2_4x4 >= h->i_height_in_spu) {
        return 0;
    } else {
        return cur_slice_idx == h->scu_data[(y2_4x4 >> 1) * h->i_width_in_scu + (x2_4x4 >> 1)].i_slice_nr;
    }
}

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE
uint32_t get_intra_neighbors(xrddec_t *h, int x_4x4, int y_4x4, int bsx, int bsy, int cur_slice_idx)
{
    /* 1. ڿǷͬһSlice */
    int b_LEFT      = is_block_available(h, x_4x4, y_4x4, -1, 0, cur_slice_idx);
    int b_TOP       = is_block_available(h, x_4x4, y_4x4,  0, -1, cur_slice_idx);
    int b_TOP_LEFT  = is_block_available(h, x_4x4, y_4x4, -1, -1, cur_slice_idx);
    int b_LEFT_DOWN = is_block_available(h, x_4x4, y_4x4, -1, (bsy >> 1) - 1, cur_slice_idx);  // (bsy >> MIN_PU_SIZE_IN_BIT << 1)
    int b_TOP_RIGHT = is_block_available(h, x_4x4, y_4x4, (bsx >> 1) - 1, -1, cur_slice_idx);  // (bsx >> MIN_PU_SIZE_IN_BIT << 1)

    int leftdown;
    int upright;
    int log2_lcu_size_in_spu = (h->i_lcu_level - B4X4_IN_BIT);
    int i_lcu_mask = (1 << log2_lcu_size_in_spu) - 1;

    /* 2. ڿǷڵǰ֮ǰع */
    x_4x4 = x_4x4 & i_lcu_mask;
    y_4x4 = y_4x4 & i_lcu_mask;

    leftdown = h->p_tab_DL_avail[((y_4x4 + (bsy >> 2) - 1) << log2_lcu_size_in_spu) + (x_4x4)];
    upright  = h->p_tab_TR_avail[((y_4x4) << log2_lcu_size_in_spu) + (x_4x4 + (bsx >> 2) - 1)];

    b_LEFT_DOWN = b_LEFT_DOWN && leftdown;
    b_TOP_RIGHT = b_TOP_RIGHT && upright;

    return (b_LEFT << MD_I_LEFT) | (b_TOP << MD_I_TOP) | (b_TOP_LEFT << MD_I_TOP_LEFT) |
        (b_TOP_RIGHT << MD_I_TOP_RIGHT) | (b_LEFT_DOWN << MD_I_LEFT_DOWN);
}

/* ---------------------------------------------------------------------------
 */
static void ALWAYS_INLINE mem_repeat_p(pel_t *dst, pel_t val, size_t num)
{
    while (num--) {
        *dst++ = val;
    }
}

/* ---------------------------------------------------------------------------
 */
static void ALWAYS_INLINE memcpy_vh_pp_c(pel_t *dst, pel_t *src, int i_src, size_t num)
{
    while (num--) {
        *dst++ = *src;
        src += i_src;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_ver_c(pel_t *src, pel_t *dst, int i_dst, int mode, int width, int height)
{
    pel_t *p_src = src + 1;
    int y;

    UNUSED_PARAMETER(mode);

    for (y = height; y != 0; y--) {
        memcpy(dst, p_src, width * sizeof(pel_t));
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_hor_c(pel_t *src, pel_t *dst, int i_dst, int mode, int width, int height)
{
    pel_t *p_src = src - 1;
    int x, y;

    UNUSED_PARAMETER(mode);

    for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
            dst[x] = p_src[-y];
        }
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_dc_c(pel_t *src, pel_t *dst, int i_dst, int mode, int width, int height)
{
    int b_top  = mode >> 8;
    int b_left = mode & 0xFF;
    pel_t *p_src = src - 1;
    int dc_value = 0;
    int x, y;

    /* get DC value */
    if (b_left) {
        for (y = 0; y < height; y++) {
            dc_value += p_src[-y];
        }

        p_src = src + 1;
        if (b_top) {
            for (x = 0; x < width; x++) {
                dc_value += p_src[x];
            }

            dc_value += ((width + height) >> 1);
            dc_value = (dc_value * (512 / (width + height))) >> 9;
        } else {
            dc_value += height / 2;
            dc_value /= height;
        }
    } else {
        p_src = src + 1;
        if (b_top) {
            for (x = 0; x < width; x++) {
                dc_value += p_src[x];
            }

            dc_value += width / 2;
            dc_value /= width;
        } else {
            dc_value = 1 << (g_bit_depth - 1);
        }
    }

    /* fill the block */
    x        = (1 << g_bit_depth) - 1;     /* max pixel value */
    dc_value = XRDDEC_CLIP3(0, x, dc_value);
    for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
            dst[x] = (pel_t)dc_value;
        }
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_plane_c(pel_t *src, pel_t *dst, int i_dst, int mode, int width, int height)
{
    int ib_mult[5] = { 13, 17, 5, 11, 23 };
    int ib_shift[5] = { 7, 10, 11, 15, 19 };
    int im_h = ib_mult [tab_log2size[width ] - 2];
    int im_v = ib_mult [tab_log2size[height] - 2];
    int is_h = ib_shift[tab_log2size[width ] - 2];
    int is_v = ib_shift[tab_log2size[height] - 2];
    int iW2 = width >> 1;
    int iH2 = height >> 1;
    int iH = 0;
    int iV = 0;
    int iA, iB, iC;
    int x, y;
    int iTmp, iTmp2;
    int max_val = (1 << g_bit_depth) - 1;
    pel_t *p_src;

    UNUSED_PARAMETER(mode);

    p_src = src + 1;
    p_src += (iW2 - 1);
    for (x = 1; x < iW2 + 1; x++) {
        iH += x * (p_src[x] - p_src[-x]);
    }

    p_src = src - 1;
    p_src -= (iH2 - 1);
    for (y = 1; y < iH2 + 1; y++) {
        iV += y * (p_src[-y] - p_src[y]);
    }

    p_src = src;
    iA = (p_src[-1 - (height - 1)] + p_src[1 + width - 1]) << 4;
    iB = ((iH << 5) * im_h + (1 << (is_h - 1))) >> is_h;
    iC = ((iV << 5) * im_v + (1 << (is_v - 1))) >> is_v;

    iTmp = iA - (iH2 - 1) * iC - (iW2 - 1) * iB + 16;
    for (y = 0; y < height; y++) {
        iTmp2 = iTmp;
        for (x = 0; x < width; x++) {
            dst[x] = (pel_t)XRDDEC_CLIP3(0, max_val, iTmp2 >> 5);
            iTmp2 += iB;
        }

        dst += i_dst;
        iTmp += iC;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_bilinear_c(pel_t *src, pel_t *dst, int i_dst, int mode, int width, int height)
{
    itr_t pTop[MAX_CU_SIZE];
    itr_t pLeft[MAX_CU_SIZE];
    itr_t pT[MAX_CU_SIZE];
    itr_t pL[MAX_CU_SIZE];
    itr_t wy[MAX_CU_SIZE];
    int ishift_x  = tab_log2size[width];
    int ishift_y  = tab_log2size[height];
    int ishift    = XRDDEC_MIN(ishift_x, ishift_y);
    int ishift_xy = ishift_x + ishift_y + 1;
    int offset    = 1 << (ishift_x + ishift_y);
    int a, b, c, w, wxy, t;
    int predx;
    int x, y;
    int max_value = (1 << g_bit_depth) - 1;

    UNUSED_PARAMETER(mode);

    for (x = 0; x < width; x++) {
        pTop[x] = src[1 + x];
    }

    for (y = 0; y < height; y++) {
        pLeft[y] = src[-1 - y];
    }

    a = pTop[width - 1];
    b = pLeft[height - 1];
    c = (width == height) ? (a + b + 1) >> 1 : (((a << ishift_x) + (b << ishift_y)) * 13 + (1 << (ishift + 5))) >> (ishift + 6);
    w = (c << 1) - a - b;

    for (x = 0; x < width; x++) {
        pT[x] = (itr_t)(b - pTop[x]);
        pTop[x] <<= ishift_y;
    }
    t = 0;
    for (y = 0; y < height; y++) {
        pL[y] = (itr_t)(a - pLeft[y]);
        pLeft[y] <<= ishift_x;
        wy[y] = (itr_t)t;
        t += w;
    }

    for (y = 0; y < height; y++) {
        predx = pLeft[y];
        wxy = -wy[y];

        for (x = 0; x < width; x++) {
            predx += pL[y];
            wxy += wy[y];
            pTop[x] += pT[x];
            dst[x] = (pel_t)XRDDEC_CLIP3(0, max_value, (((predx << ishift_y) + (pTop[x] << ishift_x) + wxy + offset) >> ishift_xy));
        }

        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static int get_context_pixel(int mode, int uiXYflag, int iTempD, int *offset)
{
    int imult = tab_auc_dir_dxdy[uiXYflag][mode][0];
    int ishift = tab_auc_dir_dxdy[uiXYflag][mode][1];
    int iTempDn = iTempD * imult >> ishift;

    *offset = ((iTempD * imult * 32) >> ishift) - iTempDn * 32;

    return iTempDn;
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_ang_x_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    int iDx = tab_auc_dir_dx[dir_mode];
    int iDy = tab_auc_dir_dy[dir_mode];
    int iX;
    int c1, c2, c3, c4;
    int i, j;
    pel_t *dst_base = dst + iDy * i_dst + iDx;

    for (j = 0; j < bsy; j++, iDy++) {
        iX = get_context_pixel(dir_mode, 0, j + 1, &c4);
        c1 = 32 - c4;
        c2 = 64 - c4;
        c3 = 32 + c4;

        i = 0;

        for (; i < bsx; i++) {
            dst[i] = (pel_t)((src[iX] * c1 + src[iX + 1] * c2 + src[iX + 2] * c3 + src[iX + 3] * c4 + 64) >> 7);
            iX++;
        }

        dst_base += i_dst;
        dst += i_dst;
    }
}


/* ---------------------------------------------------------------------------
 */
static void intra_pred_ang_y_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    int offsets[64];
    int xsteps[64];
    int iDx = tab_auc_dir_dx[dir_mode];
    int iDy = tab_auc_dir_dy[dir_mode];
    int i, j;
    int offset;
    int iY;
    pel_t *dst_base = dst + iDy * i_dst + iDx;

    for (i = 0; i < bsx; i++) {
        xsteps[i] = get_context_pixel(dir_mode, 1, i + 1, &offsets[i]);
    }

    for (j = 0; j < bsy; j++) {
        for (i = 0; i < bsx; i++) {
            int idx;
            iY = j + xsteps[i];
            idx = -iY;
            offset = offsets[i];
            dst[i] = (pel_t)((src[idx] * (32 - offset) + src[idx - 1] * (64 - offset) + src[idx - 2] * (32 + offset) + src[idx - 3] * offset + 64) >> 7);
        }
        dst_base += i_dst;
        dst += i_dst;
    }
}


/* ---------------------------------------------------------------------------
 */
static void intra_pred_ang_xy_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    ALIGN16(int xoffsets[64]);
    ALIGN16(int xsteps[64]);
    const int iDx = tab_auc_dir_dx[dir_mode];
    const int iDy = tab_auc_dir_dy[dir_mode];
    pel_t *dst_base = dst - iDy * i_dst - iDx;
    int i, j, iXx, iYy;
    int offsetx, offsety;

    for (i = 0; i < bsx; i++) {
        xsteps[i] = get_context_pixel(dir_mode, 1, i + 1, &xoffsets[i]);
    }

    for (j = 0; j < bsy; j++) {
        iXx = -get_context_pixel(dir_mode, 0, j + 1, &offsetx);

        for (i = 0; i < bsx; i++) {
#if !BUGFIX_PREDICTION_INTRA
            if (j >= iDy && i >= iDx) {
                dst[i] = dst_base[i];
            } else {
#endif
                iYy = j - xsteps[i];

                if (iYy <= -1) {
                    dst[i] = (pel_t)((src[ iXx + 2] * (32 - offsetx) + src[ iXx + 1] * (64 - offsetx) + src[ iXx] * (32 + offsetx) + src[ iXx - 1] * offsetx + 64) >> 7);
                } else {
                    offsety = xoffsets[i];
                    dst[i] = (pel_t)((src[-iYy - 2] * (32 - offsety) + src[-iYy - 1] * (64 - offsety) + src[-iYy] * (32 + offsety) + src[-iYy + 1] * offsety + 64) >> 7);
                }
#if !BUGFIX_PREDICTION_INTRA
            }
#endif
            iXx++;
        }
        dst_base += i_dst;
        dst      += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCUϱ߽PU
 */
static 
void fill_reference_samples_0_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    int num_padding = 0;

    /* fill default value */
    mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        gf_xrddec.fast_memcpy(&EP[1], &pLcuEP[1], bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        gf_xrddec.fast_memcpy(&EP[bsx + 1], &pLcuEP[bsx + 1], bsx * sizeof(pel_t));
    } else {
        mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        /* fill left pixels */
        memcpy(&EP[-bsy], &pLcuEP[-bsy], bsy * sizeof(pel_t));
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        memcpy(&EP[-2 * bsy], &pLcuEP[-2 * bsy], bsy * sizeof(pel_t));
    } else {
        mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pLcuEP[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pLcuEP[1];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pLcuEP[-1];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCUϱ߽PU
 */
static 
void fill_reference_samples_x_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    const pel_t *pL = pTL + i_TL;
    int num_padding = 0;

    /* fill default value */
    mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        gf_xrddec.fast_memcpy(&EP[1], &pLcuEP[1], bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        gf_xrddec.fast_memcpy(&EP[bsx + 1], &pLcuEP[bsx + 1], bsx * sizeof(pel_t));
    } else {
        mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        const pel_t *p_l = pL;
        int y;
        /* fill left pixels */
        for (y = 0; y < bsy; y++) {
            EP[-1 - y] = *p_l;
            p_l += i_TL;
        }
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        int y;
        const pel_t *p_l = pL + bsy * i_TL;

        for (y = 0; y < bsy; y++) {
            EP[-bsy - 1 - y] = *p_l;
            p_l += i_TL;
        }
    } else {
        mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pLcuEP[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pLcuEP[1];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pL[0];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCU߽ϵPU
 */
static 
void fill_reference_samples_y_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    const pel_t *pT = pTL + 1;
    int num_padding = 0;

    /* fill default value */
    mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        gf_xrddec.fast_memcpy(&EP[1], pT, bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        gf_xrddec.fast_memcpy(&EP[bsx + 1], &pT[bsx], bsx * sizeof(pel_t));
    } else {
        mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        /* fill left pixels */
        memcpy(&EP[-bsy], &pLcuEP[-bsy], bsy * sizeof(pel_t));
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        memcpy(&EP[-2 * bsy], &pLcuEP[-2 * bsy], bsy * sizeof(pel_t));
    } else {
        mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pLcuEP[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pT[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pLcuEP[-1];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCUڲڱ߽ϵPU
 */
static 
void fill_reference_samples_xy_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    const pel_t *pT = pTL + 1;
    const pel_t *pL = pTL + i_TL;
    int num_padding = 0;

    /* fill default value */
    mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        gf_xrddec.fast_memcpy(&EP[1], pT, bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        gf_xrddec.fast_memcpy(&EP[bsx + 1], &pT[bsx], bsx * sizeof(pel_t));
    } else {
        mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        const pel_t *p_l = pL;
        int y;
        /* fill left pixels */
        for (y = 0; y < bsy; y++) {
            EP[-1 - y] = *p_l;
            p_l += i_TL;
        }
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        int y;
        const pel_t *p_l = pL + bsy * i_TL;

        for (y = 0; y < bsy; y++) {
            EP[-bsy - 1 - y] = *p_l;
            p_l += i_TL;
        }
    } else {
        mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pTL[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pT[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pL[0];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * make intra prediction for luma block
 */
void xrddec_get_intra_pred(xrddec_row_rec_t *row_rec, cu_t *p_cu, int predmode, int ctu_x, int ctu_y, int bsx, int bsy)
{
    const int xy = ((ctu_y != 0) << 1) + (ctu_x != 0);
    pel_t *EP = row_rec->buf_edge_pixels + (MAX_CU_SIZE << 2) - 1;
    int b8_x = (ctu_x >> MIN_PU_SIZE_IN_BIT) + row_rec->ctu.i_spu_x;
    int b8_y = (ctu_y >> MIN_PU_SIZE_IN_BIT) + row_rec->ctu.i_spu_y;
    int i_pred = row_rec->ctu.i_fdec[0];
    pel_t *p_pred = row_rec->ctu.p_fdec[0] + ctu_y * i_pred + ctu_x;
    pel_t *pTL;
    int i_src;
    uint32_t avail;

    assert(predmode >= 0 && predmode < NUM_INTRA_MODE);
    avail = get_intra_neighbors(row_rec->h, b8_x, b8_y, bsx, bsy, p_cu->i_slice_nr);

    row_rec->b_block_avail_top  = (bool_t)IS_NEIGHBOR_AVAIL(avail, MD_I_TOP ); // used for second transform
    row_rec->b_block_avail_left = (bool_t)IS_NEIGHBOR_AVAIL(avail, MD_I_LEFT); // used for second transform

    i_src = i_pred;
    pTL   = p_pred - i_src - 1;

    gf_xrddec.fill_edge_f[xy](pTL, i_src, row_rec->ctu_border[0].rec_top + ctu_x - ctu_y, EP, avail, bsx, bsy);

    intra_pred(EP, p_pred, i_pred, predmode, bsy, bsx, avail);
}

/* ---------------------------------------------------------------------------
 * make intra prediction for chroma block
 */
void xrddec_get_intra_pred_chroma(xrddec_row_rec_t *row_rec, cu_t *p_cu, int ctu_c_x, int ctu_c_y)
{
    static const int TAB_CHROMA_MODE_TO_REAL_MODE[NUM_INTRA_MODE_CHROMA] = {
        DC_PRED, DC_PRED, HOR_PRED, VERT_PRED, BI_PRED
    };
    const int xy = ((ctu_c_y != 0) << 1) + (ctu_c_x != 0);
    pel_t *EP_u     = row_rec->buf_edge_pixels + (MAX_CU_SIZE << 1) - 1;
    pel_t *EP_v     = EP_u + (MAX_CU_SIZE << 2);
    int bsize_c     = 1 << (p_cu->i_cu_level - 1);
    int b8_x        = ((ctu_c_x << 1) >> MIN_PU_SIZE_IN_BIT) + row_rec->ctu.i_spu_x;
    int b8_y        = ((ctu_c_y << 1) >> MIN_PU_SIZE_IN_BIT) + row_rec->ctu.i_spu_y;
    int luma_mode   = p_cu->intra_pred_modes[0];
    int chroma_mode = p_cu->c_ipred_mode;
    int real_mode   = (chroma_mode == DM_PRED_C) ? luma_mode : TAB_CHROMA_MODE_TO_REAL_MODE[chroma_mode];
    uint32_t avail;

    /* Ԥλ */
    int i_pred      = row_rec->ctu.i_fdec[1];
    pel_t *p_pred_u = row_rec->ctu.p_fdec[1] + ctu_c_y * i_pred + ctu_c_x;
    pel_t *p_pred_v = row_rec->ctu.p_fdec[2] + ctu_c_y * i_pred + ctu_c_x;

    /* UVϽصλ */
    int i_src       = i_pred;
    pel_t *pTL_u    = p_pred_u - i_src - 1;
    pel_t *pTL_v    = p_pred_v - i_src - 1;

    /* ο߽жο߽ */
    avail = get_intra_neighbors(row_rec->h, b8_x, b8_y, bsize_c << 1, bsize_c << 1, p_cu->i_slice_nr);

    gf_xrddec.fill_edge_f[xy](pTL_u, i_src, row_rec->ctu_border[1].rec_top + ctu_c_x - ctu_c_y, EP_u, avail, bsize_c, bsize_c);
    gf_xrddec.fill_edge_f[xy](pTL_v, i_src, row_rec->ctu_border[2].rec_top + ctu_c_x - ctu_c_y, EP_v, avail, bsize_c, bsize_c);

    /* ִԤ */
    intra_pred(EP_u, p_pred_u, i_pred, real_mode, bsize_c, bsize_c, avail);
    intra_pred(EP_v, p_pred_v, i_pred, real_mode, bsize_c, bsize_c, avail);
}


/* ---------------------------------------------------------------------------
 */
void xrddec_intra_pred_init(uint32_t cpuid, ao_funcs_t *pf)
{
#define ANG_X_OFFSET    3
#define ANG_XY_OFFSET   13
#define ANG_Y_OFFSET    25
    int i;
    intra_pred_t *ipred = pf->intraf;

    pf->fill_edge_f[0]      = fill_reference_samples_0_c;
    pf->fill_edge_f[1]      = fill_reference_samples_x_c;
    pf->fill_edge_f[2]      = fill_reference_samples_y_c;
    pf->fill_edge_f[3]      = fill_reference_samples_xy_c;
    ipred[DC_PRED   ] = intra_pred_dc_c;                // 0
    ipred[PLANE_PRED] = intra_pred_plane_c;             // 1
    ipred[BI_PRED   ] = intra_pred_bilinear_c;          // 2

    for (i = ANG_X_OFFSET; i < VERT_PRED; i++) {
        ipred[i     ] = intra_pred_ang_x_c;             // 3 ~ 11
    }
    ipred[VERT_PRED ] = intra_pred_ver_c;               // 12

    for (i = ANG_XY_OFFSET; i < HOR_PRED; i++) {
        ipred[i     ] = intra_pred_ang_xy_c;            // 13 ~ 23
    }

    ipred[HOR_PRED  ] = intra_pred_hor_c;               // 24
    for (i = ANG_Y_OFFSET; i < NUM_INTRA_MODE; i++) {
        ipred[i     ] = intra_pred_ang_y_c;             // 25 ~ 32
    }
#undef ANG_X_OFFSET
#undef ANG_XY_OFFSET
#undef ANG_Y_OFFSET
}
