/*
 * xrddec.h
 *
 * Description of this file:
 *    Decoder functions definition of the xrddec library
 *
 * --------------------------------------------------------------------------
 *
 *    xrddec - video decoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#ifndef __XRDDEC_H__
#define __XRDDEC_H__

#ifdef __cplusplus
extern "C" {    // only need to export C interface if used by C++ source code
#endif
    
#if !defined(_STDINT_H) && !defined(_STDINT_H_) && !defined(_STDINT_H_INCLUDED) && !defined(_STDINT) &&\
    !defined(_SYS_STDINT_H_) && !defined(_INTTYPES_H) && !defined(_INTTYPES_H_) && !defined(_INTTYPES)
# ifdef _MSC_VER
#  pragma message("You must include stdint.h or inttypes.h before xrddec.h")
# else
#  warning You must include stdint.h or inttypes.h before xrddec.h
# endif
#endif

/* XRDDEC API version, version 2 the non-caller version(truncated), version 3 save all the input bytes */
#define XRDDEC_API_VERSION          1

/**
 * ===========================================================================
 * const defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * picture type */
#define AVS2DEC_PIC_I       0         /* picture-I */
#define AVS2DEC_PIC_P       1         /* picture-P */
#define AVS2DEC_PIC_B       2         /* picture-B */
#define AVS2DEC_PIC_G       3         /* picture-G */
#define AVS2DEC_PIC_F       4         /* picture-F */
#define AVS2DEC_PIC_S       5         /* picture-S */

/* ---------------------------------------------------------------------------
 * profile id */
#define PROFILE_MAIN_PIC    0x12      /* main picture profile */
#define PROFILE_MAIN        0x20      /* main         profile */
#define PROFILE_MAIN10      0x22      /* main 10bit   profile */

/* ---------------------------------------------------------------------------
 * log level
 */
enum xrddec_log_level_e {
    AVS2_LOG_DEBUG   = 0,
    AVS2_LOG_INFO    = 1,
    AVS2_LOG_WARNING = 2,
    AVS2_LOG_ERROR   = 3,
    AVS2_LOG_FATAL   = 4,
    AVS2_LOG_MAX     = 5
};

/* ---------------------------------------------------------------------------
 * information of return value for decode/flush()
 */
enum xrddec_ret_e {
    XRDDEC_END         = -1,   /* Decoding ended: no more bit-stream to decode and no more frames to output */
    XRDDEC_DEFAULT     = 0,    /* Decoding but no output */
    XRDDEC_GOT_FRAME   = 1,    /* Decoding get frame */
    XRDDEC_GOT_HEADER  = 2,    /* Decoding get sequence header */
    XRDDEC_GOT_BOTH    = 3,    /* Decoding get sequence header and frame together */
};

/**
 * ===========================================================================
 * interface struct type defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * information of sequence header
 */
typedef struct xrddec_seq_info_t {
    uint32_t        profile_id;       /* profile ID */
    uint32_t        level_id;         /* level   ID */
    uint32_t        progressive;      /* progressive sequence (0: interlace, 1: progressive) */
    uint32_t        horizontal_size;  /* image width */
    uint32_t        vertical_size;    /* image height */
    uint32_t        chroma_format;    /* chroma format(1: 4:2:0, 2: 4:2:2) */
    uint32_t        aspect_ratio;     /* 2: 4:3,  3: 16:9 */
    uint32_t        low_delay;        /* low delay */
    uint32_t        bitrate;          /* bitrate (bps) */
    uint32_t        internal_bitdepth;/* internal sample bit depth */
    uint32_t        output_bitdepth;  /* output sample bit depth */
    uint32_t        bytes_per_sample; /* bytes per sample */
    float           frame_rate;       /* frame rate */
    uint32_t        frame_rate_code;  /* frame rate code, mpeg12 [1...8] */
} xrddec_seq_info_t;  

/* ---------------------------------------------------------------------------
 * packet of bitstream
 */
typedef struct xrddec_packet_t {
    uint8_t        *data;             /* bitstream */
    int             len;              /* bytes of the bitstream */
    int             marker;           /* force a frame boundary after this packet */
    int64_t         pts;              /* presentation time stamp */
    int64_t         dts;              /* decoding time stamp */
} xrddec_packet_t;

/* ---------------------------------------------------------------------------
 * decoded picture
 */
typedef struct xrddec_picture_t {
    void           *magic;            /* must be the 1st member variable. do not change it */
    /* picture information */
    uint8_t        *planes[3];        /* picture planes */
    int             width[3];         /* picture width in pixels */
    int             lines[3];         /* picture height in pixels */
    /* strideһͼռõֽ, һɫͨص */
    int             pic_order_count;  /* picture number */
    int             type;             /* picture type of the corresponding frame */
    int             QP;               /* QP of the corresponding picture */
    int64_t         pts;              /* presentation time stamp */
    int64_t         dts;              /* decoding time stamp */
    int             i_pic_planes;     /* number of planes */
    int             bytes_per_sample; /* number of bytes for each sample */
    int             pic_bit_depth;    /* number of bytes for each sample */
#if XRDDEC_API_VERSION >= 2
    int             ret_type;         /* return type, xrddec_ret_e */
#endif
    int             pic_decode_error; /* is there any decoding error of this frame? */
} xrddec_picture_t;

#if XRDDEC_API_VERSION < 2
/* ---------------------------------------------------------------------------
 * Function   : output one picture or a sequence header
 * Parameters :
 *       [in] : pic     - picture
 *       [in] : headset - sequence header
 *       [in] : errCode - possible errors while decoding @pic
 *       [in] : opaque  - user data
 */
typedef void(*xrddec_output_func)(xrddec_picture_t *pic, xrddec_seq_info_t *headset, int errCode, void *opaque);

/* ---------------------------------------------------------------------------
 * parameters for create an AVS2 decoder
 */
typedef struct xrddec_param_t {
    int               threads;        /* decoding threads: 0 for auto */
    xrddec_output_func    output_f;       /* decoder picture output (MUST exist) */
    int               i_info_level;   /* 0: All; 1: Only warning and errors; 2: only errors */
    void             *opaque;         /* user data */
} xrddec_param_t;
#else
/* ---------------------------------------------------------------------------
 * parameters for create an AVS2 decoder
 */
typedef struct xrddec_param_t {
    int               threads;        /* decoding threads: 0 for auto */
    int               i_info_level;   /* 0: All; 1: Only warning and errors; 2: only errors */
    void             *opaque;         /* used for future extension */
} xrddec_param_t;
#endif


/**
 * ===========================================================================
 * interface function declares (XRDDEC library APIs for AVS2 video decoder)
 * ===========================================================================
 */

/**
 * ---------------------------------------------------------------------------
 * Function   : open an AVS2 decoder
 * Parameters :
 *   [in/out] : param - pointer to struct xrddec_param_t
 * Return     : handle of the decoder, zero for failure
 * ---------------------------------------------------------------------------
 */
void *
xrddec_decoder_open(xrddec_param_t *param);

/**
 * ---------------------------------------------------------------------------
 * Function   : decode one frame
 * Parameters :
 *       [in] : decoder - pointer to the AVS2 decoder handler
 *   [in/out] : packet  - pointer to struct xrddec_packet_t
 * Return     : 0 for successful, otherwise -1
 * ---------------------------------------------------------------------------
 */
#if XRDDEC_API_VERSION < 2
int xrddec_decoder_decode(void *decoder, xrddec_packet_t *packet);
#else
int xrddec_decoder_decode(void *decoder, xrddec_packet_t *packet, xrddec_seq_info_t *headerset, xrddec_picture_t *out_frame);
#endif

/**
 * ---------------------------------------------------------------------------
 * Function   : flush the decoder
 * Parameters :
 *       [in] : decoder - decoder handle
 * Return     : none
 * ---------------------------------------------------------------------------
 */
#if XRDDEC_API_VERSION < 2
void xrddec_decoder_flush(void *decoder);
#else
int xrddec_decoder_flush(void *decoder, xrddec_seq_info_t *headerset, xrddec_picture_t *out_frame);
#endif

#if XRDDEC_API_VERSION >= 2
/**
 * ---------------------------------------------------------------------------
 * Function   : release one output frame
 * Parameters :
 *       [in] : decoder   - decoder handle
 *            : out_frame - frame to recycle
 * Return     : none
 * ---------------------------------------------------------------------------
 */
void xrddec_decoder_frame_unref(void *decoder, xrddec_picture_t *out_frame);
#endif

/**
 * ---------------------------------------------------------------------------
 * Function   : close the AVS2 decoder
 * Parameters :
 *       [in] : decoder - decoder handle
 * Return     : none
 * ---------------------------------------------------------------------------
 */
void xrddec_decoder_close(void *decoder);

#ifdef __cplusplus
}
#endif

#endif // __XRDDEC_H__
