/*
 * common.c
 *
 * Description of this file:
 *    misc common functionsdefinition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"

#include <stdarg.h>
#include <ctype.h>
#if SYS_WINDOWS
#include <sys/types.h>
#include <sys/timeb.h>
#else
#include <sys/time.h>
#endif
#include <time.h>

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

/**
 * ===========================================================================
 * global variables
 * ===========================================================================
 */

int g_bit_depth   = 8;
int max_pel_value = 255;
int g_dc_value    = 128;

const float FRAME_RATE[8] = {
    24000.0f / 1001.0f, 24.0f, 25.0f, 30000.0f / 1001.0f, 30.0f, 50.0f, 60000.0f / 1001.0f, 60.0f
};

#if XRD_TRACE
FILE *h_trace = NULL;           /* global file handle for trace file */
int g_sym_count = 0;            /* global symbol count for trace */
int g_bit_count = 0;            /* global bit    count for trace */
#endif

#if PTW32_STATIC_LIB
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

/* this is a global in pthread-win32 to indicate if it has been
 * initialized or not */
extern int ptw32_processInitialized;

#endif

#if XRD_TRACE
/**
 * ===========================================================================
 * trace file
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
int xrd_trace_init(xrd_param_t *param)
{
    if (strlen(param->psz_trace_file) > 0) {
        /* create or truncate the trace file */
        h_trace = fopen(param->psz_trace_file, "wt");
        if (!h_trace) {
            xrd_log(NULL, XRD_LOG_ERROR, "trace: can't write to %s\n", param->psz_trace_file);
            return -1;
        } else if (!xrd_is_regular_file(fileno(h_trace))) {
            xrd_log(NULL, XRD_LOG_ERROR, "trace: incompatible with non-regular file %s\n", param->psz_trace_file);
            return -1;
        }
    }

    return 0;
}

/* ---------------------------------------------------------------------------
 */
void xrd_trace_destroy(void)
{
    if (h_trace) {
        fclose(h_trace);
    }
}

/* ---------------------------------------------------------------------------
 */
int xrd_trace(const char *psz_fmt, ...)
{
    int len = 0;

    /* append to the trace file */
    if (h_trace) {
        va_list arg;
        va_start(arg, psz_fmt);

        len = vfprintf(h_trace, psz_fmt, arg);
        fflush(h_trace);
        va_end(arg);
    }

    return len;
}

#endif

/**
 * ===========================================================================
 * xrd_log
 * ===========================================================================
 */

#ifdef _MSC_VER
/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE
void xrd_set_font_color(int color)
{
    static const WORD colors[] = {
        FOREGROUND_INTENSITY | FOREGROUND_RED,                     // ɫ
        FOREGROUND_INTENSITY | FOREGROUND_RED | FOREGROUND_GREEN,  // ɫ
        FOREGROUND_INTENSITY | FOREGROUND_GREEN,                   // ɫ
        FOREGROUND_INTENSITY | FOREGROUND_GREEN | FOREGROUND_BLUE, // cyan
        FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE,       // ɫ
    };
    color = XRD_MIN(4, color);
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), colors[color]);
}
#endif

/* ---------------------------------------------------------------------------
 */
static void
xrd_log_default(int i_log_level, const char *psz_fmt)
{
#if !defined(_MSC_VER)
    static const char str_color_clear[] = "\033[0m";  // "\033[0m"
    static const char str_color[][16] = {
    /*        red        yellow        green          cyan        (default)  */
        "\033[1;31m", "\033[1;33m", "\033[1;32m", "\033[1;36m", "\033[0m"
    };
    const char *cur_color = str_color[i_log_level];
#endif
    static const char *null_prefix = "";
    const char *psz_prefix = null_prefix;

    switch (i_log_level) {
    case XRD_LOG_ERROR:
        psz_prefix = "[error]: ";
        break;
    case XRD_LOG_WARNING:
        psz_prefix = "[warning]: ";
        break;
    case XRD_LOG_INFO:
        psz_prefix = "[info]: ";
        break;
    case XRD_LOG_DEBUG:
        psz_prefix = "[debug]: ";
        break;
    case XRD_LOG_NOPREFIX:
        psz_prefix = null_prefix;
#if !defined(_MSC_VER)
        cur_color = str_color[3];
#endif
        break;
    default:
        psz_prefix = "[unknown]: ";
#if !defined(_MSC_VER)
        cur_color  = str_color[0];
#endif
        break;
    }
#if defined(_MSC_VER)
    xrd_set_font_color(i_log_level); /* set color */
    fprintf(stdout, "%s%s", psz_prefix, psz_fmt);
    xrd_set_font_color(4);     /* restore to white color */
#else
    if (i_log_level != XRD_LOG_INFO && i_log_level != XRD_LOG_NOPREFIX) {
        fprintf(stdout, "%s%s%s%s", cur_color, psz_prefix, psz_fmt, str_color_clear);
    } else {
        fprintf(stdout, "%s%s", psz_prefix, psz_fmt);
    }
#endif
}

/* ---------------------------------------------------------------------------
 */
void xrd_log(void *p, int i_log_level, const char *psz_fmt, ...)
{
    xrd_t *h = (xrd_t *)p;
    if (h == NULL || (i_log_level & 0x0F) <= h->param.i_log_level) {
        va_list arg;
        char str_in[2048];
        va_start(arg, psz_fmt);
        vsprintf(str_in, psz_fmt, arg);
        xrd_log_default(i_log_level, str_in);
        va_end(arg);
    }
}


/**
 * ===========================================================================
 * utilities
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * get time
 */
int64_t xrd_mdate(void)
{
#if SYS_WINDOWS
    LARGE_INTEGER nFreq;
    if (QueryPerformanceFrequency(&nFreq)) { // طʾӲָ֧߾ȼ
        LARGE_INTEGER t1;
        QueryPerformanceCounter(&t1);
        return (int64_t)(1000000 * t1.QuadPart / (double)nFreq.QuadPart);
    } else {  // Ӳ֧£ʹú뼶ϵͳʱ
        struct timeb tb;
        ftime(&tb);
        return ((int64_t)tb.time * 1000 + (int64_t)tb.millitm) * 1000;
    }
#else
    struct timeval tv_date;
    gettimeofday(&tv_date, NULL);
    return (int64_t)tv_date.tv_sec * 1000000 + (int64_t)tv_date.tv_usec;
#endif
}


/**
 * ===========================================================================
 * win32 thread
 * ===========================================================================
 */

#if HAVE_WIN32THREAD || PTW32_STATIC_LIB
/* ---------------------------------------------------------------------------
 * state of the threading library being initialized
 */
static volatile LONG xrd_threading_is_init = 0;

static void xrd_threading_destroy(void)
{
#if PTW32_STATIC_LIB
    pthread_win32_thread_detach_np();
    pthread_win32_process_detach_np();
#else
    xrd_win32_threading_destroy();
#endif
}

/* ---------------------------------------------------------------------------
 */
int xrd_threading_init(void)
{
    /* if already init, then do nothing */
    if (InterlockedCompareExchange(&xrd_threading_is_init, 1, 0)) {
        return 0;
    }
#if PTW32_STATIC_LIB
    /* if static pthread-win32 is already initialized, then do nothing */
    if (ptw32_processInitialized) {
        return 0;
    }
    if (!pthread_win32_process_attach_np()) {
        return -1;
    }
#else
    if (xrd_win32_threading_init()) {
        return -1;
    }
#endif

    /* register cleanup to run at process termination */
    atexit(xrd_threading_destroy);

    return 0;
}
#endif

/* ---------------------------------------------------------------------------
 */
int xrd_create_thread(xrd_pthread_t *tid, xrd_tfunc_t tfunc, void *targ)
{
    if (g_funcs.set_deblock_const != NULL) {
        /* set deblock constants for each RDO thread */
        g_funcs.set_deblock_const();
    }
//#if HAVE_WIN32THREAD
    return xrd_pthread_create(tid, NULL, tfunc, targ);
//#else
//    xrd_pthread_attr_t attr;
//    int ret_code;
//
//    /* init thread attribute object & set detach state */
//    xrd_pthread_attr_init(&attr);
//    xrd_pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
//
//    /* create thread */
//    ret_code = xrd_pthread_create(tid, &attr, tfunc, targ);
//
//    /* destroy thread attribute object */
//    xrd_pthread_attr_destroy(&attr);
//    
//    return ret_code;
//#endif
}



