/*
 * cudata.h
 *
 * Description of this file:
 *    CU-Data functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#ifndef __XRD_CUDATA_H__
#define __XRD_CUDATA_H__

void cu_get_mvds(xrd_t *h, cu_t *p_cu);

void lcu_start_init_pos   (xrd_t *h, int i_lcu_x, int i_lcu_y);
void lcu_start_init_pixels(xrd_t *h, int i_lcu_x, int i_lcu_y);
void lcu_end(xrd_t *h, int i_lcu_x, int i_lcu_y);


/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE int clip_qp(xrd_t *h, int i_qp)
{
    /* AVS2-P2 ͼ  picture_qp */
    int max_qp = MAX_QP + (h->param.sample_bit_depth - 8) * 8;
    return XRD_MAX(MIN_QP, XRD_MIN(max_qp, i_qp));
}

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE void cu_init_transform_block(int i_cu_level, int i_tu_split, int b8, cb_t *p_tb)
{
    static const cb_t TRANS_BLOCK_INFO[TU_SPLIT_TYPE_NUM][4] = {// [tu_split_type][block]
        //    x, y, w, h          x, y, w, h          x, y, w, h          x, y, w, h           for block 0, 1, 2 and 3
        { { { 0, 0, 8, 8 } }, { { 0, 0, 0, 0 } }, { { 0, 0, 0, 0 } }, { { 0, 0, 0, 0 } } }, // 0: 8x8, ---, ---, --- (TU_SPLIT_NON   )
        { { { 0, 0, 8, 2 } }, { { 0, 2, 8, 2 } }, { { 0, 4, 8, 2 } }, { { 0, 6, 8, 2 } } }, // 2: 8x2, 8x2, 8x2, 8x2 (TU_SPLIT_HOR   )
        { { { 0, 0, 2, 8 } }, { { 2, 0, 2, 8 } }, { { 4, 0, 2, 8 } }, { { 6, 0, 2, 8 } } }, // 3: 2x8, 2x8, 2x8, 2x8 (TU_SPLIT_VER   )
        { { { 0, 0, 4, 4 } }, { { 4, 0, 4, 4 } }, { { 0, 4, 4, 4 } }, { { 4, 4, 4, 4 } } }  // 1: 4x4, 4x4, 4x4, 4x4 (TU_SPLIT_CROSS )
    };
    static const cb_t CHROMAW_BLOCK_INFO[2] = {
        { { 0, 8, 4, 4 } }, { { 4, 8, 4, 4 } }
    };
    const int shift_bits = (i_cu_level - MIN_CU_SIZE_IN_BIT);

    if (b8 < 4) {
        p_tb->v = TRANS_BLOCK_INFO[i_tu_split][b8].v << shift_bits;
    } else {
        p_tb->v = CHROMAW_BLOCK_INFO[b8 - 4].v << shift_bits;
    }
}


/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE int cu_get_qp(const xrd_t *h, const cu_info_t *p_cu_info)
{
    UNUSED_PARAMETER(h);
    return p_cu_info->i_cu_qp;
}


/* ---------------------------------------------------------------------------
 * 
 */
static ALWAYS_INLINE
int cu_get_slice_index(const xrd_t *h, int scu_x, int scu_y)
{
    int lcu_shift = (h->i_lcu_level - MIN_CU_SIZE_IN_BIT);
    int lcu_xy = (scu_y >> lcu_shift) * h->i_width_in_lcu
               + (scu_x >> lcu_shift);
    return h->lcu_slice_idx[lcu_xy];
}


/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE int cu_get_chroma_qp(xrd_t *h, int luma_qp, int uv)
{
    int QP;
    UNUSED_PARAMETER(uv);

    int bit_depth_offset = (h->param.sample_bit_depth - 8) << 3;
    luma_qp -= bit_depth_offset;
    luma_qp = luma_qp < 0 ? luma_qp : tab_qp_scale_chroma[luma_qp];
    QP  = XRD_CLIP3(0, 63 + bit_depth_offset, luma_qp + bit_depth_offset);
    return QP;
}

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE
cu_layer_t *cu_get_layer(xrd_t *h, int i_cu_level)
{
    return &h->lcu.cu_layer[i_cu_level - MIN_CU_SIZE_IN_BIT];
}

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE
cu_mode_t *cu_get_layer_mode(xrd_t *h, int i_cu_level)
{
    return &h->lcu.cu_layer[i_cu_level - MIN_CU_SIZE_IN_BIT].cu_mode;
}

static ALWAYS_INLINE
cu_parallel_t *cu_get_enc_context(xrd_t *h, int i_cu_level)
{
#if PARALLEL_INSIDE_CTU
    return &h->lcu.cu_enc[i_cu_level - MIN_CU_SIZE_IN_BIT];
#else
    UNUSED_PARAMETER(i_cu_level);
    return &h->lcu.cu_enc[0];
#endif
}


#endif  // __XRD_CUDATA_H__
