/*
 * intra.c
 *
 * Description of this file:
 *    Intra prediction functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"
#include "block_info.h"
#include "cudata.h"
#include "cpu.h"

// ---------------------------------------------------------------------------
// disable warning
#if defined(_MSC_VER) || defined(__ICL)
#pragma warning(disable: 4100)  // unreferenced formal parameter
#endif


/**
 * ===========================================================================
 * local tables
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
static const int16_t tab_log2size[MAX_CU_SIZE + 1] = {
    -1, -1, -1, -1,  2, -1, -1, -1,
     3, -1, -1, -1, -1, -1, -1, -1,
     4, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
     5, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1,
    6
};

/* ---------------------------------------------------------------------------
 */
static const char tab_auc_dir_dx[NUM_INTRA_MODE] = {
      0, 0,  0,
     11, 2, 11, 1,  8, 1,  4, 1,  1,                    /* X  */
      0,
      1, 1,  4, 1,  8, 1, 11, 2, 11, 4, 8,              /* XY */
      0,
      8, 4, 11, 2, 11, 1,  8, 1                         /* Y  */
};

/* ---------------------------------------------------------------------------
 */
static const char tab_auc_dir_dy[NUM_INTRA_MODE] = {
      0,  0,  0,
     -4, -1, -8, -1, -11, -2, -11, -4, -8,              /* X  */
      0,
      8,  4, 11,  2,  11,  1,   8,  1,  4,  1,  1,      /* XY */
      0,
     -1, -1, -4, -1,  -8, -1, -11, -2                   /* Y  */
};

/* ---------------------------------------------------------------------------
 */
static const char tab_auc_dir_dxdy[2][NUM_INTRA_MODE][2] = {
    {
        // dx/dy
        { 0,0}, {0,0}, { 0,0},
        {11,2}, {2,0}, {11,3}, {1,0}, {93,7}, {1,1}, {93,8}, {1,2}, { 1,3},                 /* X  */
        { 0,0},
        { 1,3}, {1,2}, {93,8}, {1,1}, {93,7}, {1,0}, {11,3}, {2,0}, {11,2}, {4,0}, {8,0},   /* XY */
        { 0,0},
        { 8,0}, {4,0}, {11,2}, {2,0}, {11,3}, {1,0}, {93,7}, {1,1},                         /* Y  */
    }, {
        // dy/dx
        { 0,0}, {0,0}, { 0,0},
        {93,8}, {1,1}, {93,7}, {1,0}, {11,3}, {2,0}, {11,2}, {4,0}, { 8,0},                 /* X  */
        { 0,0},
        { 8,0}, {4,0}, {11,2}, {2,0}, {11,3}, {1,0}, {93,7}, {1,1}, {93,8}, {1,2}, {1,3},   /* XY */
        { 0,0},
        { 1,3}, {1,2}, {93,8}, {1,1}, {93,7}, {1,0}, {11,3}, {2,0}                          /* Y  */
    }
};


/**
 * ===========================================================================
 * local function definition
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
static void intra_pred_ver_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    pel_t *p_src = src + 1;
    int y;

    for (y = 0; y < bsy; y++) {
        g_funcs.fast_memcpy(dst, p_src, bsx * sizeof(pel_t));
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_hor_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    pel_t *p_src = src - 1;

    while (bsy-- != 0) {
        g_funcs.mem_repeat_p(dst, *p_src--, bsx);
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 * NOTE: dir_mode = (bAboveAvail << 8) + (bLeftAvail)
 */
static void intra_pred_dc_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    int avail_top  = dir_mode >> 8;
    int avail_left = dir_mode & 0xFF;
    int sum_left   = 0;
    int sum_above  = 0;
    int dc_value;
    int x, y;
    pel_t *p_src;

    p_src = src - 1;
    for (y = 0; y < bsy; y++) {
        sum_left += p_src[-y];
    }

    p_src = src + 1;
    for (x = 0; x < bsx; x++) {
        sum_above += p_src[x];
    }

    if (avail_left && avail_top) {
        x = bsx + bsy;
        dc_value = ((sum_left + sum_above + (x >> 1)) * (512 / x)) >> 9;
    } else if (avail_left) {
        dc_value = (sum_left  + (bsy >> 1)) >> xrd_log2u(bsy);
    } else if (avail_top) {
        dc_value = (sum_above + (bsx >> 1)) >> xrd_log2u(bsx);
    } else {
        dc_value = g_dc_value;
    }

    for (y = 0; y < bsy; y++) {
        g_funcs.mem_repeat_p(dst, (pel_t)dc_value, bsx);
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_plane_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    /*                 size in bits:       2   3   4   5   6 */
    static const int ib_mult [8] = { 0, 0, 13, 17,  5, 11, 23, 0 };
    static const int ib_shift[8] = { 0, 0,  7, 10, 11, 15, 19, 0 };
    const int mult_h  = ib_mult [tab_log2size[bsx]];
    const int mult_v  = ib_mult [tab_log2size[bsy]];
    const int shift_h = ib_shift[tab_log2size[bsx]];
    const int shift_v = ib_shift[tab_log2size[bsy]];
    const int W2   = bsx >> 1;              /* half block width */
    const int H2   = bsy >> 1;              /* half block height */
    const int vmax = (1 << g_bit_depth) - 1;  /* max value of pixel */
    int H = 0;
    int V = 0;
    int a, b, c;
    int x, y;
    pel_t *p_src;

    /* calculate H and V */
    p_src = src + W2;
    for (x = 1; x < W2 + 1; x++) {
        H += x * (p_src[x] - p_src[-x]);
    }
    p_src = src - H2;
    for (y = 1; y < H2 + 1; y++) {
        V += y * (p_src[-y] - p_src[y]);
    }

    a  = (src[-bsy] + src[bsx]) << 4;
    b  = ((H << 5) * mult_h + (1 << (shift_h - 1))) >> shift_h;
    c  = ((V << 5) * mult_v + (1 << (shift_v - 1))) >> shift_v;
    a += 16 - b * (W2 - 1) - c * (H2 - 1);

    for (y = 0; y < bsy; y++) {
        int pix = a;
        for (x = 0; x < bsx; x++) {
            dst[x] = (pel_t)XRD_CLIP3(0, vmax, pix >> 5);
            pix   += b;
        }
        dst += i_dst;
        a   += c;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_bilinear_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    itr_t pTop[MAX_CU_SIZE], pLeft[MAX_CU_SIZE], pT[MAX_CU_SIZE], pL[MAX_CU_SIZE], wy[MAX_CU_SIZE];
    int shift_x  = tab_log2size[bsx];
    int shift_y  = tab_log2size[bsy];
    int shift    = XRD_MIN(shift_x, shift_y);
    int shift_xy = shift_x + shift_y + 1;
    int offset   = 1 << (shift_x + shift_y);
    int vmax     = max_pel_value;    // max value of pixel
    int a, b, c, t, wxy, temp;
    int predx, val;
    int x, y;
    pel_t *p_src;

    p_src = src + 1;
    for (x = 0; x < bsx; x++) {
        pTop[x] = p_src[x];
    }
    p_src = src - 1;
    for (y = 0; y < bsy; y++) {
        pLeft[y] = p_src[-y];
    }

    a = pTop [bsx - 1];
    b = pLeft[bsy - 1];
    c = (bsx == bsy) ? (a + b + 1) >> 1 : (((a << shift_x) + (b << shift_y)) * 13 + (1 << (shift + 5))) >> (shift + 6);
    t = (c << 1) - a - b;

    for (x = 0; x < bsx; x++) {
        pT  [x]   = (itr_t)(b - pTop[x]);
        pTop[x] <<= shift_y;
    }

    temp = 0;
    for (y = 0; y < bsy; y++) {
        pL   [y]   = (itr_t)(a - pLeft[y]);
        pLeft[y] <<= shift_x;
        wy   [y]   = (itr_t)temp;
        temp      += t;
    }

    for (y = 0; y < bsy; y++) {
        predx = pLeft[y];
        wxy   = -wy[y];
        for (x = 0; x < bsx; x++) {
            predx   += pL[y];
            wxy     += wy[y];
            pTop[x] += pT[x];
            val      = ((predx << shift_y) + (pTop[x] << shift_x) + wxy + offset) >> shift_xy;
            dst[x]   = (pel_t)XRD_CLIP3(0, vmax, val);
        }
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE int get_context_pixel(int dir_mode, int xy_flag, int temp_d, int *offset)
{
    const int mult  = tab_auc_dir_dxdy[xy_flag][dir_mode][0];
    const int shift = tab_auc_dir_dxdy[xy_flag][dir_mode][1];
    int temp_dn;

    temp_d *= mult;
    temp_dn = temp_d >> shift;
    *offset = ((temp_d << 5) >> shift) - (temp_dn << 5);

    return temp_dn;
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_ang_x_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    int f0, f1, f2, f3;
    int i, j;
    int iX;

    for (j = 0; j < bsy; j++) {
        iX = get_context_pixel(dir_mode, 0, j + 1, &f3);
        f0 = 32 - f3;
        f1 = 64 - f3;
        f2 = 32 + f3;

        for (i = 0; i < bsx; i++) {
            dst[i] = (pel_t)((src[iX    ] * f0 +
                              src[iX + 1] * f1 +
                              src[iX + 2] * f2 +
                              src[iX + 3] * f3 + 64) >> 7);
            iX++;
        }

        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_ang_y_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    int offsets[64];
    int xsteps[64];
    int offset;
    int i, j;
    int iY;

    for (i = 0; i < bsx; i++) {
        xsteps[i] = get_context_pixel(dir_mode, 1, i + 1, &offsets[i]);
    }

    for (j = 0; j < bsy; j++) {
        for (i = 0; i < bsx; i++) {
            iY     = j + xsteps[i];
            offset = offsets[i];
            dst[i] = (pel_t)((src[-iY    ] * (32 - offset) +
                              src[-iY - 1] * (64 - offset) +
                              src[-iY - 2] * (32 + offset) +
                              src[-iY - 3] * (     offset) + 64) >> 7);
        }
        dst += i_dst;
    }
}

/* ---------------------------------------------------------------------------
 */
static void intra_pred_ang_xy_c(pel_t *src, pel_t *dst, int i_dst, int dir_mode, int bsx, int bsy)
{
    ALIGN16(int xoffsets[64]);
    ALIGN16(int xsteps[64]);
    int i, j, iXx, iYy;
    int offsetx, offsety;

    for (i = 0; i < bsx; i++) {
        xsteps[i] = get_context_pixel(dir_mode, 1, i + 1, &xoffsets[i]);
    }

    for (j = 0; j < bsy; j++) {
        iXx = -get_context_pixel(dir_mode, 0, j + 1, &offsetx);

        for (i = 0; i < bsx; i++) {
            iYy = j - xsteps[i];

            if (iYy <= -1) {
                dst[i] = (pel_t)((src[ iXx + 2] * (32 - offsetx) +
                                  src[ iXx + 1] * (64 - offsetx) +
                                  src[ iXx    ] * (32 + offsetx) +
                                  src[ iXx - 1] * (     offsetx) + 64) >> 7);
            } else {
                offsety = xoffsets[i];
                dst[i] = (pel_t)((src[-iYy - 2] * (32 - offsety) +
                                  src[-iYy - 1] * (64 - offsety) +
                                  src[-iYy    ] * (32 + offsety) +
                                  src[-iYy + 1] * (     offsety) + 64) >> 7);
            }
            iXx++;
        }
        dst += i_dst;
    }
}


/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCUϱ߽PU
 */
static 
void fill_reference_samples_0_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    int num_padding = 0;

    /* fill default value */
    g_funcs.mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        g_funcs.fast_memcpy(&EP[1], &pLcuEP[1], bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        g_funcs.fast_memcpy(&EP[bsx + 1], &pLcuEP[bsx + 1], bsx * sizeof(pel_t));
    } else {
        g_funcs.mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        /* fill left pixels */
        memcpy(&EP[-bsy], &pLcuEP[-bsy], bsy * sizeof(pel_t));
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        memcpy(&EP[-2 * bsy], &pLcuEP[-2 * bsy], bsy * sizeof(pel_t));
    } else {
        g_funcs.mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pLcuEP[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pLcuEP[1];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pLcuEP[-1];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCUϱ߽PU
 */
static 
void fill_reference_samples_x_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    const pel_t *pL = pTL + i_TL;
    int num_padding = 0;

    /* fill default value */
    g_funcs.mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        g_funcs.fast_memcpy(&EP[1], &pLcuEP[1], bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        g_funcs.fast_memcpy(&EP[bsx + 1], &pLcuEP[bsx + 1], bsx * sizeof(pel_t));
    } else {
        g_funcs.mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        const pel_t *p_l = pL;
        int y;
        /* fill left pixels */
        for (y = 0; y < bsy; y++) {
            EP[-1 - y] = *p_l;
            p_l += i_TL;
        }
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        int y;
        const pel_t *p_l = pL + bsy * i_TL;

        for (y = 0; y < bsy; y++) {
            EP[-bsy - 1 - y] = *p_l;
            p_l += i_TL;
        }
    } else {
        g_funcs.mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pLcuEP[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pLcuEP[1];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pL[0];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCU߽ϵPU
 */
static 
void fill_reference_samples_y_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    const pel_t *pT = pTL + 1;
    int num_padding = 0;

    /* fill default value */
    g_funcs.mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        g_funcs.fast_memcpy(&EP[1], pT, bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        g_funcs.fast_memcpy(&EP[bsx + 1], &pT[bsx], bsx * sizeof(pel_t));
    } else {
        g_funcs.mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        /* fill left pixels */
        memcpy(&EP[-bsy], &pLcuEP[-bsy], bsy * sizeof(pel_t));
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        memcpy(&EP[-2 * bsy], &pLcuEP[-2 * bsy], bsy * sizeof(pel_t));
    } else {
        g_funcs.mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pLcuEP[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pT[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pLcuEP[-1];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/* ---------------------------------------------------------------------------
 * fill reference samples for intra prediction
 * LCUڲڱ߽ϵPU
 */
static 
void fill_reference_samples_xy_c(const pel_t *pTL, int i_TL, const pel_t *pLcuEP, pel_t *EP, uint32_t i_avai, int bsx, int bsy)
{
    const pel_t *pT = pTL + 1;
    const pel_t *pL = pTL + i_TL;
    int num_padding = 0;

    /* fill default value */
    g_funcs.mem_repeat_p(&EP[-(bsy << 1)], g_dc_value, ((bsy + bsx) << 1) + 1);

    /* get prediction pixels ---------------------------------------
     * extra pixels          | left-down pixels   | left pixels   | top-left | top pixels  | top-right pixels  | extra pixels
     * -2*bsy-4 ... -2*bsy-1 | -bsy-bsy ... -bsy-1| -bsy -3 -2 -1 |     0    | 1 2 ... bsx | bsx+1 ... bsx+bsx | 2*bsx+1 ... 2*bsx+4
     */

    /* fill top & top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        /* fill top pixels */
        g_funcs.fast_memcpy(&EP[1], pT, bsx * sizeof(pel_t));
    }

    /* fill top-right pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_RIGHT)) {
        g_funcs.fast_memcpy(&EP[bsx + 1], &pT[bsx], bsx * sizeof(pel_t));
    } else {
        g_funcs.mem_repeat_p(&EP[bsx + 1], EP[bsx], bsx);   // repeat the last pixel
    }

    /* fill extra pixels */
    num_padding = bsy * 11 / 4 - bsx + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[2 * bsx + 1], EP[2 * bsx], num_padding); // from (2*bsx) to (iX + 3) = (bsy *11/4 + bsx - 1) + 3
    }

    /* fill left & left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        const pel_t *p_l = pL;
        int y;
        /* fill left pixels */
        for (y = 0; y < bsy; y++) {
            EP[-1 - y] = *p_l;
            p_l += i_TL;
        }
    }

    /* fill left-down pixels */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT_DOWN)) {
        int y;
        const pel_t *p_l = pL + bsy * i_TL;

        for (y = 0; y < bsy; y++) {
            EP[-bsy - 1 - y] = *p_l;
            p_l += i_TL;
        }
    } else {
        g_funcs.mem_repeat_p(&EP[-(bsy << 1)], EP[-bsy], bsy);
    }

    /* fill top-left pixel */
    if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP_LEFT)) {
        EP[0] = pTL[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_TOP)) {
        EP[0] = pT[0];
    } else if (IS_NEIGHBOR_AVAIL(i_avai, MD_I_LEFT)) {
        EP[0] = pL[0];
    }

    /* fill extra pixels */
    num_padding = bsx * 11 / 4 - bsy + 4;
    if (num_padding > 0) {
        g_funcs.mem_repeat_p(&EP[-2 * bsy - num_padding], EP[-2 * bsy], num_padding); // from (-2*bsy) to (-iY - 3) = -(bsx *11/4 + bsy - 1) - 3
    }
}

/**
 * ===========================================================================
 * interface function definition
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
void xrd_intra_pred_init(intrinsic_func_t *pf)
{
#define ANG_X_OFFSET    3
#define ANG_XY_OFFSET   13
#define ANG_Y_OFFSET    25
    int i;

    intra_pred_t *ipred = pf->intraf;

    pf->fill_edge_f[0] = fill_reference_samples_0_c;
    pf->fill_edge_f[1] = fill_reference_samples_x_c;
    pf->fill_edge_f[2] = fill_reference_samples_y_c;
    pf->fill_edge_f[3] = fill_reference_samples_xy_c;
    ipred[DC_PRED   ] = intra_pred_dc_c;                // 0
    ipred[PLANE_PRED] = intra_pred_plane_c;             // 1
    ipred[BI_PRED   ] = intra_pred_bilinear_c;          // 2

    for (i = ANG_X_OFFSET; i < VERT_PRED; i++) {
        ipred[i     ] = intra_pred_ang_x_c;             // 3 ~ 11
    }
    ipred[VERT_PRED ] = intra_pred_ver_c;               // 12

    for (i = ANG_XY_OFFSET; i < HOR_PRED; i++) {
        ipred[i     ] = intra_pred_ang_xy_c;            // 13 ~ 23
    }

    ipred[HOR_PRED  ] = intra_pred_hor_c;               // 24
    for (i = ANG_Y_OFFSET; i < NUM_INTRA_MODE; i++) {
        ipred[i     ] = intra_pred_ang_y_c;             // 25 ~ 32
    }
#undef ANG_X_OFFSET
#undef ANG_XY_OFFSET
#undef ANG_Y_OFFSET
}
