/*
 * aec_bitest.c
 *
 * Description of this file:
 *    AEC functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"
#include "aec.h"
#include "bitstream.h"
#include "block_info.h"

#if OPT_FAST_BITS_EST
/* ---------------------------------------------------------------------------
 * sqrt table */
static const uint8_t sqrt_tab[256] = {
    181, 180, 180, 179, 178, 177, 177, 176, 175, 175, 174, 173, 172, 172, 171, 170,
    169, 169, 168, 167, 166, 165, 165, 164, 163, 162, 162, 161, 160, 159, 158, 158,
    157, 156, 155, 154, 153, 153, 152, 151, 150, 149, 148, 148, 147, 146, 145, 144,
    143, 142, 141, 140, 140, 139, 138, 137, 136, 135, 134, 133, 132, 131, 130, 129,
    128, 127, 126, 125, 124, 123, 122, 121, 120, 119, 118, 116, 115, 114, 113, 112,
    111, 110, 108, 107, 106, 105, 104, 102, 101, 100,  99,  97,  96,  95,  93,  92,
     91,  89,  88, 86,   85,  83,  82,  80,  78,  77,  75,  73,  72,  70,  68,  66,
     64,  62,  60, 58,   55,  53,  51,  48,  45,  42,  39,  36,  32,  28,  23,  16,
      0,  16,  23, 28,   32,  36,  39,  42,  45,  48,  51,  53,  55,  58,  60,  62,
     64,  66,  68, 70,   72,  73,  75,  77,  78,  80,  82,  83,  85,  86,  88,  89,
     91,  92,  93, 95,   96,  97,  99, 100, 101, 102, 104, 105, 106, 107, 108, 110,
    111, 112, 113, 114, 115, 116, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127,
    128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 140, 141, 142,
    143, 144, 145, 146, 147, 148, 148, 149, 150, 151, 152, 153, 153, 154, 155, 156,
    157, 158, 158, 159, 160, 161, 162, 162, 163, 164, 165, 165, 166, 167, 168, 169,
    169, 170, 171, 172, 172, 173, 174, 175, 175, 176, 177, 177, 178, 179, 180, 180
};

/**
 * ===========================================================================
 * B-EST(bit-estimator) model (used for fast bits estimation)
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * get self-information for B-EST(bit-estimator)
 */
int cavs2_best_self_information(coeff_t *p_qcoef, int stride, int bsize)
{
    const uint8_t* tab = sqrt_tab + 128;
    int info = stride + 8;      /* this value will be divided by 16.0f */
    int i, j;

    for (j = 0; j < bsize; j++) {
        for (i = 0; i < bsize; i++) {
            info += *(tab + (char)p_qcoef[i]);
        }
        p_qcoef += stride;
    }

    return info;
}

/* ---------------------------------------------------------------------------
 * get self-information for B-EST(bit-estimator)
 */
int cavs2_best_self_information_on_level(coeff_t *p_level, int pairs, int bsize)
{
    const uint8_t* tab = sqrt_tab + 128;
    int info = bsize + 8;       /* this value will be divided by 16.0f */
    int i;

    for (i = 0; i < pairs; i++) {
        info += *(tab + (char)p_level[i]);
    }

    return info;
}

/* ---------------------------------------------------------------------------
 * init B-EST(bit-estimator)
 */
void cavs2_best_init(bit_est_t *p_best)
{
    float x  = 1.0;
    float xx = 1.0;
    float y  = 0.4f;
    float xy = 0.4f;
    int i;

    /* set default parameters */
    p_best->pos       = 0;
    p_best->alpha     = 4.0f;   /* default */
    p_best->beta      = 0.0f;

    p_best->sum_bits  = 0.0f;
    p_best->sum_bits2 = 0.0f;
    p_best->sum_info  = 0.0f;
    p_best->sum_info2 = 0.0f;
    for (i = 0; i < CAVS2_BEST_NUM; i++) {
        p_best->info [i] =  x;
        p_best->info2[i] = xx;
        p_best->bits [i] =  y;
        p_best->bits2[i] = xy;

        p_best->sum_info  +=  x;
        p_best->sum_info2 += xx;
        p_best->sum_bits  +=  y;
        p_best->sum_bits2 += xy;
    }
}

/* ---------------------------------------------------------------------------
 * update the B-EST(bit-estimator) model after coding one TU
 */
static void cavs2_best_update(bit_est_t *p_best, int bits, float info)
{
    float d;                    /* denominator */
    float x  = info;
    float xx = info * info;
    float y  = (float)bits;
    float xy = info * bits;
    const int pos = p_best->pos;      /* current position */

    /* -------------------------------------------------------------
     * update the data in slide window
     */

    /* remove the oldest data in current position */
    p_best->sum_info  -= p_best->info [pos];
    p_best->sum_info2 -= p_best->info2[pos];
    p_best->sum_bits  -= p_best->bits [pos];
    p_best->sum_bits2 -= p_best->bits2[pos];

    /* fill the new data */
    p_best->info [pos] = x;
    p_best->info2[pos] = xx;
    p_best->bits [pos] = y;
    p_best->bits2[pos] = xy;

    /* update the sum and counter */
    p_best->sum_info  += x;
    p_best->sum_info2 += xx;
    p_best->sum_bits  += y;
    p_best->sum_bits2 += xy;

    /* point to next position and reset the array element */
    p_best->pos = (p_best->pos + 1) & (CAVS2_BEST_NUM - 1);

    /* -------------------------------------------------------------
     * update the model parameters
     */

    // n(a^2+b^2+c^2...) >= (a+b+c...)^2 ==> d >= 0
    d = CAVS2_BEST_NUM * p_best->sum_info2 - p_best->sum_info * p_best->sum_info;

    if (d > 10.0f) {
        p_best->alpha = (p_best->sum_bits2 * CAVS2_BEST_NUM - p_best->sum_info * p_best->sum_bits ) / d;
        p_best->beta  = (p_best->sum_info2 * p_best->sum_bits - p_best->sum_info * p_best->sum_bits2) / d;
    }
}
#endif
