/*
 * encoder.h
 *
 * Description of this file:
 *    Encoder functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#ifndef __XRD_ENCODER_H__
#define __XRD_ENCODER_H__

#include "xrdenc.h"

/**
 * ===========================================================================
 * interface function defines (pre-encode processing)
 * ===========================================================================
 */

int      send_frame_to_enc_queue(xrd_handler_t *h_mgr, xrd_frame_t *frm);

void     xrde_get_frame_lambda(xrd_t *h, xrd_frame_t *cur_frm, int i_qp);

int      xrde_refine_qp(xrd_t *h);

/**
 * ===========================================================================
 * interface function defines (encode processing)
 * ===========================================================================
 */

int      encoder_check_parameters(xrd_param_t *param);

xrd_t *encoder_open  (xrd_param_t *param, xrd_handler_t *h_mgr);
int      encoder_encode(xrd_handler_t *h_mgr, xrd_frame_t *frame);
void     encoder_close (xrd_handler_t *h_mgr);

int      encoder_contexts_init(xrd_t *h, xrd_handler_t *h_mgr);
void     encoder_write_rec_frame(xrd_handler_t *h_mgr);
void     encoder_fetch_one_encoded_frame(xrd_handler_t *h_mgr, xrd_outpacket_t *packet, int is_flush);

void     xrd_reconfigure_encoder(xrd_t *h);

/**
 * ===========================================================================
 * API interface
 * ===========================================================================
 */
/**
 * ---------------------------------------------------------------------------
 * Function   : Output help parameters
 * Parameters :
 *      [in ] : param - pointer to struct xrd_param_t
 *      [out] : none
 * Return     : none
 * ---------------------------------------------------------------------------
 */
void xrd_encoder_opt_help(void);

/**
 * ---------------------------------------------------------------------------
 * Function   : initialize default parameters for the XRDENC video encoder
 * Parameters :
 *      [in ] : none
 * Return     : parameter handler, can be further configured
 * ---------------------------------------------------------------------------
 */
xrd_param_t *xrd_encoder_opt_alloc(void);

/**
 * ---------------------------------------------------------------------------
 * Function   : Parsing encoding parameters
 * Parameters :
 *      [in ] : param - pointer to struct xrd_param_t
 *      [in ] : argc  - number of command line parameters
 *      [in ] : argv  - pointer to parameter strings
 * Return     : int   - zero for success, otherwise failed
 * ---------------------------------------------------------------------------
 */
int xrd_encoder_opt_set(xrd_param_t *param, int argc, char *argv[]);

/**
 * ---------------------------------------------------------------------------
 * Function   : Parsing encoding parameters
 * Parameters :
 *      [in ] : param - pointer to struct xrd_param_t
 *      [in ] : name  - name of parameter
 *      [in ] : value_string - parameter value
 *      [in ] : value_i      - when value_string is null, use this value
 * Return     : int   - zero for success, otherwise failed
 * ---------------------------------------------------------------------------
 */
int xrd_encoder_opt_set2(xrd_param_t *param, const char *name, const char *value_string);

/**
 * ---------------------------------------------------------------------------
 * Function   : get value of a specific parameter
 * Parameters :
 *      [in ] : param - pointer to struct xrd_param_t
 *      [in ] : name  - name of a parameter (input, output, width, height, frames)
 * Return     : const char *: value string
 * ---------------------------------------------------------------------------
 */
const char *xrd_encoder_opt_get(xrd_param_t *param, const char *name);

/**
 * ---------------------------------------------------------------------------
 * Function   : free memory of parameter
 * Parameters :
 *      [in ] : none
 *      [out] : parameter handler, can be further configured
 * Return     : none
 * ---------------------------------------------------------------------------
 */
void xrd_encoder_opt_destroy(xrd_param_t *param);


/**
 * ===========================================================================
 * interface function declares: encoding
 * ===========================================================================
 */

/**
 * ---------------------------------------------------------------------------
 * Function   : get buffer for the encoder caller
 * Parameters :
 *      [in ] : coder - pointer to wrapper of the XRDENC encoder
 *            : pic   - pointer to struct xrd_picture_t
 *      [out] : none
 * Return     : zero for success, otherwise failed
 * ---------------------------------------------------------------------------
 */
int xrd_encoder_get_buffer(void *coder, xrd_picture_t *pic);

/**
 * ---------------------------------------------------------------------------
 * Function   : create and initialize the XRDENC video encoder
 * Parameters :
 *      [in ] : param     - pointer to struct xrd_param_t
 *            : dump_func - pointer to struct xrd_dump_func_t
 *            : opaque    - user data
 *      [out] : none
 * Return     : handle of XRDENC encoder wrapper, none zero for success, otherwise false
 * ---------------------------------------------------------------------------
 */
#if XRDENC_API_VERSION < 2
void *xrd_encoder_create(xrd_param_t *param, xrd_dump_func_t dump_func, void *opaque);
#else
void *xrd_encoder_create(xrd_param_t *param);
#endif

/**
 * ---------------------------------------------------------------------------
 * Function   : destroy the XRDENC video encoder
 * Parameters :
 *      [in ] : coder - pointer to wrapper of the XRDENC encoder
 *      [out] : none
 * Return     : none
 * Note       : this API is *NOT* thread-safe, 
 *              and can not be called simultaneously with other APIs.
 * ---------------------------------------------------------------------------
 */
void xrd_encoder_destroy(void *coder);

/**
 * ---------------------------------------------------------------------------
 * Function   : write (send) data to the XRDENC encoder
 * Parameters :
 *      [in ] : coder - pointer to wrapper of the XRDENC encoder
 *            : pic   - pointer to struct xrd_picture_t
 *      [out] : none
 * Return     : zero for success, otherwise failed
 * ---------------------------------------------------------------------------
 */

#if XRDENC_API_VERSION < 2
int xrd_encoder_encode(void *coder, xrd_picture_t *pic);
#else
int xrd_encoder_encode(void *coder, xrd_picture_t *pic, xrd_outpacket_t *packet);
#endif

/**
 * ---------------------------------------------------------------------------
 * Function   : label a packet to be recycled
 * Parameters :
 *      [in ] : coder    - pointer to wrapper of the XRDENC encoder
 *            : packet   - pointer to struct xrd_outpacket_t
 *      [out] : none
 * Return     : zero for success, otherwise failed
 * ---------------------------------------------------------------------------
 */
int xrd_encoder_packet_unref(void *coder, xrd_outpacket_t *packet);

#endif  // __XRD_ENCODER_H__
