/*
 * me.c
 *
 * Description of this file:
 *    ME functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"
#include "block_info.h"
#include "me.h"
#include "common/cpu.h"
#include "common/mc.h"
#include "predict.h"


/**
 * ===========================================================================
 * type defines
 * ===========================================================================
 */

typedef struct mv_info {
    int     bdir;               /* best direction */
    dist_t  bcost;              /* best cost      */
    int     bdist;              /* best distance  */
} mv_info;


/**
 * ===========================================================================
 * local/global variables
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * big hexagon for UMH */
static const int8_t HEX4[16][2] = {
    {  0, -4 }, { 0,  4 }, { -2, -3 }, { 2, -3 },
    { -4, -2 }, { 4, -2 }, { -4, -1 }, { 4, -1 },
    { -4,  0 }, { 4,  0 }, { -4,  1 }, { 4,  1 },
    { -4,  2 }, { 4,  2 }, { -2,  3 }, { 2,  3 }
};

/* ---------------------------------------------------------------------------
 * radius 2 hexagon
 * repeated entries are to avoid having to compute mod6 every time */
const int8_t HEX2[8][2] = {
    {-1, -2},       /* 0,       0(6)   5        */
    {-2,  0},       /* 1,                       */
    {-1,  2},       /* 2,                       */
    { 1,  2},       /* 3,   1(7)   *       4    */
    { 2,  0},       /* 4,                       */
    { 1, -2},       /* 5,                       */
    {-1, -2},       /* 6,       2      3        */
    {-2,  0}        /* 7,                       */
};

/* ---------------------------------------------------------------------------
 * (x - 1) % 6 */
const int8_t M1MOD6[8] = { 5, 0, 1, 2, 3, 4, 5, 0 };

/* ---------------------------------------------------------------------------
 * radius 1 diamond
 * repeated entries are to avoid having to compute mod4 every time */
const int8_t DIA1[6][2] = {
    { 0, -1},       /* 0,                       */
    {-1,  0},       /* 1,           0(4)        */
    { 0,  1},       /* 2,                       */
    { 1,  0},       /* 3,      1(5) *   3       */
    { 0, -1},       /* 4,                       */
    {-1,  0}        /* 5,           2           */
};

/* ---------------------------------------------------------------------------
 * (x - 1) % 4 */
const int8_t M1MOD4[6] = { 3, 0, 1, 2, 3, 0 };

/* ---------------------------------------------------------------------------
 * uneven multi-hexagon-grid: 5x5 */
static int8_t GRID[24][2] = {
    { -1, -1 }, {  0, -1 }, { 1, -1 },  /* inside 8 points */
    { -1,  0 },             { 1,  0 },
    { -1,  1 }, {  0,  1 }, { 1,  1 },
    { -2, -2 }, { -1, -2 }, { 0,  2 }, { 1, -2 }, { 2, -2 }, /* outside 16 points */
    { -2, -1 },                                   { 2, -1 },
    { -2,  0 },                                   { 2,  0 },
    { -2,  1 },                                   { 2,  1 },
    { -2,  2 }, { -1,  2 }, { 0,  2 }, { 1,  2 }, { 2,  2 }
};

/* ---------------------------------------------------------------------------
 * ڷ */
static const int8_t Spiral[9][2] = {
    {  0,  0 }, {  0, -1 }, {  0, 1 },
    { -1, -1 }, {  1, -1 }, { -1, 0 },
    {  1,  0 }, { -1,  1 }, {  1, 1 }
};

/* ---------------------------------------------------------------------------
 * offsets for Two Point Search (TZ) */
static const int offsets[16][2] =
{
    /*    X   X    */
    /*  X 1 2 3 X  */
    /*    4 * 5    */
    /*  X 6 7 8 X  */
    /*    X   X    */
    { -1,  0 }, {  0, -1 },
    { -1, -1 }, {  1, -1 },
    { -1,  0 }, {  1,  0 },
    { -1,  1 }, { -1, -1 },
    {  1, -1 }, {  1,  1 },
    { -1,  0 }, {  0,  1 },
    { -1,  1 }, {  1,  1 },
    {  1,  0 }, {  0,  1 },
};

static const int i_org = FENC_STRIDE;

/**
 * ===========================================================================
 * macros
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * early termination */
#define EARLY_TERMINATION(pred_sad) \
    if (bcost < (pred_sad) * beta3) {\
        goto umh_step_3;\
    } else if (bcost < (pred_sad) * beta2) {\
        goto umh_step_2;\
    }


/**
 * ===========================================================================
 * calculate cost for integer pixel motion search
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
#define CAL_COST_IPEL(mx, my) \
    g_funcs.pixf.sad[i_pixel](p_org, i_org,\
        p_fref + (my) * i_fref + (mx), i_fref) + MV_COST_IPEL(mx, my)

/* ---------------------------------------------------------------------------
 */
#define ME_COST_IPEL(mx, my) \
    if (CHECK_MV_RANGE(mx, my)) {\
        int cost = g_funcs.pixf.sad[i_pixel](p_org, i_org,\
                   p_fref + (my) * i_fref + (mx), i_fref) + MV_COST_IPEL(mx, my);\
        COPY3_IF_LT(bcost, cost, bmx, mx, bmy, my);\
    }

/* ---------------------------------------------------------------------------
 */
#define ME_COST_IPEL_DIR(mx, my, d) \
    if (CHECK_MV_RANGE(mx, my)) {\
        int cost = g_funcs.pixf.sad[i_pixel](p_org, i_org,\
                   p_fref + (my) * i_fref + (mx), i_fref) + MV_COST_IPEL(mx, my);\
        COPY4_IF_LT(bcost, cost, bmx, mx, bmy, my, dir, d);\
    }

/* ---------------------------------------------------------------------------
 */
#define ME_COST_IPEL_X3(m0x, m0y, m1x, m1y, m2x, m2y) \
{\
    pel_t *pix_base = p_fref + omy * i_fref + omx;\
    g_funcs.pixf.sad_x3[i_pixel](p_org,\
        pix_base + (m0y) * i_fref + (m0x),\
        pix_base + (m1y) * i_fref + (m1x),\
        pix_base + (m2y) * i_fref + (m2x),\
        i_fref, costs);\
    costs[0] += MV_COST_IPEL(omx + (m0x), omy + (m0y));\
    costs[1] += MV_COST_IPEL(omx + (m1x), omy + (m1y));\
    costs[2] += MV_COST_IPEL(omx + (m2x), omy + (m2y));\
    COPY3_IF_LT(bcost, costs[0], bmx, omx + (m0x), bmy, omy + (m0y));\
    COPY3_IF_LT(bcost, costs[1], bmx, omx + (m1x), bmy, omy + (m1y));\
    COPY3_IF_LT(bcost, costs[2], bmx, omx + (m2x), bmy, omy + (m2y));\
}

/* ---------------------------------------------------------------------------
 */
#define ME_COST_IPEL_X3_DIR(m0x, m0y, d0, m1x, m1y, d1, m2x, m2y, d2) \
{\
    pel_t *pix_base = p_fref + omy * i_fref + omx;\
    g_funcs.pixf.sad_x3[i_pixel](p_org,\
        pix_base + (m0y) * i_fref + (m0x),\
        pix_base + (m1y) * i_fref + (m1x),\
        pix_base + (m2y) * i_fref + (m2x),\
        i_fref, costs);\
    costs[0] += MV_COST_IPEL(omx + (m0x), omy + (m0y));\
    costs[1] += MV_COST_IPEL(omx + (m1x), omy + (m1y));\
    costs[2] += MV_COST_IPEL(omx + (m2x), omy + (m2y));\
    COPY4_IF_LT(bcost, costs[0], bmx, omx + (m0x), bmy, omy + (m0y), dir, d0);\
    COPY4_IF_LT(bcost, costs[1], bmx, omx + (m1x), bmy, omy + (m1y), dir, d1);\
    COPY4_IF_LT(bcost, costs[2], bmx, omx + (m2x), bmy, omy + (m2y), dir, d2);\
}

/* ---------------------------------------------------------------------------
 */
#define ME_COST_IPEL_X4(m0x, m0y, m1x, m1y, m2x, m2y, m3x, m3y) \
{\
    if (CHECK_MV_RANGE_X4(omx + m0x, omy + m0y, omx + m1x, omy + m1y, omx + m2x, omy + m2y, omx + m3x, omy + m3y)) {  \
        pel_t *pix_base = p_fref + omy * i_fref + omx;\
        g_funcs.pixf.sad_x4[i_pixel](p_org,\
            pix_base + (m0y) * i_fref + (m0x),\
            pix_base + (m1y) * i_fref + (m1x),\
            pix_base + (m2y) * i_fref + (m2x),\
            pix_base + (m3y) * i_fref + (m3x),\
            i_fref, costs);\
        costs[0] += MV_COST_IPEL(omx + (m0x), omy + (m0y));\
        costs[1] += MV_COST_IPEL(omx + (m1x), omy + (m1y));\
        costs[2] += MV_COST_IPEL(omx + (m2x), omy + (m2y));\
        costs[3] += MV_COST_IPEL(omx + (m3x), omy + (m3y));\
        COPY3_IF_LT(bcost, costs[0], bmx, omx + (m0x), bmy, omy + (m0y));\
        COPY3_IF_LT(bcost, costs[1], bmx, omx + (m1x), bmy, omy + (m1y));\
        COPY3_IF_LT(bcost, costs[2], bmx, omx + (m2x), bmy, omy + (m2y));\
        COPY3_IF_LT(bcost, costs[3], bmx, omx + (m3x), bmy, omy + (m3y));\
    } else {                    \
        ME_COST_IPEL(omx + m0x, omy + m0y); \
        ME_COST_IPEL(omx + m1x, omy + m1y); \
        ME_COST_IPEL(omx + m2x, omy + m2y); \
        ME_COST_IPEL(omx + m3x, omy + m3y); \
    } \
}

/* ---------------------------------------------------------------------------
 */
#define ME_COST_IPEL_X4_DIR(m0x, m0y, d0, m1x, m1y, d1, m2x, m2y, d2, m3x, m3y, d3) \
{\
    pel_t *pix_base = p_fref + omy * i_fref + omx;\
    g_funcs.pixf.sad_x4[i_pixel](p_org,\
        pix_base + (m0y) * i_fref + (m0x),\
        pix_base + (m1y) * i_fref + (m1x),\
        pix_base + (m2y) * i_fref + (m2x),\
        pix_base + (m3y) * i_fref + (m3x), i_fref, costs);\
    costs[0] += MV_COST_IPEL(omx + (m0x), omy + (m0y));\
    costs[1] += MV_COST_IPEL(omx + (m1x), omy + (m1y));\
    costs[2] += MV_COST_IPEL(omx + (m2x), omy + (m2y));\
    costs[3] += MV_COST_IPEL(omx + (m3x), omy + (m3y));\
    COPY4_IF_LT(bcost, costs[0], bmx, omx + (m0x), bmy, omy + (m0y), dir, d0);\
    COPY4_IF_LT(bcost, costs[1], bmx, omx + (m1x), bmy, omy + (m1y), dir, d1);\
    COPY4_IF_LT(bcost, costs[2], bmx, omx + (m2x), bmy, omy + (m2y), dir, d2);\
    COPY4_IF_LT(bcost, costs[3], bmx, omx + (m3x), bmy, omy + (m3y), dir, d3);\
}

/* ---------------------------------------------------------------------------
 * for TZ */
#define ME_COST_IPEL_DIR_DIST(mx, my, direction, dist) \
    if (CHECK_MV_RANGE(mx, my)) {\
        dist_t cost = g_funcs.pixf.sad[i_pixel](p_org, i_org,\
                      p_fref + (my) * i_fref + (mx), i_fref) + MV_COST_IPEL(mx, my);\
        COPY5_IF_LT(mvinfo.bcost, cost, bmx, mx, bmy, my, mvinfo.bdir, direction, mvinfo.bdist, dist);\
    }

/* ---------------------------------------------------------------------------
 * for TZ */
#define ME_COST_IPEL_X4_TZ(m0x, m0y, p0, d0, m1x, m1y, p1, d1, m2x, m2y, p2, d2, m3x, m3y, p3, d3) \
{\
    if (CHECK_MV_RANGE_X4(m0x, m0y, m1x, m1y, m2x, m2y, m3x, m3y)) {  \
        g_funcs.pixf.sad_x4[i_pixel](p_org,\
            p_fref + (m0x) + (m0y) * i_fref,\
            p_fref + (m1x) + (m1y) * i_fref,\
            p_fref + (m2x) + (m2y) * i_fref,\
            p_fref + (m3x) + (m3y) * i_fref,\
            i_fref, costs);\
        (costs)[0] += MV_COST_IPEL(m0x, m0y);\
        (costs)[1] += MV_COST_IPEL(m1x, m1y);\
        (costs)[2] += MV_COST_IPEL(m2x, m2y);\
        (costs)[3] += MV_COST_IPEL(m3x, m3y);\
        COPY5_IF_LT(mvinfo.bcost, costs[0], bmx, m0x, bmy, m0y, mvinfo.bdir, p0, mvinfo.bdist, d0);\
        COPY5_IF_LT(mvinfo.bcost, costs[1], bmx, m1x, bmy, m1y, mvinfo.bdir, p1, mvinfo.bdist, d1);\
        COPY5_IF_LT(mvinfo.bcost, costs[2], bmx, m2x, bmy, m2y, mvinfo.bdir, p2, mvinfo.bdist, d2);\
        COPY5_IF_LT(mvinfo.bcost, costs[3], bmx, m3x, bmy, m3y, mvinfo.bdir, p3, mvinfo.bdist, d3);\
    } else { \
        ME_COST_IPEL_DIR_DIST(m0x, m0y, p0, d0); \
        ME_COST_IPEL_DIR_DIST(m1x, m1y, p1, d1); \
        ME_COST_IPEL_DIR_DIST(m2x, m2y, p2, d2); \
        ME_COST_IPEL_DIR_DIST(m3x, m3y, p3, d3); \
    }\
}

/* ---------------------------------------------------------------------------
 * diamond:     1
 *            1 0 1
 *              1    */
#define DIA_ITER(mx, my) \
{\
    omx = mx;\
    omy = my;\
    ME_COST_IPEL_X4(0,-1, -1,0, 1,0, 0,1);\
}


/**
 * ===========================================================================
 * calculate cost for fractional pixel refine
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
#define ME_COST_QPEL(mx, my) \
{\
    pel_t *p_pred = p_filtered[(((my) & 3) << 2) + ((mx) & 3)] + i_offset\
                  + ((my) >> 2) * i_fref + ((mx) >> 2); \
    cost = g_funcs.pixf.fpel_cmp[i_pixel](p_org, i_org, p_pred, i_fref) + MV_COST_FPEL(mx, my);\
}

/* ---------------------------------------------------------------------------
 */
#define ME_COST_QPEL_SYM \
{\
    int mx_sym;\
    int my_sym;\
    \
    cost = MAX_DISTORTION;\
    if (h->i_type == SLICE_TYPE_B) {\
        mx_sym = -scale_mv_skip  (   mx, distance_bwd, distance_fwd);\
        my_sym = -scale_mv_skip_y(h, my, distance_bwd, distance_fwd);\
    } else {\
        mx_sym = scale_mv_skip  (   mx, distance_bwd, distance_fwd);\
        my_sym = scale_mv_skip_y(h, my, distance_bwd, distance_fwd);\
    }\
    \
    if (CHECK_MV_RANGE(mx, my) && CHECK_MV_RANGE(mx_sym, my_sym)) {\
        int xx1 = mx     >> 2;\
        int yy1 = my     >> 2;\
        int xx2 = mx_sym >> 2;\
        int yy2 = my_sym >> 2;\
        pel_t *p_src1 = p_filtered1[((my     & 3) << 2) + (mx     & 3)] + i_offset + yy1 * i_fref + xx1;\
        pel_t *p_src2 = p_filtered2[((my_sym & 3) << 2) + (mx_sym & 3)] + i_offset + yy2 * i_fref + xx2;\
        pel_t *p_pred = buf_pixel_temp;\
        \
        g_funcs.pixf.avg[i_pixel](p_pred, 64, p_src1, i_fref, p_src2, i_fref, 32); \
        cost = g_funcs.pixf.fpel_cmp[i_pixel](p_org, i_org, p_pred, MAX_CU_SIZE)\
             + MV_COST_FPEL(mx, my);\
    }\
}

/* ---------------------------------------------------------------------------
 */
#define ME_COST_QPEL_BID \
    if (CHECK_MV_RANGE(mx, my) && CHECK_MV_RANGE(mx_bid, my_bid)) {\
        int xx1 = mx     >> 2;\
        int yy1 = my     >> 2;\
        pel_t *p_src1 = p_filtered1[((my     & 3) << 2) + (mx     & 3)] + i_offset + yy1 * i_fref + xx1;\
        int distortion = g_funcs.pixf.fpel_cmp[i_pixel](buf_pixel_temp, MAX_CU_SIZE, p_src1, i_fref) >> 1;\
        \
        cost = distortion + MV_COST_FPEL(mx, my) + mv_bid_bit;\
    } else {\
        cost = MAX_DISTORTION;\
    }


/**
 * ===========================================================================
 * local function defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * candMV1/4ȵֵµMVӦ2ʱµMVֵΧڣ򷵻1ʾµMVӦ
 * candMV1/4ֵΧڣµMVõʱMVֵΧ򷵻1ʾµMVӦ
 * 򣬷0ֵʾµMVӦü
 */
static int pmvr_adapt_mv(int *mx, int *my, int ctr_x, int ctr_y,
    int mv_x, int mv_y, int step_x, int step_y)
{
    if (XRD_ABS(mv_x - ctr_x) > TH_PMVR || XRD_ABS(mv_y - ctr_y) > TH_PMVR) {
        *mx = mv_x + step_x * 2;
        *my = mv_y + step_y * 2;
        return (XRD_ABS(*mx - ctr_x) <= TH_PMVR && XRD_ABS(*my - ctr_y) <= TH_PMVR);
    } else {
        *mx = mv_x + step_x;
        *my = mv_y + step_y;
        return (XRD_ABS(*mx - ctr_x) > TH_PMVR || XRD_ABS(*my - ctr_y) > TH_PMVR);
    }
}

/* ---------------------------------------------------------------------------
 */
static int ALWAYS_INLINE
mv_roundclip(int16_t (*dst)[2], int16_t (*mvc)[2], int i_mvc, int mv_min[2], int mv_max[2], uint32_t pmv)
{
    int cnt = 0;
    int i;

    for (i = 0; i < i_mvc; i++) {
        int mx = IPEL(mvc[i][0]);
        int my = IPEL(mvc[i][1]);
        uint32_t mv = MAKEDWORD(mx, my);

        if (!mv || mv == pmv) {
            continue;
        }

        dst[cnt][0] = (int16_t)XRD_CLIP3(mv_min[0], mv_max[0], mx);
        dst[cnt][1] = (int16_t)XRD_CLIP3(mv_min[1], mv_max[1], my);
        cnt++;
    }

    return cnt;
}

/* ---------------------------------------------------------------------------
 */
static int ALWAYS_INLINE
mv_clip(int16_t (*dst)[2], int16_t (*mvc)[2], int i_mvc, int mv_min[2], int mv_max[2], uint32_t pmv)
{
    int cnt = 0;
    int i;

    for (i = 0; i < i_mvc; i++) {
        int mx = mvc[i][0];
        int my = mvc[i][1];
        uint32_t mv = M32(mvc[i]);

        if (!mv || mv == pmv) {
            continue;
        }

        dst[cnt][0] = (int16_t)XRD_CLIP3(mv_min[0], mv_max[0], mx);
        dst[cnt][1] = (int16_t)XRD_CLIP3(mv_min[1], mv_max[1], my);
        cnt++;
    }

    return cnt;
}


/* ---------------------------------------------------------------------------
 * sub pixel block motion search
 */
static dist_t
me_subpel_refine(xrd_t *h, xrd_me_t *p_me, dist_t imv_cost, int *cost_neighbors)
{
    pel_t  *p_org     = p_me->p_fenc;
    pel_t **p_filtered = p_me->p_fref_1st->filtered;
    int i_fref   = p_me->p_fref_1st->i_stride[IMG_Y];
    int pmx      = p_me->mvp.x;
    int pmy      = p_me->mvp.y;
    int i_pixel  = p_me->i_pixel;
    int i_offset = p_me->i_bias;
    const uint16_t *p_cost_mvx = h->mvbits - p_me->mvp.x;
    const uint16_t *p_cost_mvy = h->mvbits - p_me->mvp.y;
    int lambda = h->i_lambda_factor;
    const int search_pos2 = 9;
    const int search_pos4 = 9;
    dist_t bcost;
    int ctr_x = (pmx >> 1) << 1;
    int ctr_y = (pmy >> 1) << 1;
    int pos, cost;
    int mx, my, bmx, bmy;
    mv_t bmv;

    // convert search center to quarter-pel units
    bmx = p_me->bmv.x;
    bmy = p_me->bmv.y;
    bmv = p_me->bmv;

    if (h->param.enable_hadamard) {
        ME_COST_QPEL(bmx, bmy);
        bcost = cost;
    } else {
        bcost = p_me->bcost;
    }

    /* -------------------------------------------------------------
     * half-pel refine */
    int skip_quar_pixel = 0;
    // loop over search positions
    for (pos = 1; pos < search_pos2; pos++) {
        mx = bmx + (Spiral[pos][0] << 1);
        my = bmy + (Spiral[pos][1] << 1);
        ME_COST_QPEL(mx, my);
        if (cost < bcost) {
            bcost = cost;
            bmv.v = MAKEDWORD(mx, my);
        }
    }

    bmx = bmv.x;
    bmy = bmv.y;

    /* -------------------------------------------------------------
     * quarter-pel refine */

    if (!skip_quar_pixel) {
        // loop over search positions
        for (pos = 1; pos < search_pos4; pos++) {
            if (h->param.enable_pmvr) {
                if (pmvr_adapt_mv(&mx, &my, ctr_x, ctr_y, bmx, bmy, Spiral[pos][0], Spiral[pos][1])) {
                    continue;
                }
            } else {
                mx = bmx + Spiral[pos][0];    // quarter-pel units
                my = bmy + Spiral[pos][1];    // quarter-pel units
            }

            // set motion vector cost
            ME_COST_QPEL(mx, my);
            if (cost < bcost) {
                bcost = cost;
                bmv.v = MAKEDWORD(mx, my);
            }
        }
    }
    // save the results
    p_me->bmv   = bmv;
    p_me->bcost = bcost;
    p_me->mvcost[PDIR_FWD] = MV_COST_FPEL(bmv.x,bmv.y);
    return bcost;
}


/**
 * ===========================================================================
 * interface function defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * initialize the motion search
 */
int xrd_me_get_buf_size(xrd_param_t *param)
{
    int me_range    = XRD_MAX(256, param->search_range);
    int subpel_num  = 4 * (2 * me_range + 3);
    int max_mv_bits = 5 + 2 * (int)ceil(log(subpel_num + 1) / log(2) + 1e-10);
    int max_mvd     = (1 << ((max_mv_bits >> 1))) - 1;
    int mem_size;

    /* buffer size for mvbits */
    mem_size = (max_mvd * 2 + 1) * sizeof(uint16_t) + CACHE_LINE_SIZE;

    return mem_size;
}

/**
 * ---------------------------------------------------------------------------
 * Function   : initialize the motion search module
 * Parameters :
 *      [in ] : h   - pointer to struct xrd_t, the HiPE encoder
 *      [out] : none
 * Return     : none
 * ---------------------------------------------------------------------------
 */
void xrd_me_init(xrd_t *h, uint8_t **mem_base)
{
    uint8_t *mbase  = *mem_base;
    int me_range    = XRD_MAX(256, h->param.search_range);
    int subpel_num  = 4 * (2 * me_range + 3);
    int max_mv_bits = 5 + 2 * (int)ceil(log(subpel_num + 1) / log(2) + 1e-10);
    int max_mvd     = (1 << ((max_mv_bits >> 1))) - 1;
    int bits, i, imin, imax;

    /* set pointer of mvbits */
    h->mvbits  = (uint16_t *)mbase;
    h->mvbits += max_mvd;       // reset the array offset
    mbase     += (max_mvd * 2 + 1) * sizeof(uint16_t);
    ALIGN_POINTER(mbase);

    *mem_base = mbase;

    // init array of motion vector bits
    h->mvbits[0] = 1;
    for (bits = 3; bits <= max_mv_bits; bits += 2) {
        imax = 1 << (bits >> 1);
        imin = imax >> 1;

        for (i = imin; i < imax; i++) {
            h->mvbits[-i] = h->mvbits[i] = (uint16_t)bits;
        }
    }
}

/* ---------------------------------------------------------------------------
 * i_qp: QP of P/F frame
 * TODO: call this function before encoding P/F frame?
 */
void xrd_me_init_umh_threshold(xrd_t *h, double *bsize, int i_qp)
{
    const int quant_coef[6] = { 13107, 11916, 10082, 9362, 8192, 7282 };
    int gb_qp_per = i_qp / 6;
    int gb_qp_rem = i_qp % 6;
    int gb_q_bits = 15 + gb_qp_per;
    int gb_qp_const = (h->i_type == SLICE_TYPE_I) ? ((1 << gb_q_bits) / 3) : ((1 << gb_q_bits) / 6);
    int threshold_4x4 = ((1 << gb_q_bits) - gb_qp_const) / quant_coef[gb_qp_rem];
    double quantize_step = threshold_4x4 / (4 * 5.61f);

    memset(bsize, 0, MAX_INTER_MODES * sizeof(double));
    bsize[PRED_nRx2N] = (16 * 16) * quantize_step;
    bsize[PRED_nLx2N] = 4 * bsize[PRED_nRx2N];
    bsize[PRED_2NxnD] = 4 * bsize[PRED_nRx2N];
    bsize[PRED_2NxnU] = 4 * bsize[PRED_2NxnD];
    bsize[PRED_Nx2N ] = 4 * bsize[PRED_2NxnU];
    bsize[PRED_2NxN ] = 4 * bsize[PRED_2NxnU];
    bsize[PRED_2Nx2N] = 4 * bsize[PRED_2NxN ];
}

// int g_me_time[4] = { 0 };

/* ---------------------------------------------------------------------------
 * return minimum motion cost after search
 */
dist_t xrd_me_search(xrd_t *h, xrd_me_t *p_me, int16_t(*mvc)[2], int i_mvc)
{
    /* special version of pack to allow shortcuts in CHECK_MV_RANGE */
    ALIGN16(int costs[16]);
    double beta2  = p_me->beta2 + 1;
    double beta3  = p_me->beta3 + 1;
    pel_t *p_org = p_me->p_fenc;
    pel_t *p_fref = p_me->p_fref_1st->planes[IMG_Y] + p_me->i_bias;
    int i_fref    = p_me->p_fref_1st->i_stride[IMG_Y];
    int i_pixel   = p_me->i_pixel;
    int mv_x_min  = p_me->mv_min_fpel[0];
    int mv_y_min  = p_me->mv_min_fpel[1];
    int mv_x_max  = p_me->mv_max_fpel[0];
    int mv_y_max  = p_me->mv_max_fpel[1];
    int me_range  = h->param.search_range;
    int lambda    = h->i_lambda_factor; // factor for determining Lagrangian's motion cost
    const uint32_t mv_min = pack16to32_mask2(-mv_x_min, -mv_y_min);
    const uint32_t mv_max = pack16to32_mask2(mv_x_max, mv_y_max) | 0x8000;
    const uint16_t *p_cost_mvx = h->mvbits - p_me->mvp.x;
    const uint16_t *p_cost_mvy = h->mvbits - p_me->mvp.y;
    uint32_t pmv;
    dist_t bcost = MAX_DISTORTION;
    int bmx = 0, bmy = 0;
    int omx, omy;
    int i, j, dir, idx;

    // g_me_time[0]++;
    /* -------------------------------------------------------------
     * try MVP and some key searching points */
    pmv = MAKEDWORD(mvc[0][0], mvc[0][1]);   /* mvc[0][] is the MVP */

    {
        int center_x = mvc[0][0];
        int center_y = mvc[0][1];
        int i_mvc_start = 1;
        ME_COST_IPEL(center_x, center_y);

        DIA_ITER(bmx, bmy);

        if (center_x != 0 || center_y != 0) {
            ME_COST_IPEL(0, 0);
            i_mvc_start++;

            if (bmx != center_x || bmy != center_y) {
                DIA_ITER(bmx, bmy);
            }
        }

        center_x = bmx;
        center_y = bmy;

        for (i = i_mvc_start; i < i_mvc; i++) {
            int mx = mvc[i][0];
            int my = mvc[i][1];
            ME_COST_IPEL(mx, my);
        }

        /* ǰMV MVPΧһС */
        if (MAKEDWORD(center_x, center_y) != MAKEDWORD(bmx, bmy)) {
            DIA_ITER(bmx, bmy);
        }
    }

    if (bcost == MAX_DISTORTION) {
        goto _me_error;         /* me failed */
    }

    /* -------------------------------------------------------------
     * search using different method */
    switch (h->me_method_frame) {
    case XRD_ME_TZ:         /* TZ */
    {
        const int RasterDistance = 16;
        const int EarlyExitIters = 3;
        int bdist;
        mv_info mvinfo;
        int tz_iters = 0;
        int i_dist;

        mvinfo.bcost = bcost;
        mvinfo.bdist = 0;
        mvinfo.bdir  = 0;
        
        /* direction */
        /*     2     */
        /*   4 * 5   */
        /*     7     */
        omx = bmx;
        omy = bmy;
        ME_COST_IPEL_X4_TZ(omx + 0, omy - 1, 2, 1,
                           omx - 1, omy + 0, 4, 1,
                           omx + 1, omy + 0, 5, 1,
                           omx + 0, omy + 1, 7, 1);
        if (mvinfo.bcost < bcost) {
            tz_iters = 0;
        } else {
            ++tz_iters;
        }
        for (i_dist = 2; i_dist <= me_range; i_dist <<= 1) {
            /*          2           points 2, 4, 5, 7 are i_dist
             *        1   3         points 1, 3, 6, 8 are i_dist/2
             *      4   *   5
             *        6   8
             *          7           */
            int i_dist2 = i_dist >> 1;
            omx = bmx;
            omy = bmy;
            ME_COST_IPEL_X4_TZ(omx,           omy - i_dist,  2, i_dist,
                               omx - i_dist2, omy - i_dist2, 1, i_dist2,
                               omx + i_dist2, omy - i_dist2, 3, i_dist2,
                               omx - i_dist,  omy,           4, i_dist);
            
            ME_COST_IPEL_X4_TZ(omx + i_dist,  omy,           5, i_dist,
                               omx - i_dist2, omy + i_dist2, 6, i_dist2,
                               omx + i_dist2, omy + i_dist2, 8, i_dist2,
                               omx,           omy + i_dist,  7, i_dist);
            if (mvinfo.bcost < bcost) {
                tz_iters = 0;
            } else if (++tz_iters >= EarlyExitIters) {
                break;
            }
        }
        bcost = mvinfo.bcost;
        bdist = mvinfo.bdist;
        dir   = mvinfo.bdir;

        if (bdist == 1) {
            if (!dir) {
                break;
            } else {
                goto umh_step_3;
            }
        }

        /* raster search refinement if original search distance was too big */
        if (bdist > RasterDistance) {
            const int iRasterDist  = RasterDistance >> 1;
            const int iRasterDist2 = RasterDistance >> 2;
            int rmv_y_min = XRD_MAX(mv_y_min, bmy - RasterDistance + 2);
            int rmv_y_max = XRD_MIN(mv_y_min, bmy + RasterDistance - 2);
            int rmv_x_min = XRD_MAX(mv_y_min, bmx - RasterDistance + 2);
            int rmv_x_max = XRD_MIN(mv_y_min, bmx + RasterDistance - 2);
            for (j = rmv_y_min; j < rmv_y_max; j += iRasterDist) {
                for (i = rmv_x_min; i < rmv_x_max; i += iRasterDist) {
                    ME_COST_IPEL_X4(i, j, i, j + iRasterDist2, i + iRasterDist2, j, i + iRasterDist2, j + iRasterDist2);
                }
            }
            bdist = iRasterDist;
        } else {
            bdist = mvinfo.bdist;
        }

        mvinfo.bcost = bcost;
        mvinfo.bdist = 0;
        mvinfo.bdir = 0;

        /* raster refinement */
        for (i_dist = bdist >> 1; i_dist > 1; i_dist >>= 1) {
            int i_dist2 = i_dist >> 1;
            omx = bmx;
            omy = bmy;
            ME_COST_IPEL_X4_TZ(omx,           omy - i_dist,  2, i_dist,
                               omx - i_dist2, omy - i_dist2, 1, i_dist2,
                               omx + i_dist2, omy - i_dist2, 3, i_dist2,
                               omx - i_dist,  omy,           4, i_dist);
            
            ME_COST_IPEL_X4_TZ(omx + i_dist,  omy,           5, i_dist,
                               omx - i_dist2, omy + i_dist2, 6, i_dist2,
                               omx + i_dist2, omy + i_dist2, 8, i_dist2,
                               omx,           omy + i_dist,  7, i_dist);
        }
        omx = bmx;
        omy = bmy;
        ME_COST_IPEL_X4_TZ(omx + 0, omy - 1, 2, 1,
                           omx - 1, omy + 0, 4, 1,
                           omx + 1, omy + 0, 5, 1,
                           omx + 0, omy + 1, 7, 1);



        /* star refinement */
        if (mvinfo.bdist > 0) {
            omx = bmx;
            omy = bmy;
            ME_COST_IPEL_X4_TZ(omx + 0, omy - 1, 2, 1,
                               omx - 1, omy + 0, 4, 1,
                               omx + 1, omy + 0, 5, 1,
                               omx + 0, omy + 1, 7, 1);
            for (i_dist = 2; i_dist <= 8; i_dist <<= 1) {
                int i_dist2 = i_dist >> 1;
                omx = bmx;
                omy = bmy;
                ME_COST_IPEL_X4_TZ(omx,           omy - i_dist,  2, i_dist,
                                   omx - i_dist2, omy - i_dist2, 1, i_dist2,
                                   omx + i_dist2, omy - i_dist2, 3, i_dist2,
                                   omx - i_dist,  omy,           4, i_dist);
            
                ME_COST_IPEL_X4_TZ(omx + i_dist,  omy,           5, i_dist,
                                   omx - i_dist2, omy + i_dist2, 6, i_dist2,
                                   omx + i_dist2, omy + i_dist2, 8, i_dist2,
                                   omx,           omy + i_dist,  7, i_dist);
            }
        }

        bcost = mvinfo.bcost;
        bdist = mvinfo.bdist;
        dir = mvinfo.bdir;
        if (bdist == 1 && !dir) {
            break;
        }
        goto umh_step_3;
    }
    case XRD_ME_UMH:        /* UMH */
        /* http://www.cnblogs.com/TaigaCon/archive/2014/06/16/3788984.html
         * 0. ʼ */
        pmv = MAKEDWORD(bmx, bmy);
        // g_me_time[1]++;

        // early termination algorithm
        EARLY_TERMINATION(p_me->pred_sad_uplayer);
        EARLY_TERMINATION(p_me->pred_sad);

        // umh_step_1:
        /* UMH 1. Unsymmetrical-cross search ǶԳʮ */
        // g_me_time[2]++;
        omx = bmx;
        omy = bmy;
        for (i = 1; i <= me_range; i += 2) {
            ME_COST_IPEL(omx + i, omy);
            ME_COST_IPEL(omx - i, omy);
        }
        for (j = 1; j <= me_range / 2; j += 2) {
            ME_COST_IPEL(omx, omy + j);
            ME_COST_IPEL(omx, omy - j);
        }

        // early termination algorithm
        EARLY_TERMINATION(p_me->pred_sad);

        /* UMH 2. Spiral search  */
        omx = bmx;
        omy = bmy;
        for (i = 0; i < 24; i++) {
            ME_COST_IPEL(omx + GRID[i][0], omy + GRID[i][1]);
        }

        // early termination algorithm
        EARLY_TERMINATION(p_me->pred_sad);

        // big hexagon
        for (j = 1; j <= me_range / 4; j++) {
            omx = bmx;
            omy = bmy;
            for (i = 0; i < 16; i++) {
                ME_COST_IPEL(omx + HEX4[i][0] * j, omy + HEX4[i][1] * j);
            }
            if (bmx != omx || bmy != omy) {
                EARLY_TERMINATION(p_me->pred_sad);
            }
        }

        /* !!! NO break statement here */
    case XRD_ME_HEX:        /* hexagon search */
umh_step_2 :                  /* UMH 3. Uneven Multi-Hexagon-grid Search ģ */
        // g_me_time[3]++;
        dir = 0;                                        /*   6   5   */
        omx = bmx;                                      /*           */
        omy = bmy;                                      /* 1   *   4 */
        ME_COST_IPEL_X3_DIR(-1,-2,6,  1,-2,5, -2,0,1);  /*           */
        ME_COST_IPEL_X3_DIR( 2, 0,4, -1, 2,2,  1,2,3);  /*   2   3   */

        if (dir) {
            const int8_t (*hex)[2];
            /* UMH 4. Extended Hexagon-based Search ģ巴 */
            idx = dir - 1;      /* start array index */
            /* half hexagon, not overlapping the previous iteration */
            for (i = 0; i < me_range && CHECK_MV_RANGE(bmx, bmy); i++) {
                dir = 0;
                omx = bmx;
                omy = bmy;
                hex = &HEX2[idx];
                ME_COST_IPEL_X3_DIR(hex[0][0],hex[0][1],1, hex[1][0],hex[1][1],2, hex[2][0],hex[2][1],3);
                if (!dir) {
                    break;      /* early terminate */
                }
                idx = M1MOD6[dir + idx - 1];    /* next start array index */
            }
        }
        /* !!! NO break statement here */
    case XRD_ME_DIA:        /* diamond search */
umh_step_3:                   /* UMH 5. the third step with a small search pattern Сģ巴 */
        dir = 0;
        if (CHECK_MV_RANGE(bmx, bmy)) {
            omx = bmx;                                          /*    4    */
            omy = bmy;                                          /*  1 * 3  */
            ME_COST_IPEL_X4_DIR(0,-1,4, -1,0,1, 1,0,3, 0,1,2);  /*    2    */
        }
        if (dir) {
            const int8_t (*dia)[2];
            idx = dir - 1;      /* start array index */
            /* half diamond, not overlapping the previous iteration */
            for (i = 0; i < me_range && CHECK_MV_RANGE(bmx, bmy); i++) {
                dir = 0;
                omx = bmx;
                omy = bmy;
                dia = &DIA1[idx];
                ME_COST_IPEL_X3_DIR(dia[0][0],dia[0][1],1, dia[1][0],dia[1][1],2, dia[2][0],dia[2][1],3);
                if (!dir) {
                    break;      /* early terminate */
                }
                idx = M1MOD4[dir + idx - 1];    /* next start array index */
            }
        }
        break;
    default:                    /* XRD_ME_FS: full search */
        omx = bmx;
        omy = bmy;
        for (j = -me_range; j < me_range; j++) {
            for (i = -me_range; i < me_range; i++) {
                ME_COST_IPEL(omx + i, omy + j);
            }
        }
        break;
    }

    /* -------------------------------------------------------------
     * store the results of fullpel search */
    p_me->bmv.v  = MAKEDWORD(FPEL(bmx), FPEL(bmy));
    p_me->bmv2.v = MAKEDWORD(bmx, bmy);
    p_me->bcost  = bcost;
    p_me->bcost2 = bcost;
    p_me->mvcost[PDIR_FWD] = MV_COST_IPEL(bmx, bmy);

    /* -------------------------------------------------------------
     * sub-pel refine */
    bcost = me_subpel_refine(h, p_me, bcost, costs);


_me_error:
    return bcost;
}

/* ---------------------------------------------------------------------------
 * find motion vector for forward dual hypothesis prediction (sub-pel search)
 * return minimum motion cost after search
 */
dist_t xrd_me_search_sym(xrd_t *h, xrd_me_t *p_me, pel_t *buf_pixel_temp, mv_t *mv)
{
    const int search_pos2 = 5;  // search positions for    half-pel search  (default: 9)
    const int search_pos4 = 5;  // search positions for quarter-pel search  (default: 9)
    pel_t **p_filtered1 = p_me->p_fref_1st->filtered;
    pel_t **p_filtered2 = p_me->p_fref_2nd->filtered;
    pel_t *p_org = p_me->p_fenc;
    int distance_fwd = p_me->i_distance_1st;
    int distance_bwd = p_me->i_distance_2nd;
    int i_pixel  = p_me->i_pixel;
    int i_offset = p_me->i_bias;
    int ctr_x    = (p_me->mvp1.x >> 1) << 1;
    int ctr_y    = (p_me->mvp1.y >> 1) << 1;
    int mv_x_min = p_me->mv_min[0];
    int mv_y_min = p_me->mv_min[1];
    int mv_x_max = p_me->mv_max[0];
    int mv_y_max = p_me->mv_max[1];
    int lambda   = h->i_lambda_factor;
    int min_pos2 = (h->param.enable_hadamard ? 0 : 1);
    int max_pos2 = (h->param.enable_hadamard ? XRD_MAX(1, search_pos2) : search_pos2);
    const uint32_t mv_min = pack16to32_mask2(-mv_x_min, -mv_y_min);
    const uint32_t mv_max = pack16to32_mask2(mv_x_max, mv_y_max) | 0x8000;
    const uint16_t *p_cost_mvx = h->mvbits - p_me->mvp.x;
    const uint16_t *p_cost_mvy = h->mvbits - p_me->mvp.y;
    mv_t bmv = *mv;  // best mv
    dist_t bcost = MAX_DISTORTION;
    dist_t cost;
    int pos;
    int mx, my;
    int i_fref = p_me->p_fref_1st->i_stride[IMG_Y];

    // loop over search positions
    for (pos = min_pos2; pos < max_pos2; pos++) {
        mx = mv->x + (Spiral[pos][0] << 1);    // quarter-pel units
        my = mv->y + (Spiral[pos][1] << 1);    // quarter-pel units

        ME_COST_QPEL_SYM;
        if (cost < bcost) {
            bcost = cost;
            bmv.v = MAKEDWORD(mx, my);
        }
    }

    mv->v = bmv.v;

    /* -------------------------------------------------------------
     * quarter-pel refine */

    // loop over search positions
    for (pos = 1; pos < search_pos4; pos++) {
        if (h->param.enable_pmvr) {
            if (pmvr_adapt_mv(&mx, &my, ctr_x, ctr_y, mv->x, mv->y, Spiral[pos][0], Spiral[pos][1])) {
                continue;
            }
        } else {
            mx = mv->x + Spiral[pos][0];    // quarter-pel units
            my = mv->y + Spiral[pos][1];    // quarter-pel units
        }

        ME_COST_QPEL_SYM;
        if (cost < bcost) {
            bcost = cost;
            bmv.v = MAKEDWORD(mx, my);
        }
    }

    mv->v = bmv.v;
    p_me->mvcost[PDIR_SYM] = MV_COST_FPEL(bmv.x, bmv.y);

    // return minimum motion cost
    return bcost;
}

/* ---------------------------------------------------------------------------
 * return minimum motion cost after search (sub-pel search)
 */
dist_t xrd_me_search_bid(xrd_t *h, xrd_me_t *p_me, pel_t *buf_pixel_temp, mv_t *fwd_mv, mv_t *bwd_mv, cu_parallel_t *p_enc)
{
    pel_t **p_filtered1 = p_me->p_fref_1st->filtered;
    pel_t **p_filtered2 = p_me->p_fref_2nd->filtered;
    pel_t *p_org = p_me->p_fenc;
    const int search_pos2 = 9;  // search positions for    half-pel search  (default: 9)
    const int search_pos4 = 9;  // search positions for quarter-pel search  (default: 9)
    int i_pixel  = p_me->i_pixel;
    int i_offset = p_me->i_bias;
    int ctr_x    = (p_me->mvp1.x >> 1) << 1;
    int ctr_y    = (p_me->mvp1.y >> 1) << 1;
    int mv_x_min = p_me->mv_min[0];
    int mv_y_min = p_me->mv_min[1];
    int mv_x_max = p_me->mv_max[0];
    int mv_y_max = p_me->mv_max[1];
    int lambda   = h->i_lambda_factor;
    int min_pos2 = (h->param.enable_hadamard ? 0 : 1);
    int max_pos2 = (h->param.enable_hadamard ? XRD_MAX(1, search_pos2) : search_pos2);
    int block_w = p_me->i_block_w;
    int xx2;
    int yy2;
    int mv_bid_bit;
    const uint32_t mv_min = pack16to32_mask2(-mv_x_min, -mv_y_min);
    const uint32_t mv_max = pack16to32_mask2(mv_x_max, mv_y_max) | 0x8000;
    const uint16_t *p_cost_mvx = h->mvbits - p_me->mvp1.x;
    const uint16_t *p_cost_mvy = h->mvbits - p_me->mvp1.y;
    const uint16_t *p_cost_bix = h->mvbits - p_me->mvp2.x;
    const uint16_t *p_cost_biy = h->mvbits - p_me->mvp2.y;
    mv_t bmv = *fwd_mv; // best mv
    dist_t bcost = MAX_DISTORTION;
    dist_t cost;
    int mx, my, mx_bid, my_bid;
    int pos;
    int i_fref = p_me->p_fref_1st->i_stride[IMG_Y];
    coeff_t *cur_blk = p_enc->coeff_blk;

    mx_bid = bwd_mv->x;
    my_bid = bwd_mv->y;

    //ѱֵԤֵļ㹫ʽΪ2ֵ-Ԥֵ
    xx2 = mx_bid >> 2;
    yy2 = my_bid >> 2;
    mv_bid_bit = MV_COST_FPEL_BID(mx_bid, my_bid);

    if (CHECK_MV_RANGE(mx_bid, my_bid)) {
        pel_t *p_src2 = p_filtered2[((my_bid & 3) << 2) + (mx_bid & 3)];

        if (p_src2 != NULL) {
            p_src2 += i_offset + yy2 * i_fref + xx2;
            g_funcs.pixf.sub_ps[i_pixel](cur_blk, block_w, p_org, p_src2, FENC_STRIDE, i_fref);//M-A
        } else {
            ALIGN32(pel_t tmp_pred[MAX_CU_SIZE * MAX_CU_SIZE]);
            mv_t mvt;
            mvt.x = (int16_t)mx_bid;
            mvt.y = (int16_t)my_bid;
            get_mv_for_mc(h, &mvt, p_me->i_pix_x, p_me->i_pix_y, block_w, p_me->i_block_h);
            mc_luma(tmp_pred, MAX_CU_SIZE, mvt.x, mvt.y, block_w, p_me->i_block_h, p_me->p_fref_2nd);
            g_funcs.pixf.sub_ps[i_pixel](cur_blk, block_w, p_org, tmp_pred, FENC_STRIDE, MAX_CU_SIZE);//M-A
        }
        g_funcs.pixf.add_ps[i_pixel](buf_pixel_temp, MAX_CU_SIZE, p_org, cur_blk, FENC_STRIDE, block_w);//M-A+M
    }

    // loop over search positions
    for (pos = min_pos2; pos < max_pos2; pos++) {
        mx = fwd_mv->x + (Spiral[pos][0] << 1);    // quarter-pel units
        my = fwd_mv->y + (Spiral[pos][1] << 1);    // quarter-pel units

        ME_COST_QPEL_BID;
        if (cost < bcost) {
            bcost = cost;
            bmv.v = MAKEDWORD(mx, my);
        }
    }

    fwd_mv->v = bmv.v;

    /* -------------------------------------------------------------
     * quarter-pel refine */

    // loop over search positions
    for (pos = 1; pos < search_pos4; pos++) {
        if (h->param.enable_pmvr) {
            if (pmvr_adapt_mv(&mx, &my, ctr_x, ctr_y, fwd_mv->x, fwd_mv->y, Spiral[pos][0], Spiral[pos][1])) {
                continue;
            }
        } else {
            mx = fwd_mv->x + Spiral[pos][0];    // quarter-pel units
            my = fwd_mv->y + Spiral[pos][1];    // quarter-pel units
        }

        ME_COST_QPEL_BID;
        if (cost < bcost) {
            bcost = cost;
            bmv.v = MAKEDWORD(mx, my);
        }
    }

    fwd_mv->v = bmv.v;
    p_me->mvcost[PDIR_BID] = MV_COST_FPEL(bmv.x, bmv.y) + MV_COST_FPEL_BID(mx_bid, my_bid);

    // return minimum motion cost
    return bcost;
}
