/*
 * rdo.c
 *
 * Description of this file:
 *    RDO functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"
#include "rdo.h"
#include "cudata.h"
#include "aec.h"
#include "common/mc.h"
#include "transform.h"
#include "block_info.h"
#include "wquant.h"
#include "me.h"
#include "cpu.h"
#include "predict.h"
#include "ratecontrol.h"
#include "rdoq.h"


/**
 * ===========================================================================
 * local/global variables
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 */
static const float SUBCU_COST_RATE[2][4] = {
    {0.50f, 0.75f, 0.97f, 1.0f},   /* ֡CUCostһ㶼ϴ */
    {0.75f, 0.90f, 0.99f, 1.0f},   /* ֡£SkipCostС */
};

static const int tab_pdir_bskip[DS_MAX_NUM] = { PDIR_BID, PDIR_BWD, PDIR_SYM, PDIR_FWD };

extern double tab_thres_cu_non_split_intra[CTU_DEPTH][3];


/* ---------------------------------------------------------------------------
 */
static const int8_t NUM_PREDICTION_UNIT[MAX_PRED_MODES] = {// [mode]
    1, // 0: 8x8, ---, ---, --- (PRED_SKIP   )
    1, // 1: 8x8, ---, ---, --- (PRED_2Nx2N  )
    2, // 2: 8x4, 8x4, ---, --- (PRED_2NxN   )
    2, // 3: 4x8, 4x8, ---, --- (PRED_Nx2N   )
    2, // 4: 8x2, 8x6, ---, --- (PRED_2NxnU  )
    2, // 5: 8x6, 8x2, ---, --- (PRED_2NxnD  )
    2, // 6: 2x8, 6x8, ---, --- (PRED_nLx2N  )
    2, // 7: 6x8, 2x8, ---, --- (PRED_nRx2N  )
    1, // 8: 8x8, ---, ---, --- (PRED_I_2Nx2N)
    4, // 9: 4x4, 4x4, 4x4, 4x4 (PRED_I_NxN  )
    4, //10: 8x2, 8x2, 8x2, 8x2 (PRED_I_2Nxn )
    4  //11: 2x8, 2x8, 2x8, 2x8 (PRED_I_nx2N )
};

static const cb_t CODING_BLOCK_INFO[MAX_PRED_MODES + 1][4] = {// [mode][block]
    //   x, y, w, h        x, y, w, h        x, y, w, h        x, y, w, h          for block 0, 1, 2 and 3
    { { {0, 0, 8, 8} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 0: 8x8, ---, ---, --- (PRED_SKIP   )
    { { {0, 0, 8, 8} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 1: 8x8, ---, ---, --- (PRED_2Nx2N  )
    { { {0, 0, 8, 4} }, { {0, 4, 8, 4} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 2: 8x4, 8x4, ---, --- (PRED_2NxN   )
    { { {0, 0, 4, 8} }, { {4, 0, 4, 8} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 3: 4x8, 4x8, ---, --- (PRED_Nx2N   )
    { { {0, 0, 8, 2} }, { {0, 2, 8, 6} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 4: 8x2, 8x6, ---, --- (PRED_2NxnU  )
    { { {0, 0, 8, 6} }, { {0, 6, 8, 2} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 5: 8x6, 8x2, ---, --- (PRED_2NxnD  )
    { { {0, 0, 2, 8} }, { {2, 0, 6, 8} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 6: 2x8, 6x8, ---, --- (PRED_nLx2N  )
    { { {0, 0, 6, 8} }, { {6, 0, 2, 8} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 7: 6x8, 2x8, ---, --- (PRED_nRx2N  )
    { { {0, 0, 8, 8} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} }, { {0, 0, 0, 0} } }, // 8: 8x8, ---, ---, --- (PRED_I_2Nx2N)
    { { {0, 0, 4, 4} }, { {4, 0, 4, 4} }, { {0, 4, 4, 4} }, { {4, 4, 4, 4} } }, // 9: 4x4, 4x4, 4x4, 4x4 (PRED_I_NxN  )
    { { {0, 0, 8, 2} }, { {0, 2, 8, 2} }, { {0, 4, 8, 2} }, { {0, 6, 8, 2} } }, //10: 8x2, 8x2, 8x2, 8x2 (PRED_I_2Nxn )
    { { {0, 0, 2, 8} }, { {2, 0, 2, 8} }, { {4, 0, 2, 8} }, { {6, 0, 2, 8} } }, //11: 2x8, 2x8, 2x8, 2x8 (PRED_I_nx2N )
    { { {0, 0, 4, 4} }, { {4, 0, 4, 4} }, { {0, 4, 4, 4} }, { {4, 4, 4, 4} } }, // X: 4x4, 4x4, 4x4, 4x4
};

static const int8_t TU_SPLIT_TYPE[MAX_PRED_MODES][2] = {  // [mode][(NsqtEnable or SdipEnables) and cu_level > B8X8_IN_BIT]
    //  split_type for block non-sdip/nsqt:[0] and sdip/nsqt:[1]
    { TU_SPLIT_CROSS,   TU_SPLIT_CROSS   },   // 0: 8x8, ---, ---, --- (PRED_SKIP   )
    { TU_SPLIT_CROSS,   TU_SPLIT_CROSS   },   // 1: 8x8, ---, ---, --- (PRED_2Nx2N  )
    { TU_SPLIT_CROSS,   TU_SPLIT_HOR     },   // 2: 8x4, 8x4, ---, --- (PRED_2NxN   )
    { TU_SPLIT_CROSS,   TU_SPLIT_VER     },   // 3: 4x8, 4x8, ---, --- (PRED_Nx2N   )
    { TU_SPLIT_CROSS,   TU_SPLIT_HOR     },   // 4: 8x2, 8x6, ---, --- (PRED_2NxnU  )
    { TU_SPLIT_CROSS,   TU_SPLIT_HOR     },   // 5: 8x6, 8x2, ---, --- (PRED_2NxnD  )
    { TU_SPLIT_CROSS,   TU_SPLIT_VER     },   // 6: 2x8, 6x8, ---, --- (PRED_nLx2N  )
    { TU_SPLIT_CROSS,   TU_SPLIT_VER     },   // 7: 6x8, 2x8, ---, --- (PRED_nRx2N  )
    { TU_SPLIT_NON,     TU_SPLIT_INVALID },   // 8: 8x8, ---, ---, --- (PRED_I_2Nx2N)
    { TU_SPLIT_CROSS,   TU_SPLIT_CROSS   },   // 9: 4x4, 4x4, 4x4, 4x4 (PRED_I_NxN  )
    { TU_SPLIT_INVALID, TU_SPLIT_HOR     },   //10: 8x2, 8x2, 8x2, 8x2 (PRED_I_2Nxn )
    { TU_SPLIT_INVALID, TU_SPLIT_VER     }    //11: 2x8, 2x8, 2x8, 2x8 (PRED_I_nx2N )
};

static const int8_t headerbits_skipmode[8] = { 2, 3, 4, 4, 3, 4, 5, 5 };//temporal wsm1 wsm2 wsm3 spatial_direct0 spatial_direct1 spatial_direct2 spatial_direct3
/**
 * ===========================================================================
 * local function defines (utilities)
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * CUģʽȷǰCUPUʹС֡仮֣
 */
static ALWAYS_INLINE
void cu_init_pu_inter(xrd_t *h, cu_info_t *p_cu_info, int i_level, int i_mode)
{
    const int shift_bits = i_level - MIN_CU_SIZE_IN_BIT;
    const int8_t block_num = NUM_PREDICTION_UNIT[i_mode];
    int ds_mode = p_cu_info->directskip_mhp_idx;
    int i;
    cb_t *p_cb = p_cu_info->cb;

    // set for each block
    if (i_mode == PRED_SKIP) {
        ///! һЩSkip/DirectģʽCU8x8PUֳ4
        if (i_level > 3 && (h->i_type == SLICE_TYPE_P || (h->i_type == SLICE_TYPE_F && ds_mode == DS_NONE)
            || (h->i_type == SLICE_TYPE_B && ds_mode == DS_NONE))) {
            p_cu_info->num_pu = 4;
            for (i = 0; i < 4; i++) {
                p_cb[i].v = CODING_BLOCK_INFO[PRED_I_nx2N + 1][i].v << shift_bits;
            }
        } else {
            p_cu_info->num_pu = 1;
            memset(p_cu_info->cb, 0, sizeof(p_cu_info->cb));
            p_cb[0].v = CODING_BLOCK_INFO[PRED_SKIP][0].v << shift_bits;
        }
    } else {
        p_cu_info->num_pu = block_num;
        for (i = 0; i < block_num; i++) {
            p_cb[i].v = CODING_BLOCK_INFO[i_mode][i].v << shift_bits;
        }
    }
}

/* ---------------------------------------------------------------------------
 * CUģʽȷǰCUPUʹС֡ڻ֣
 */
static ALWAYS_INLINE
void cu_init_pu_intra(xrd_t *h, cu_info_t *p_cu_info, int i_level, int i_mode)
{
    const int shift_bits = i_level - MIN_CU_SIZE_IN_BIT;
    const int8_t block_num = NUM_PREDICTION_UNIT[i_mode];
    int i;
    cb_t *p_cb = p_cu_info->cb;

    UNUSED_PARAMETER(h);
    // set for each block
    p_cu_info->num_pu = block_num;
    for (i = 0; i < 4; i++) {
        p_cb[i].v = CODING_BLOCK_INFO[i_mode][i].v << shift_bits;
    }
}

/* ---------------------------------------------------------------------------
 * TU split type when TU split is enabled for current CU
 */
static ALWAYS_INLINE
void cu_set_tu_split_type(xrd_t *h, cu_info_t *p_cu_info, int transform_split_flag) 
{
    int mode = p_cu_info->i_mode;
    int level = p_cu_info->i_level;
    int enable_nsqt_sdip = IS_INTRA_MODE(mode) ? h->param.enable_sdip : h->param.enable_nsqt;

    enable_nsqt_sdip = enable_nsqt_sdip && level > B8X8_IN_BIT;
    p_cu_info->i_tu_split = transform_split_flag ? TU_SPLIT_TYPE[mode][enable_nsqt_sdip] : TU_SPLIT_NON;
    assert(p_cu_info->i_tu_split != TU_SPLIT_INVALID);
}


/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE uint32_t
cu_get_valid_modes(xrd_t *h, int frm_type, int level)
{
    return h->valid_modes[frm_type][level - MIN_CU_SIZE_IN_BIT];
}

/* ---------------------------------------------------------------------------
 */
static INLINE
void cu_init(xrd_t *h, cu_t *p_cu, cu_info_t *best, int i_level)
{
    cu_layer_t *p_layer  = cu_get_layer(h, i_level);
    int i;

    /* Ping-pong buffer */
    p_layer->buf_pred_inter      = p_layer->buf_pred_inter_luma[0];
    p_layer->buf_pred_inter_best = p_layer->buf_pred_inter_luma[1];

    /* init rec and coeff pointer */
    p_cu->cu_info.p_rec  [0]      = p_layer->rec_buf_y [0];
    p_cu->cu_info.p_coeff[0]      = p_layer->coef_buf_y[0];
    p_layer->p_rec_tmp   [0]      = p_layer->rec_buf_y [1];
    p_layer->p_coeff_tmp [0]      = p_layer->coef_buf_y[1];
    best->p_rec          [0]      = p_layer->rec_buf_y [2];
    best->p_coeff        [0]      = p_layer->coef_buf_y[2];

    p_cu->cu_info.p_rec  [1]      = p_layer->rec_buf_uv [0][0];
    p_cu->cu_info.p_coeff[1]      = p_layer->coef_buf_uv[0][0];
    p_layer->p_rec_tmp   [1]      = p_layer->rec_buf_uv [0][1];
    p_layer->p_coeff_tmp [1]      = p_layer->coef_buf_uv[0][1];
    best->p_rec          [1]      = p_layer->rec_buf_uv [0][2];
    best->p_coeff        [1]      = p_layer->coef_buf_uv[0][2];

    p_cu->cu_info.p_rec  [2]      = p_layer->rec_buf_uv [1][0];
    p_cu->cu_info.p_coeff[2]      = p_layer->coef_buf_uv[1][0];
    p_layer->p_rec_tmp   [2]      = p_layer->rec_buf_uv [1][1];
    p_layer->p_coeff_tmp [2]      = p_layer->coef_buf_uv[1][1];
    best->p_rec          [2]      = p_layer->rec_buf_uv [1][2];
    best->p_coeff        [2]      = p_layer->coef_buf_uv[1][2];

    /* init basic properties */
    p_cu->cu_info.i_cbp = 0;

    /* set qp needed in loop filter (even if constant QP is used) */
    p_cu->cu_info.i_cu_qp = h->i_qp;

    if (h->param.i_rc_method == XRD_RC_CBR_SCU) {
        int i_left_cu_qp;
        if (p_cu->i_pix_x > 0) {
            i_left_cu_qp = h->cu_info[p_cu->i_scu_xy - 1].i_cu_qp;
        } else {
            i_left_cu_qp = h->i_qp;
        }

        p_cu->cu_info.i_delta_qp = p_cu->cu_info.i_cu_qp - i_left_cu_qp;
    } else {
        p_cu->cu_info.i_delta_qp = 0;
    }

    /* ref_idx_1st[], ref_idx_2nd[] ڴ */
    memset(p_cu->cu_info.ref_idx_1st, INVALID_REF, sizeof(p_cu->cu_info.ref_idx_1st) + sizeof(p_cu->cu_info.ref_idx_2nd));

    /* init position for 4 sub-CUs */
    if (i_level > B8X8_IN_BIT) {
        for (i = 0; i < 4; i++) {
            cu_t *p_sub_cu = p_cu->sub_cu[i];

            p_sub_cu->i_pix_x          = p_cu->i_pix_x + ((i &  1) << (i_level - 1));
            p_sub_cu->i_pix_y          = p_cu->i_pix_y + ((i >> 1) << (i_level - 1));
            p_sub_cu->cu_info.i_scu_x  = p_sub_cu->i_pix_x >> MIN_CU_SIZE_IN_BIT;
            p_sub_cu->cu_info.i_scu_y  = p_sub_cu->i_pix_y >> MIN_CU_SIZE_IN_BIT;
            p_sub_cu->i_scu_xy         = p_sub_cu->cu_info.i_scu_y * h->i_width_in_mincu + p_sub_cu->cu_info.i_scu_x;
        }
    }

    /* set neighbor CUs */
    check_neighbor_cu_avail(h, p_cu, p_cu->cu_info.i_scu_x, p_cu->cu_info.i_scu_y, p_cu->i_scu_xy);

    p_cu->b_cbp_direct = 0;
}

/* ---------------------------------------------------------------------------
 * copy information of CU
 */
static ALWAYS_INLINE void cu_copy_info(cu_info_t *p_dst, const cu_info_t *p_src)
{
    const int num_bytes = sizeof(cu_info_t) - (int)((uint8_t *)&p_dst->i_level - (uint8_t *)p_dst);
    memcpy(&p_dst->i_level, &p_src->i_level, num_bytes);
}

/* ---------------------------------------------------------------------------
 * store cu parameters to best
 */
static void cu_store_parameters(xrd_t *h, cu_t *p_cu, cu_info_t *best)
{
    int mode = p_cu->cu_info.i_mode;
    cu_mode_t *p_mode = cu_get_layer_mode(h, p_cu->cu_info.i_level);

    // store best mode
    cu_copy_info(best, &p_cu->cu_info);

    /* --- reconstructed blocks ---- */
    XRD_SWAP_PTR(best->p_rec[0], p_cu->cu_info.p_rec[0]);
    XRD_SWAP_PTR(best->p_rec[1], p_cu->cu_info.p_rec[1]);
    XRD_SWAP_PTR(best->p_rec[2], p_cu->cu_info.p_rec[2]);

    /* ---- residual (coefficients) ---- */
    XRD_SWAP_PTR(best->p_coeff[0], p_cu->cu_info.p_coeff[0]);
    XRD_SWAP_PTR(best->p_coeff[1], p_cu->cu_info.p_coeff[1]);
    XRD_SWAP_PTR(best->p_coeff[2], p_cu->cu_info.p_coeff[2]);

    /* ---- prediction information ---- */
    if (!IS_INTRA_MODE(mode)) {
        memcpy(&p_mode->best_mc, &p_cu->mc, sizeof(p_cu->mc));
    }
}

/* ---------------------------------------------------------------------------
 * sets motion vectors and reference indexes for an CU
 */
static void cu_save_mvs_refs(xrd_t *h, cu_info_t *p_cu_info)
{
    int8_t *p_dirpred = h->dir_pred;
    int8_t *p_ref_1st = h->fwd_1st_ref;
    int8_t *p_ref_2nd = h->bwd_2nd_ref;
    mv_t   *p_mv_1st  = h->fwd_1st_mv;
    mv_t   *p_mv_2nd  = h->bwd_2nd_mv;
    int bx_4x4_cu     = p_cu_info->i_scu_x << (MIN_CU_SIZE_IN_BIT - MIN_PU_SIZE_IN_BIT);
    int by_4x4_cu     = p_cu_info->i_scu_y << (MIN_CU_SIZE_IN_BIT - MIN_PU_SIZE_IN_BIT);
    int w_in_4x4      = h->i_width_in_minpu;
    cu_mode_t *p_cu_mode = cu_get_layer_mode(h, p_cu_info->i_level);
    cu_mc_param_t *p_mc = &p_cu_mode->best_mc;
    int width, height;
    int k, by, bx;
    int r, c;

    for (k = 0; k < p_cu_info->num_pu; k++) {
        int8_t i_dir_pred = (int8_t)p_cu_info->b8pdir[k];
        int8_t ref_1st = (int8_t)p_cu_info->ref_idx_1st[k];
        int8_t ref_2nd = (int8_t)p_cu_info->ref_idx_2nd[k];
        mv_t   mv_1st  = p_mc->mv[k][0];
        mv_t   mv_2nd  = p_mc->mv[k][1];
        cb_t   cur_cb  = p_cu_info->cb[k];

        cur_cb.v >>= 2;
        bx         = cur_cb.x;
        by         = cur_cb.y;
        width      = cur_cb.w;
        height     = cur_cb.h;

        bx += bx_4x4_cu;
        by += by_4x4_cu;

        for (r = 0; r < height; r++) {
            int offset = (by + r) * w_in_4x4 + bx;
            for (c = 0; c < width; c++) {
                p_dirpred[offset + c] = i_dir_pred;

                p_mv_1st [offset + c] = mv_1st;
                p_mv_2nd [offset + c] = mv_2nd;

                p_ref_1st[offset + c] = ref_1st;
                p_ref_2nd[offset + c] = ref_2nd;
            }
        }
    }
}

/* ---------------------------------------------------------------------------
 * set stored cu parameters
 */
static INLINE
void cu_copy_stored_parameters(xrd_t *h, cu_t *p_cu, cu_info_t *best)
{
    int mode       = best->i_mode;
    int w_in_4x4   = h->i_width_in_minpu;
    int scu_xy     = p_cu->i_scu_xy;
    int b4x4_x     = p_cu->i_pix_x >> MIN_PU_SIZE_IN_BIT;
    int b4x4_y     = p_cu->i_pix_y >> MIN_PU_SIZE_IN_BIT;
    int pix_x      = p_cu->i_pos_x;
    int pix_y      = p_cu->i_pos_y;
    int pix_cx     = pix_x >> 1;
    int pix_cy     = pix_y >> 1;
    int blocksize  = p_cu->i_size;
    int ip_stride  = h->i_width_in_minpu + 16;
    int part_idx_c = PART_INDEX(blocksize >> 1, blocksize >> 1);
    int8_t *p_intramode   = h->ipredmode + (pix_y >> MIN_PU_SIZE_IN_BIT) * ip_stride + b4x4_x;
    const int size_in_spu = (blocksize >> MIN_PU_SIZE_IN_BIT);
    const int size_in_scu = (blocksize >> MIN_CU_SIZE_IN_BIT);
    int i, j;

    cu_copy_info(&p_cu->cu_info, best);

    //===== reconstruction values =====
    g_funcs.pixf.copy_pp[PART_INDEX(blocksize, blocksize)](h->lcu.p_fdec[0] + pix_y * FDEC_STRIDE + pix_x, FDEC_STRIDE,
                                                           best->p_rec[0], FREC_STRIDE);
    g_funcs.pixf.copy_ss[PART_INDEX(blocksize, blocksize)](h->lcu.lcu_coeff[0] + (p_cu->idx_zorder << 6), blocksize,
                                                           best->p_coeff[0], blocksize);

    g_funcs.pixf.copy_pp[part_idx_c](h->lcu.p_fdec[1] + pix_cy * FDEC_STRIDE + pix_cx, FDEC_STRIDE, 
                                     best->p_rec[1], FREC_CSTRIDE / 2);
    g_funcs.pixf.copy_pp[part_idx_c](h->lcu.p_fdec[2] + pix_cy * FDEC_STRIDE + pix_cx, FDEC_STRIDE, 
                                     best->p_rec[2], FREC_CSTRIDE / 2);
    g_funcs.pixf.copy_ss[part_idx_c](h->lcu.lcu_coeff[1] + (p_cu->idx_zorder << 4), blocksize >> 1, 
                                     best->p_coeff[1], blocksize >> 1);
    g_funcs.pixf.copy_ss[part_idx_c](h->lcu.lcu_coeff[2] + (p_cu->idx_zorder << 4), blocksize >> 1,
                                     best->p_coeff[2], blocksize >> 1);

    //===============   cbp and mode   ===============
    for (j = 0; j < size_in_scu; j++) {
        cu_info_t *p_cu_info = &h->cu_info[j * h->i_width_in_mincu + scu_xy];  // save data to cu_info
        for (i = size_in_scu; i != 0; i--) {
            cu_copy_info(p_cu_info++, best);
        }
    }

    //===============   intra pred mode   ===============
    if (IS_INTRA_MODE(mode)) {
        int n_size4 = size_in_spu >> 2;
        int k;
        int8_t intra_pred_mode;

        switch (mode) {
        case PRED_I_2Nxn:
            for (i = 0; i < 4; i++) {
                for (j = i * n_size4; j < (i + 1) * n_size4; j++) {
                    g_funcs.fast_memset(p_intramode, p_cu->cu_info.real_intra_modes[i], size_in_spu * sizeof(int8_t));
                    p_intramode += ip_stride;
                }
            }
            break;
        case PRED_I_nx2N:
            for (j = 0; j < size_in_spu; j++) {
                for (i = 0; i < 4; i++) {
                    k = i * n_size4;
                    g_funcs.fast_memset(p_intramode + k, p_cu->cu_info.real_intra_modes[i], n_size4 * sizeof(int8_t));
                }
                p_intramode += ip_stride;
            }
            break;
        case PRED_I_NxN:
            n_size4 = size_in_spu >> 1;
            for (j = 0; j < n_size4; j++) {
                for (i = 0; i < 2; i++) {
                    k = i * n_size4;
                    g_funcs.fast_memset(p_intramode + k, p_cu->cu_info.real_intra_modes[i], n_size4 * sizeof(int8_t));
                }
                p_intramode += ip_stride;
            }
            for (j = n_size4; j < size_in_spu; j++) {
                for (i = 0; i < 2; i++) {
                    k = i * n_size4;
                    g_funcs.fast_memset(p_intramode + k, p_cu->cu_info.real_intra_modes[i + 2], n_size4 * sizeof(int8_t));
                }
                p_intramode += ip_stride;
            }
            break;
        default: // PRED_2Nx2N
            intra_pred_mode = p_cu->cu_info.real_intra_modes[0];
            for (j = size_in_spu - 1; j != 0; j--) {
                p_intramode[size_in_spu - 1] = intra_pred_mode;
                p_intramode += ip_stride;
            }
            g_funcs.fast_memset(p_intramode, intra_pred_mode, size_in_spu * sizeof(int8_t));
            break;
        }
    } else if (h->i_type != SLICE_TYPE_I && h->fenc->b_enable_intra) {
        for (j = size_in_spu - 1; j != 0; j--) {
            p_intramode[size_in_spu - 1] = -1;
            p_intramode += ip_stride;
        }
        g_funcs.fast_memset(p_intramode, -1, size_in_spu * sizeof(int8_t));
    }

    //============== inter prediction information =========================
    if (h->i_type != SLICE_TYPE_I) {
        if (IS_INTER_MODE(p_cu->cu_info.i_mode)) {
            cu_save_mvs_refs(h, &p_cu->cu_info); // store mv
        } else {
            int8_t *p_dirpred = h->dir_pred    + b4x4_y * w_in_4x4 + b4x4_x;
            int8_t *p_ref_1st = h->fwd_1st_ref + b4x4_y * w_in_4x4 + b4x4_x;
            int8_t *p_ref_2nd = h->bwd_2nd_ref + b4x4_y * w_in_4x4 + b4x4_x;
            int size_b4 = blocksize >> MIN_PU_SIZE_IN_BIT;

            for (i = size_b4; i != 0; i--) {
                for (j = 0; j < size_b4; j++) {
                    p_ref_1st[j] = INVALID_REF;
                    p_ref_2nd[j] = INVALID_REF;
                    p_dirpred[j] = PDIR_INVALID;
                }
                p_dirpred += w_in_4x4;
                p_ref_1st += w_in_4x4;
                p_ref_2nd += w_in_4x4;
            }
        }
    }
}

/* ---------------------------------------------------------------------------
 * get spatial neighboring MV
 */
static ALWAYS_INLINE
void cu_get_neighbor_spatial(xrd_t *h, int cur_slice_idx, neighbor_inter_t *p_neighbor, int x4, int y4)
{
    int pos           = y4 * h->i_width_in_minpu + x4;
    int y_outside_pic = y4 < 0 || y4 >= h->i_height_in_minpu;
    int x_outside_pic = x4 < 0 || x4 >= h->i_width_in_minpu;

    // scu_xy = XRD_MIN(h->i_width_in_mincu * h->i_height_in_mincu - 1, XRD_MAX(0, scu_xy));
    if (y_outside_pic || x_outside_pic || cu_get_slice_index(h, x4 >> 1, y4 >> 1) != cur_slice_idx) {
        p_neighbor->is_available = 0;
        p_neighbor->i_dir_pred = PDIR_INVALID;
        p_neighbor->ref_idx[0] = INVALID_REF;
        p_neighbor->ref_idx[1] = INVALID_REF;
        p_neighbor->mv[0].v    = 0;
        p_neighbor->mv[1].v    = 0;
    } else {
        p_neighbor->is_available = 1;
        p_neighbor->i_dir_pred = h->dir_pred[pos];
        p_neighbor->ref_idx[0] = h->fwd_1st_ref[pos];
        p_neighbor->ref_idx[1] = h->bwd_2nd_ref[pos];
        p_neighbor->mv[0]      = h->fwd_1st_mv[pos];
        p_neighbor->mv[1]      = h->bwd_2nd_mv[pos];
    }
}

/* ---------------------------------------------------------------------------
 * get temporal MV predictor
 */
static ALWAYS_INLINE
void cu_get_neighbor_temporal(xrd_t *h, neighbor_inter_t *p_neighbor, int x4, int y4)
{
    int w_in_16x16 = (h->i_width_in_minpu + 3) >> 2;
    int pos = (y4 >> 2) * w_in_16x16 + (x4 >> 2);

    p_neighbor->is_available = 1;
    p_neighbor->i_dir_pred   = PDIR_FWD;
    p_neighbor->ref_idx[0]   = h->fref[0]->pu_ref[pos];
    p_neighbor->mv[0]        = h->fref[0]->pu_mv[pos];
    p_neighbor->ref_idx[1]   = INVALID_REF;
    p_neighbor->mv[1].v      = 0;
}

/* ---------------------------------------------------------------------------
 * get neighboring MVs for MVP
 */
void cu_get_neighbors(xrd_t *h, cu_t *p_cu, cb_t *p_cb)
{
    neighbor_inter_t *neighbors = cu_get_layer(h, p_cu->cu_info.i_level)->neighbor_inter;
    int cur_slice_idx = cu_get_slice_index(h, p_cu->i_pix_x >> MIN_CU_SIZE_IN_BIT, p_cu->i_pix_y >> MIN_CU_SIZE_IN_BIT);
    int bx_4x4 = p_cu->i_pix_x >> MIN_PU_SIZE_IN_BIT;
    int by_4x4 = p_cu->i_pix_y >> MIN_PU_SIZE_IN_BIT;
    int xx0 = (p_cb->x >> MIN_PU_SIZE_IN_BIT) + bx_4x4;
    int yy0 = (p_cb->y >> MIN_PU_SIZE_IN_BIT) + by_4x4;
    int xx1 = (p_cb->w >> MIN_PU_SIZE_IN_BIT) + xx0 - 1;
    int yy1 = (p_cb->h >> MIN_PU_SIZE_IN_BIT) + yy0 - 1;

    /* 1. check whether the top-right 4x4 block is reconstructed */
    int x_TR_4x4_in_lcu = xx1 - (h->lcu.i_scu_x << (MIN_CU_SIZE_IN_BIT - MIN_PU_SIZE_IN_BIT));
    int y_TR_4x4_in_lcu = yy0 - (h->lcu.i_scu_y << (MIN_CU_SIZE_IN_BIT - MIN_PU_SIZE_IN_BIT));
    int b_available_TR  = h->tab_avail_TR[(y_TR_4x4_in_lcu << (h->i_lcu_level - B4X4_IN_BIT)) + x_TR_4x4_in_lcu];

    /* 2. get neighboring blocks */
    /*  */
    cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_TOPLEFT ], xx0 - 1, yy0 - 1);

    /* ڵPUϢ */
    if (IS_VER_PU_PART(p_cu->cu_info.i_mode) && p_cb->x != 0) {  // CUֱΪPUҵǰPUΪұһ
        neighbor_inter_t *p_neighbor = neighbors + BLK_LEFT;
        p_neighbor->is_available = 1;
        // cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_LEFT], xx0 - 1, yy0);
        p_neighbor->i_dir_pred = p_cu->cu_info.b8pdir[0];
        p_neighbor->ref_idx[0] = p_cu->cu_info.ref_idx_1st[0];
        p_neighbor->ref_idx[1] = p_cu->cu_info.ref_idx_2nd[0];
        p_neighbor->mv[0]      = p_cu->mc.mv[0][0];
        p_neighbor->mv[1]      = p_cu->mc.mv[0][1];
        memcpy(&neighbors[BLK_LEFT2], p_neighbor, sizeof(neighbor_inter_t));
    } else {
        cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_LEFT], xx0 - 1, yy0);
        cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_LEFT2], xx0 - 1, yy1);
    }

    /* ڵPUϢ */
    if (IS_HOR_PU_PART(p_cu->cu_info.i_mode) && p_cb->y != 0) {  // CUˮƽΪPUҵǰPUΪ±һ
        neighbor_inter_t *p_neighbor = neighbors + BLK_TOP;
        p_neighbor->is_available = 1;
        // cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_LEFT], xx0 - 1, yy0);
        p_neighbor->i_dir_pred = p_cu->cu_info.b8pdir[0];
        p_neighbor->ref_idx[0] = p_cu->cu_info.ref_idx_1st[0];
        p_neighbor->ref_idx[1] = p_cu->cu_info.ref_idx_2nd[0];
        p_neighbor->mv[0]      = p_cu->mc.mv[0][0];
        p_neighbor->mv[1]      = p_cu->mc.mv[0][1];
        memcpy(&neighbors[BLK_TOP2], p_neighbor, sizeof(neighbor_inter_t));
    } else {
        cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_TOP], xx0, yy0 - 1);
        cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_TOP2], xx1, yy0 - 1);
    }

    /*  */
    cu_get_neighbor_spatial(h, cur_slice_idx, &neighbors[BLK_TOPRIGHT], b_available_TR ? xx1 + 1 : -1, yy0 - 1);

    cu_get_neighbor_temporal(h, &neighbors[BLK_COL], xx0, yy0);
}


/* ---------------------------------------------------------------------------
 * return: number of reference frames
 */
static ALWAYS_INLINE
int cu_get_mvs_for_mc(xrd_t *h, cu_t *p_cu, int pu_idx,
                      mv_t *p_mv_1st, mv_t *p_mv_2nd, 
                      int *p_ref_idx1, int *p_ref_idx2)
{
    int num_ref;            // number of reference frames
    int dmh_mode = p_cu->cu_info.dmh_mode;
    int ref_1st = p_cu->cu_info.ref_idx_1st[pu_idx]; // һǰB֡Ԥ⣩˶ʸ
    int ref_2nd = p_cu->cu_info.ref_idx_2nd[pu_idx]; // ڶB֡˫ĺ
    mv_t mv_1st, mv_2nd;    // һǰB֡Ԥ⣩͵ڶ˶ʸ

    if (h->i_type != SLICE_TYPE_B) {
        num_ref = (ref_1st != INVALID_REF) + (ref_2nd != INVALID_REF);
        mv_1st  = p_cu->mc.mv[pu_idx][0];
        mv_2nd  = p_cu->mc.mv[pu_idx][1];
        if (dmh_mode > 0) {
            num_ref = 2;
            ref_2nd = ref_1st;
            mv_2nd  = mv_1st;
            mv_1st.x -= tab_dmh_pos[dmh_mode][0];
            mv_1st.y -= tab_dmh_pos[dmh_mode][1];
            mv_2nd.x += tab_dmh_pos[dmh_mode][0];
            mv_2nd.y += tab_dmh_pos[dmh_mode][1];
        }
    } else {
        num_ref = (ref_1st != INVALID_REF) + (ref_2nd != INVALID_REF);
        if (ref_1st == INVALID_REF) {
            ref_1st = B_BWD;
            ref_2nd = INVALID_REF;
            mv_1st  = p_cu->mc.mv[pu_idx][1];
            mv_2nd.v = 0;
        } else {
            mv_1st  = p_cu->mc.mv[pu_idx][0];
            mv_2nd  = p_cu->mc.mv[pu_idx][1];
            ref_1st = B_FWD;
            ref_2nd = B_BWD;
        }
    }

    *p_mv_1st   = mv_1st;
    *p_mv_2nd   = mv_2nd;
    *p_ref_idx1 = ref_1st;
    *p_ref_idx2 = ref_2nd;

    return num_ref;
}

/* ---------------------------------------------------------------------------
 * forward quantization
 */
static INLINE
int tu_quant_forward(xrd_t *h, aec_t *p_aec, cu_t *p_cu, coeff_t *p_coeff, int i_level, int bsx, int bsy, int qp, int b_intra, int b_luma, int intra_mode)
{
    /* ((1 << x) * 5) / 31 */
    static const int tab_quant_fwd_add[] = {
         0,      0,      0,      1,      2,      5,     10,     20,
        41,     82,    165,    330,    660,   1321,   2642,   5285,
     10570,  21140,  42281,  84562, 169125, 338250, 676500, 1353001,
    };
    const int shift = 15 + LIMIT_BIT - (h->param.sample_bit_depth + 1) - i_level;
    /* TODO: ȷadd޸ĶܵӰ죬Ӱϴĳɲ */
    const int add = tab_quant_fwd_add[shift + b_intra];
    // const int add = ((1 << (shift + b_intra)) * 165) >> 10;   // (* 165 / 1024) Լ  (* 5 / 31)

    if (h->lcu.b_enable_rdoq) {
        return rdoq_block(h, p_aec, p_cu, p_coeff, bsx, bsy, i_level, qp, b_luma, intra_mode);
    } else {
#if !ENABLE_WQUANT
        return g_funcs.dctf.quant(p_coeff, bsx * bsy, tab_Q_TAB[qp], shift, add);
#else
        if (!h->WeightQuantEnable) {
            return g_funcs.dctf.quant(p_coeff, bsx * bsy, tab_Q_TAB[qp], shift, add);
        } else {
            int *levelscale = h->wq_data.levelScale[i_level - B4X4_IN_BIT][b_intra];

            return g_funcs.dctf.wquant(p_coeff, bsx * bsy, tab_Q_TAB[qp], shift, add, levelscale);
        }
#endif
    }
}

/* ---------------------------------------------------------------------------
 * inverse quantization
 */
static INLINE
void tu_quant_inverse(xrd_t *h, cu_t *p_cu, coeff_t *coef, int num_coeff, int i_level, int qp, int b_luma)
{
    const int scale = tab_IQ_TAB[qp];
    const int shift = tab_IQ_SHIFT[qp] + (h->param.sample_bit_depth + 1) + i_level - LIMIT_BIT;

#if !ENABLE_WQUANT
    UNUSED_PARAMETER(h);
    UNUSED_PARAMETER(b_luma);
    UNUSED_PARAMETER(p_cu);
    g_funcs.dctf.dequant(coef, num_coeff, scale, shift);
#else
    if (!h->WeightQuantEnable) {
        g_funcs.dctf.dequant(coef, num_coeff, scale, shift);
    } else {
        int b_hor = b_luma && p_cu->cu_info.i_tu_split == TU_SPLIT_HOR;
        int b_ver = b_luma && p_cu->cu_info.i_tu_split == TU_SPLIT_VER;
        const int16_t(*AVS_SCAN)[2] = NULL;
        int wqm_size_id = 0;
        int wqm_stride = 0;
        int wqm_shift = h->param.PicWQDataIndex == 1 ? 3 : 0;
        int xy_shift = 0;
        int16_t *wq_matrix;

        // adaptive frequency weighting quantization
        if ((h->param.enable_sdip || h->param.enable_nsqt) && (b_hor || b_ver)) {
            xy_shift = XRD_MIN(2, i_level - B4X4_IN_BIT);
            wqm_size_id = xy_shift + 1;

            if (b_hor) {
                AVS_SCAN = tab_coef_scan_list_hor[XRD_MIN(2, i_level - 2)];
            } else {
                AVS_SCAN = tab_coef_scan_list_ver[XRD_MIN(2, i_level - 2)];
            }
        } else {
            xy_shift = XRD_MIN(3, i_level - B4X4_IN_BIT);
            wqm_size_id = xy_shift + 1;

            AVS_SCAN = tab_coef_scan_list[XRD_MIN(3, i_level - 2)];
        }

        wqm_stride = 1 << (wqm_size_id + B4X4_IN_BIT);
        if (wqm_size_id == 2) {
            wqm_stride >>= 1;
        } else if (wqm_size_id == 3) {
            wqm_stride >>= 2;
        }
        wq_matrix = h->wq_data.cur_wq_matrix[wqm_size_id];

        dequant_weighted_c(coef, num_coeff, scale, shift, wqm_shift, wqm_stride, xy_shift, wq_matrix, AVS_SCAN);
    }
#endif
}

/* ---------------------------------------------------------------------------
 */
static INLINE
void tu_get_dct_coeff(xrd_t *h, coeff_t *cur_blk, int pu_size_idx, int bsx, int bsy)
{

    g_funcs.dctf.dct[pu_size_idx](cur_blk, cur_blk, bsx);

}



/**
 * ===========================================================================
 * local function defines (chroma)
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * finish transform, quantization, inverse-transform, inverse-quantization 
 * and reconstruction pixel generation of chroma block
 */
static int cu_recon_chroma(xrd_t *h, aec_t *p_aec, cu_t *p_cu, dist_t *distortion)
{
    int b_intra = IS_INTRA_MODE(p_cu->cu_info.i_mode);
    int pix_x_c = p_cu->i_pos_x >> 1;
    int pix_y_c = p_cu->i_pos_y >> CHROMA_V_SHIFT;
    int level_c = p_cu->cu_info.i_level - CHROMA_V_SHIFT;
    int bsize_c = 1 << level_c;
    int partidx_c = PART_INDEX(bsize_c, bsize_c);
    int cbp_c = 0;     // Coding Block Pattern (CBP) of chroma blocks
    int num_nonzero;   // number of non-zero coefficients
    int qp_c;
    int uv;
    cu_parallel_t *p_enc = cu_get_enc_context(h, p_cu->cu_info.i_level);
    coeff_t *cur_blk = p_enc->coeff_blk;
    pel_t *p_pred;

    /* prediction buffer of chroma blocks */
    if (b_intra) {
        p_pred = p_enc->intra_pred_c[p_cu->cu_info.i_intra_mode_c];
    } else {
        p_pred = p_enc->buf_pred_inter_c;
    }

     for (uv = 0; uv < 2; uv++) {
        pel_t *p_fdec = p_cu->cu_info.p_rec[uv + 1];
        pel_t *p_fenc = h->lcu.p_fenc[uv + 1] + pix_y_c * FENC_STRIDE + pix_x_c;

        g_funcs.pixf.sub_ps[partidx_c](cur_blk, bsize_c, p_fenc, p_pred, FENC_STRIDE, FREC_CSTRIDE);

        // DCT, quantization, inverse quantization, IDCT, and reconstruction
        tu_get_dct_coeff(h, cur_blk, partidx_c, bsize_c, bsize_c);

        qp_c = cu_get_qp(h, &p_cu->cu_info);
#if ENABLE_WQUANT
        qp_c += (uv == 0 ? h->param.chroma_quant_param_delta_u : h->param.chroma_quant_param_delta_v);
#endif

        qp_c = cu_get_chroma_qp(h, qp_c, uv);

        num_nonzero = tu_quant_forward(h, p_aec, p_cu, cur_blk, level_c, bsize_c, bsize_c, qp_c, b_intra, 0, DC_PRED);
        cbp_c |= (num_nonzero != 0) << (4 + uv);

        if (num_nonzero) {
            // TODO: ݿ
            g_funcs.pixf.copy_ss[partidx_c](p_cu->cu_info.p_coeff[uv + 1], bsize_c, cur_blk, bsize_c);

            tu_quant_inverse(h, p_cu, cur_blk, bsize_c * bsize_c, level_c, qp_c, 0);
            g_funcs.dctf.idct[partidx_c](cur_blk, cur_blk, bsize_c);

            g_funcs.pixf.add_ps[partidx_c](p_fdec, FREC_CSTRIDE / 2, p_pred, cur_blk, FREC_CSTRIDE, bsize_c);
        } else {
            g_funcs.pixf.copy_pp[partidx_c](p_fdec, FREC_CSTRIDE / 2, p_pred, FREC_CSTRIDE);
        }

        distortion[uv] = g_funcs.pixf.ssd[partidx_c](p_fenc, FENC_STRIDE, p_fdec, FREC_CSTRIDE / 2);

        p_pred += (FREC_CSTRIDE >> 1);  // uvoffset
    }

    return cbp_c;
}

/* ---------------------------------------------------------------------------
 * get max available bits for left residual coding
 */
static ALWAYS_INLINE
int rdo_get_left_bits(xrd_t *h, rdcost_t min_rdcost, dist_t distortion)
{
    rdcost_t f_lambda = h->f_lambda_mode;
    double f_left_bits = ((min_rdcost - distortion) * h->f_lambda_1th) + 1;
    int left_bits;

    left_bits = (int)XRD_CLIP3F(0.0f, 32766.0f, f_left_bits);    // clipһ
    if (left_bits * f_lambda + distortion <= min_rdcost) {
        left_bits++;    // ⸡֤ﵽֵʱrdcostmin_rdcost
    }

    return left_bits;
}


/**
 * ===========================================================================
 * local function defines (intra)
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * finish transform, quantization, inverse-transform, inverse-quantization 
 * and reconstruction pixel generation of a intra luma block
 */
static INLINE
int cu_recon_intra_luma(xrd_t *h, aec_t *p_aec, cu_t *p_cu, pel_t *p_pred, int bsx, int bsy, 
                        int block_x, int block_y, int idx_tu, int intra_pred_mode, dist_t *distortion)
{
    int used_wavelet = (p_cu->cu_info.i_level == B64X64_IN_BIT && p_cu->cu_info.i_tu_split != TU_SPLIT_CROSS);
    int i_tu_level = p_cu->cu_info.i_level - (p_cu->cu_info.i_tu_split != TU_SPLIT_NON);
    int pos_x = p_cu->i_pos_x + block_x;
    int pos_y = p_cu->i_pos_y + block_y;
    int part_idx = PART_INDEX(bsx, bsy);
    int w_tr = bsx >> used_wavelet;
    int h_tr = bsy >> used_wavelet;
    int num_non_zero;
    int b_2nd_trans = h->param.enable_secT;
    cu_parallel_t *p_enc  = cu_get_enc_context(h, p_cu->cu_info.i_level);
    pel_t      *p_fenc    = h->lcu.p_fenc[0] + pos_y * FENC_STRIDE + pos_x;
    pel_t      *p_fdec    = p_cu->cu_info.p_rec[0] + block_y * FREC_STRIDE + block_x;
    coeff_t    *cur_blk   = p_enc->coeff_blk;
    coeff_t    *p_coeff_y = p_cu->cu_info.p_coeff[0] + (idx_tu << ((p_cu->cu_info.i_level - 1) << 1));
    int b_top  = IS_NEIGHBOR_AVAIL(p_cu->block_avail, MD_I_TOP);
    int b_left = IS_NEIGHBOR_AVAIL(p_cu->block_avail, MD_I_LEFT);

    // get prediction and prediction error
    g_funcs.pixf.sub_ps[PART_INDEX(bsx, bsy)](cur_blk, bsx, p_fenc, p_pred, FENC_STRIDE, bsx);

    // block transform
    if (part_idx == LUMA_4x4) {
        if (b_2nd_trans) {
            g_funcs.dctf.transform_4x4_2nd(cur_blk, w_tr);
        } else {
            g_funcs.dctf.dct[LUMA_4x4](cur_blk, cur_blk, 4);     /* 4x4 dct */
        }
    } else {
        tu_get_dct_coeff(h, cur_blk, part_idx, w_tr, h_tr);

        if (b_2nd_trans) {
            g_funcs.dctf.transform_2nd(cur_blk, w_tr, intra_pred_mode, b_top, b_left);
        }
    }

    // quantization
    num_non_zero = tu_quant_forward(h, p_aec, p_cu, cur_blk, i_tu_level, w_tr, h_tr, cu_get_qp(h, &p_cu->cu_info), 1, 1, intra_pred_mode);

    if (num_non_zero) {
        g_funcs.pixf.copy_ss[PART_INDEX(w_tr, h_tr)](p_coeff_y, w_tr, cur_blk, w_tr);

        // inverse quantization
        tu_quant_inverse(h, p_cu, cur_blk, w_tr * h_tr, i_tu_level, cu_get_qp(h, &p_cu->cu_info), 1);

        // inverse transform
        if (part_idx == LUMA_4x4) {
            if (b_2nd_trans) {
                g_funcs.dctf.inv_transform_4x4_2nd(cur_blk, w_tr);
            } else {
                g_funcs.dctf.idct[LUMA_4x4](cur_blk, cur_blk, 4);    /* 4x4 idct */
            }
        } else {
            if (b_2nd_trans) {
                g_funcs.dctf.inv_transform_2nd(cur_blk, w_tr, intra_pred_mode, b_top, b_left);
            }

            g_funcs.dctf.idct[part_idx](cur_blk, cur_blk, w_tr);
        }

        g_funcs.pixf.add_ps[part_idx](p_fdec, FREC_STRIDE, p_pred, cur_blk, bsx, bsx);
    } else {
        g_funcs.pixf.copy_pp[part_idx](p_fdec, FREC_STRIDE, p_pred, bsx);
    }

    // get distortion (SSD) of current block
    *distortion = g_funcs.pixf.ssd[part_idx](p_fenc, FENC_STRIDE, p_fdec, FREC_STRIDE);

    return num_non_zero;
}

/* ---------------------------------------------------------------------------
 * get the MPMs of an intra block at (pos_x, pos_y) as 4x4 address
 */
static ALWAYS_INLINE
void xrd_get_mpms(xrd_t *h, cu_t *p_cu, int blockidx, int pos_y, int pos_x, int mpm[2])
{
    int ip_stride = h->i_width_in_minpu + 16;
    int top_mode  = h->ipredmode[((pos_y >> MIN_PU_SIZE_IN_BIT) - 1) * ip_stride + pos_x];
    int left_mode = h->ipredmode[(pos_y >> MIN_PU_SIZE_IN_BIT) * ip_stride + pos_x - 1];

    if (blockidx != 0) {
        if (p_cu->cu_info.i_mode == PRED_I_2Nxn) {
            top_mode = p_cu->cu_info.real_intra_modes[blockidx - 1];
        } else if (p_cu->cu_info.i_mode == PRED_I_nx2N) {
            left_mode = p_cu->cu_info.real_intra_modes[blockidx - 1];
        } else if (p_cu->cu_info.i_mode == PRED_I_NxN) {
            switch (blockidx) {
            case 1:
                left_mode = p_cu->cu_info.real_intra_modes[0];
                break;
            case 2:
                top_mode  = p_cu->cu_info.real_intra_modes[0];
                break;
            case 3:
                top_mode  = p_cu->cu_info.real_intra_modes[1];
                left_mode = p_cu->cu_info.real_intra_modes[2];
                break;
            default: // case 0:
                break;
            }
        }
    }

    top_mode  = (top_mode  < 0) ? DC_PRED : top_mode;
    left_mode = (left_mode < 0) ? DC_PRED : left_mode;
    mpm[0] = XRD_MIN(top_mode, left_mode);
    mpm[1] = XRD_MAX(top_mode, left_mode);

    if (mpm[0] == mpm[1]) {
        mpm[0] = DC_PRED;
        mpm[1] = (mpm[1] == DC_PRED) ? BI_PRED : mpm[1];
    }
}


/* ---------------------------------------------------------------------------
 * ֡PUַʽRDCostŵPUַʽ
 */
static void cu_check_intra(xrd_t *h, aec_t *p_aec, cu_t *p_cu, cu_info_t *best, int mode, rdcost_t *min_rdcost)
{
    int level = p_cu->cu_info.i_level;
    cu_layer_t    *p_layer = cu_get_layer(h, level);
    cu_parallel_t *p_enc   = cu_get_enc_context(h, level);
    rdcost_t rdcost_luma = 0;
    rdcost_t rdcost = MAX_COST;
    rdcost_t min_mode_rdcost = MAX_COST;
    pel_t *rec_bak_y = best->p_rec[0];
    pel_t *p_best_part[4];
    int blockidx;
    int num_luma_block = mode != PRED_I_2Nx2N ? 4 : 1;
    int b_need_swap_buf = 0;
    int pix_x_c = p_cu->i_pos_x >> 1;
    int pix_y_c = p_cu->i_pos_y >> CHROMA_V_SHIFT;
    intra_candidate_t *p_candidates = p_layer->intra_candidates;

    /* ȷPU */
    cu_init_pu_intra(h, &p_cu->cu_info, level, mode);

    /* ȷTU */
    cu_set_tu_split_type(h, &p_cu->cu_info, mode != PRED_I_2Nx2N);

    h->copy_aec_state_rdo(&p_layer->cs_rdo, p_aec);
    p_cu->cu_info.i_cbp = 0;

    p_cu->intra_avail = (uint8_t)xrd_intra_get_cu_neighbors(h, p_cu, p_cu->i_pix_x, p_cu->i_pix_y, p_cu->i_size);

    /* 1, intra luma prediction and mode decision */
    for (blockidx = 0; blockidx < num_luma_block; blockidx++) {
        int mpm[2];  // most probable modes (MPMs) for current luma block
        int block_x = p_cu->cu_info.cb[blockidx].x;
        int block_y = p_cu->cu_info.cb[blockidx].y;
        int block_w = p_cu->cu_info.cb[blockidx].w;
        int block_h = p_cu->cu_info.cb[blockidx].h;
        int pos_x = p_cu->i_pos_x + block_x;
        int pos_y = p_cu->i_pos_y + block_y;
        int b4x4_x = (p_cu->i_pix_x + block_x) >> MIN_PU_SIZE_IN_BIT;
        dist_t best_dist = MAX_DISTORTION;
        int best_rate = INT_MAX;
        int best_mode = 0;
        int best_pmode = 0;
        int best_cbp = 0;
        pel_t *p_fenc = h->lcu.p_fenc[0] + pos_y * FENC_STRIDE + pos_x;
        rdcost_t best_rdcost = MAX_COST;
        int i;
        int num_for_rdo;
        intra_candidate_t *p_candidates = p_layer->intra_candidates;   // candidate list, reserving the cost
        /* init */
        xrd_get_mpms(h, p_cu, blockidx, pos_y, b4x4_x, mpm);

        for (i = 0; i < INTRA_MODE_NUM_FOR_RDO; i++) {
            p_candidates[i].mode = 0;
            p_candidates[i].cost = MAX_COST;
        }

        /* conduct prediction and get intra prediction direction candidates for RDO */
        num_for_rdo = h->lcu.get_intra_dir_for_rdo_luma(h, p_cu, p_candidates, p_fenc, mpm, blockidx, 
                                                        block_x, block_y, block_w, block_h);

        // store the coding state
        h->copy_aec_state_rdo(&p_layer->cs_rdo, p_aec);

        /* RDO */
        for (i = 0; i < num_for_rdo; i++) {
            rdcost_t rdcost;
            dist_t dist_curr;     // ǰ֡ڿʧ
            int rate_curr = 0; // ǰ֡ڿʣ
            int mode = p_candidates[i].mode;
            pel_t *p_pred = p_enc->intra_pred[mode];

            // get and check rate_chroma-distortion cost
            int mode_idx_aec = (mpm[0] == mode) ? -2 : ((mpm[1] == mode) ? -1 : (mpm[0] > mode ? mode : (mpm[1] > mode ? mode - 1 : mode - 2)));
            int num_nonzero;

            num_nonzero = cu_recon_intra_luma(h, p_aec, p_cu, p_pred,
                                              block_w, block_h, block_x, block_y,
                                              blockidx, mode, &dist_curr);
            num_nonzero = !!num_nonzero;

            {
                int used_wavelet = (p_cu->cu_info.i_level == B64X64_IN_BIT && p_cu->cu_info.i_tu_split != TU_SPLIT_CROSS);
                int w_tr = block_w >> used_wavelet;
                int i_tu_level = p_cu->cu_info.i_level - (p_cu->cu_info.i_tu_split != TU_SPLIT_NON) - used_wavelet;
                int rate_luma_mode; 
                coeff_t *p_coeff_y = p_cu->cu_info.p_coeff[0] + (blockidx << ((p_cu->cu_info.i_level - 1) << 1));

                // get rate for intra prediction mode
                rate_luma_mode = p_aec->binary.write_intra_pred_mode(p_aec, mode_idx_aec);

                // get rate for luminance coefficients
                // TODO: ٶ CBP 0/1Ӱ죬ʵӦӰ
                if (num_nonzero) {
                    int bits_left = rdo_get_left_bits(h, best_rdcost, dist_curr) - rate_luma_mode;
                    rate_curr = p_aec->binary.est_luma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info, p_coeff_y,
                                                                   i_tu_level, xrd_log2u(w_tr),
                                                                   1, mode, bits_left);
                    rate_luma_mode += rate_curr;
                }

                // calculate RD-cost and return it
                rdcost = dist_curr + h->f_lambda_mode * rate_luma_mode;
            }

            // choose best mode
            if (rdcost < best_rdcost) {
                XRD_SWAP_PTR(p_cu->cu_info.p_rec[0], p_layer->p_rec_tmp[0]);
                XRD_SWAP_PTR(p_cu->cu_info.p_coeff[0], p_layer->p_coeff_tmp[0]);

                // set best mode update minimum cost
                best_dist = dist_curr;
                best_rate = rate_curr;
                best_rdcost = rdcost;
                best_mode = mode;
                best_pmode = mode_idx_aec;
                best_cbp = num_nonzero;   // flag if dct-coefficients must be coded
                h->copy_aec_state_rdo(&p_enc->cs_tu, p_aec);
            }

            h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);

        }   // for (i = 0; i < num_for_rdo; i++)

        /* change the coding state to BEST */
        if (best_rate < INT_MAX) {
            if (p_cu->cu_info.i_mode != PRED_I_2Nx2N) {
                /* TODO: @luofl ﲻܿعֵ fdec */
                g_funcs.pixf.copy_pp[PART_INDEX(block_w, block_h)](h->lcu.p_fdec[0] + pos_y * FDEC_STRIDE + pos_x, FDEC_STRIDE,
                                                                   p_layer->p_rec_tmp[0] + block_y * FREC_STRIDE + block_x, FREC_STRIDE);
            }

            /* copy coefficients and reconstructed data for best mode */
            XRD_SWAP_PTR(p_cu->cu_info.p_rec[0], p_layer->p_rec_tmp[0]);
            XRD_SWAP_PTR(p_cu->cu_info.p_coeff[0], p_layer->p_coeff_tmp[0]);
            p_best_part[blockidx] = p_cu->cu_info.p_rec[0];

            /* set intra mode prediction */
            p_cu->cu_info.pred_intra_modes[blockidx] = (int8_t)best_pmode;
            p_cu->cu_info.real_intra_modes[blockidx] = (int8_t)best_mode;

            /* copy coding state */
            h->copy_aec_state_rdo(p_aec, &p_enc->cs_tu);
        }

        /* ģʽ״̬ʧ桢ȷųԤģʽCBP */
        rdcost_luma += best_dist + h->f_lambda_mode * best_rate;
        p_cu->cu_info.i_cbp |= (best_cbp) << blockidx;

        /* ȿRDOǰֹ */
        if (rdcost_luma >= *min_rdcost) {
            p_layer->mode_rdcost[mode] = MAX_COST;        /* set the cost for SDIP fast algorithm */
            h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);
            return;  // ȿrdcostѾǰֵֹͣɫȿģʽ
        }
    }
    p_cu->feature.rdcost_luma = rdcost_luma;

    /* 2, store best luma reconstruction pixels */
    for (blockidx = 0; blockidx < num_luma_block; blockidx++) {
        if (p_best_part[blockidx] != p_cu->cu_info.p_rec[0]) {
            int offset = p_cu->cu_info.cb[blockidx].y * FREC_STRIDE + p_cu->cu_info.cb[blockidx].x;
            int offset_coeff = blockidx << ((p_cu->cu_info.i_level - 1) << 1);
            int w_tr = p_cu->cu_info.cb[0].w;
            int h_tr = p_cu->cu_info.cb[0].h;
            int part_idx = PART_INDEX(w_tr, h_tr);
            g_funcs.pixf.copy_pp[part_idx](p_cu->cu_info.p_rec[0]   + offset, FREC_STRIDE, p_layer->p_rec_tmp[0]  + offset, p_cu->i_size);
            g_funcs.pixf.copy_ss[part_idx](p_cu->cu_info.p_coeff[0] + offset_coeff, w_tr, p_layer->p_coeff_tmp[0] + offset_coeff, w_tr);
        }
    }

    /* 3, Chroma mode decision and CU mode updating */
    int i_left_cmode = DM_PRED_C;
    /* check left */
    if (p_cu->p_left_cu != NULL) {
        i_left_cmode = p_cu->p_left_cu->i_intra_mode_c;
    }

    if (h->param.chroma_format != CHROMA_400) {
        int lmode;
        int num_rdo_chroma_mode;
        int idx_chroma_mode;
        int tmp_cbp_luma = p_cu->cu_info.i_cbp;

        lmode = tab_intra_mode_luma2chroma[p_cu->cu_info.real_intra_modes[0]];
        num_rdo_chroma_mode = h->get_intra_candidates_chroma(h, p_cu, level - 1, pix_y_c, pix_x_c, p_candidates);

        for (idx_chroma_mode = 0; idx_chroma_mode < num_rdo_chroma_mode; idx_chroma_mode++) {
            dist_t dist_chroma = 0;  // ɫȿָ
            dist_t dist_uv[2];
            int rate_chroma = 0;
            int bits_left;
            int predmode_c = p_candidates[idx_chroma_mode].mode;
            int cbp_c;

            if (predmode_c != DM_PRED_C && predmode_c == lmode) {
                continue;
            }
            p_cu->cu_info.i_intra_mode_c = (int8_t)predmode_c;

            /* RDO̵ɫȿع̣任任عֵ */
            cbp_c = cu_recon_chroma(h, p_aec, p_cu, dist_uv);
            dist_chroma = dist_uv[0] + dist_uv[1];

            p_cu->cu_info.i_cbp = (int8_t)(tmp_cbp_luma + cbp_c);

            /* ------- GET RATE -------- */
            rate_chroma = p_aec->binary.est_cu_header(p_aec, h, &p_cu->cu_info, i_left_cmode);
            rate_chroma += p_aec->binary.write_cu_cbp_dqp(p_aec, h, &p_cu->cu_info, h->i_slice_index, h->lcu.last_dequant);

            bits_left = rdo_get_left_bits(h, *min_rdcost - rdcost_luma, dist_chroma);

            if (p_cu->cu_info.i_cbp & (1 << 4)) {
                int cur_bits_left = bits_left - rate_chroma;
                rate_chroma += p_aec->binary.est_chroma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info,
                                                                    p_cu->cu_info.p_coeff[1],
                                                                    level - 1, cur_bits_left);
            }
            if (p_cu->cu_info.i_cbp & (1 << 5)) {
                int cur_bits_left = bits_left - rate_chroma;
                rate_chroma += p_aec->binary.est_chroma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info,
                                                                    p_cu->cu_info.p_coeff[2],
                                                                    level - 1, cur_bits_left);
            }

            rdcost = dist_chroma + h->f_lambda_mode * rate_chroma + rdcost_luma;

            min_mode_rdcost = XRD_MIN(rdcost, min_mode_rdcost);

            if (rdcost < *min_rdcost) {
                *min_rdcost = rdcost;
                h->copy_aec_state_rdo(&p_layer->cs_cu, p_aec);    /* store coding state for the best mode */
                cu_store_parameters(h, p_cu, best);
                b_need_swap_buf = 1;
            }

            h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);
        }
    } else {   /* YUV400 */
        /* ------- GET RATE -------- */
        int rate_hdr = p_aec->binary.est_cu_header(p_aec, h, &p_cu->cu_info, i_left_cmode);
        rate_hdr += p_aec->binary.write_cu_cbp_dqp(p_aec, h, &p_cu->cu_info, h->i_slice_index, h->lcu.last_dequant);
        rdcost = h->f_lambda_mode * rate_hdr + rdcost_luma;

        if (rdcost < *min_rdcost) {
            *min_rdcost = rdcost;
            h->copy_aec_state_rdo(&p_layer->cs_cu, p_aec);    /* store coding state for the best mode */
            cu_store_parameters(h, p_cu, best);
            b_need_swap_buf = 1;
        }
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);
    }

    /* 4, confirm the buffer pointers and record the best information */
    if (best->p_rec[0] == rec_bak_y && b_need_swap_buf) {
        XRD_SWAP_PTR(best->p_rec[0],   p_cu->cu_info.p_rec[0]);
        XRD_SWAP_PTR(best->p_coeff[0], p_cu->cu_info.p_coeff[0]);
    }

    p_layer->mode_rdcost[mode] = min_mode_rdcost;    /* store the cost for SDIP fast algorithm */
}

/* ---------------------------------------------------------------------------
 * int scrFlag = 0;             // 0=noSCR, 1=strongSCR, 2=jmSCR
 */
static INLINE int 
tu_recon_inter_luma(xrd_t *h, aec_t *p_aec, cu_t *p_cu,
                    int i_level, int8_t *cbp, int blockidx, coeff_t *cur_blk, 
                    int x_pu, int y_pu, int w_pu, int h_pu)
{
    cu_layer_t *p_layer = cu_get_layer(h, p_cu->cu_info.i_level);
    int used_wavelet = (p_cu->cu_info.i_level == B64X64_IN_BIT && p_cu->cu_info.i_tu_split != TU_SPLIT_CROSS);
    int part_idx = PART_INDEX(w_pu, h_pu);
    int w_tr = w_pu >> used_wavelet;
    int h_tr = h_pu >> used_wavelet;
    int num_non_zero = 0;
    pel_t *p_fdec = p_cu->cu_info.p_rec[0] + y_pu * FREC_STRIDE + x_pu;
    pel_t *p_pred = p_layer->buf_pred_inter + y_pu * FREC_STRIDE + x_pu;
    coeff_t *coeff_y = p_cu->cu_info.p_coeff[0] + (blockidx << ((p_cu->cu_info.i_level - 1) << 1));

    tu_get_dct_coeff(h, cur_blk, part_idx, w_tr, h_tr);

    num_non_zero = tu_quant_forward(h, p_aec, p_cu, cur_blk, i_level, w_tr, h_tr,
                                    cu_get_qp(h, &p_cu->cu_info), 0, 1, DC_PRED);

    if (num_non_zero != 0) {
        *cbp |= (1 << blockidx);    // ָλΪ 1
        g_funcs.pixf.copy_ss[PART_INDEX(w_tr, h_tr)](coeff_y, w_tr, cur_blk, w_tr);

        tu_quant_inverse(h, p_cu, cur_blk, w_tr * h_tr, i_level, cu_get_qp(h, &p_cu->cu_info), 1);
        g_funcs.dctf.idct[part_idx](cur_blk, cur_blk, w_tr);

        g_funcs.pixf.add_ps[part_idx](p_fdec, FREC_STRIDE, p_pred, cur_blk, FREC_STRIDE, w_pu);
    } else {
        /* CBPָλֵCBPʼֵΪ0 */
        // ȫ鲻任ֻ追ԤֵΪعֵ
        coeff_y[0] = 0;
        if (p_cu->cu_info.i_tu_split) {
            g_funcs.pixf.copy_pp[part_idx](p_fdec, FREC_STRIDE, p_pred, FREC_STRIDE);
        }
    }

    return num_non_zero;
}


/* ---------------------------------------------------------------------------
 * ָʽع֡ԤⷽʽCUȷ
 * صǰCUʧ棨ɫȿʧ棩
 */
static
dist_t cu_recon_inter_luma(xrd_t *h, aec_t *p_aec, cu_t *p_cu,
                           int is_non_residual, int b_tu_split, 
                           int cbp_c, dist_t *distortion)
{
    cu_layer_t *p_layer = cu_get_layer(h, p_cu->cu_info.i_level);
    cu_parallel_t *p_enc = cu_get_enc_context(h, p_cu->cu_info.i_level);
    coeff_t *cur_blk   = p_enc->coeff_blk;
    coeff_t *coeff_bak = p_enc->coeff_bak;
    coeff_t *p_resi;
    int level = p_cu->cu_info.i_level;
    int num_nonzero = 0;
    int sum_dc_coeff = 0;
    int b_zero_block = 0;
    int blockidx;
    int pix_x = p_cu->i_pos_x;
    int pix_y = p_cu->i_pos_y;
    int cu_size = p_cu->i_size;
    int cu_size_2 = cu_size >> 1;
    int cu_size_4 = cu_size_2 >> 1;
    pel_t *p_fenc;
    pel_t *p_fdec;

    /* clear CBP */
    p_cu->cu_info.i_cbp = 0;

    /* encode for luma */
    cu_set_tu_split_type(h, &p_cu->cu_info, b_tu_split);

    if (is_non_residual) {  /* SKIP mode (or no residual coding) */
        int uvoffset = (FREC_CSTRIDE >> 1);
        int part_idx_c = PART_INDEX(cu_size_2, cu_size_2);
        int pix_x_c = pix_x >> 1;
        int pix_y_c = pix_y >> CHROMA_V_SHIFT;

        h->lcu.bypass_all_dmh |= (p_cu->cu_info.dmh_mode == 0);

        /* copy Y component and get distortion */
        p_fenc = h->lcu.p_fenc[0] + pix_y * FENC_STRIDE + pix_x;
        p_fdec = p_cu->cu_info.p_rec[0];
        g_funcs.pixf.copy_pp[PART_INDEX(cu_size, cu_size)](p_fdec, FREC_STRIDE, p_layer->buf_pred_inter, FREC_STRIDE);
        distortion[0] = g_funcs.pixf.ssd[PART_INDEX(cu_size, cu_size)](p_fenc, FENC_STRIDE, p_fdec, FREC_STRIDE);

        /* chroma distortion */
        if (cbp_c) {
            /* copy U component and get distortion */
            p_fenc = h->lcu.p_fenc[1] + pix_y_c * FENC_STRIDE + pix_x_c;
            p_fdec = p_cu->cu_info.p_rec[1];
            g_funcs.pixf.copy_pp[part_idx_c](p_fdec, FREC_CSTRIDE / 2, p_enc->buf_pred_inter_c, FREC_CSTRIDE);
            distortion[1] = g_funcs.pixf.ssd[part_idx_c](p_fenc, FENC_STRIDE, p_fdec, FREC_CSTRIDE / 2);

            /* copy V component and get distortion */
            p_fenc = h->lcu.p_fenc[2] + pix_y_c * FENC_STRIDE + pix_x_c;
            p_fdec = p_cu->cu_info.p_rec[2];
            g_funcs.pixf.copy_pp[part_idx_c](p_fdec, FREC_CSTRIDE / 2, p_enc->buf_pred_inter_c + uvoffset, FREC_CSTRIDE);
            distortion[2] = g_funcs.pixf.ssd[part_idx_c](p_fenc, FENC_STRIDE, p_fdec, FREC_CSTRIDE / 2);
        } 

        return distortion[0] + distortion[1] + distortion[2];
    } else if (p_cu->cu_info.i_tu_split) {
        int pix_cu_x = 0;
        int pix_cu_y = 0;

        switch (p_cu->cu_info.i_tu_split) {
        case TU_SPLIT_HOR:
            g_funcs.pixf.copy_ss[PART_INDEX(cu_size, cu_size)](cur_blk, cu_size, coeff_bak, cu_size);
            for (blockidx = 0; blockidx < 4; blockidx++, pix_cu_y += cu_size_4) {
                p_resi = cur_blk + pix_cu_y * cu_size + pix_cu_x;
                num_nonzero += tu_recon_inter_luma(h, p_aec, p_cu, level - 1, &p_cu->cu_info.i_cbp, blockidx, p_resi, pix_cu_x, pix_cu_y, cu_size, cu_size_4);
                sum_dc_coeff += XRD_ABS(p_cu->cu_info.p_coeff[0][pix_cu_y * cu_size + pix_cu_x]);
            }
            break;
        case TU_SPLIT_VER:
            for (blockidx = 0; blockidx < 4; blockidx++, pix_cu_x += cu_size_4) {
                p_resi = coeff_bak + pix_cu_y * cu_size + pix_cu_x;
                g_funcs.pixf.copy_ss[PART_INDEX(cu_size_4, cu_size)](cur_blk, cu_size_4, p_resi, cu_size);
                num_nonzero += tu_recon_inter_luma(h, p_aec, p_cu, level - 1, &p_cu->cu_info.i_cbp, blockidx, cur_blk, pix_cu_x, pix_cu_y, cu_size_4, cu_size);
                sum_dc_coeff += XRD_ABS(p_cu->cu_info.p_coeff[0][pix_cu_y * cu_size + pix_cu_x]);
            }
            break;
        default:
            for (blockidx = 0; blockidx < 4; blockidx++) {
                pix_cu_x = (blockidx & 1) * cu_size_2;
                pix_cu_y = (blockidx >> 1) * cu_size_2;
                p_resi = coeff_bak + pix_cu_y * cu_size + pix_cu_x;
                g_funcs.pixf.copy_ss[PART_INDEX(cu_size_2, cu_size_2)](cur_blk, cu_size_2, p_resi, cu_size);
                num_nonzero += tu_recon_inter_luma(h, p_aec, p_cu, level - 1, &p_cu->cu_info.i_cbp, blockidx, cur_blk, pix_cu_x, pix_cu_y, cu_size_2, cu_size_2);
                sum_dc_coeff += XRD_ABS(p_cu->cu_info.p_coeff[0][pix_cu_y * cu_size + pix_cu_x]);
            }
            break;
        }

        // ǰCUϵ LUMA_COEFF_COST DCϵ£϶Ϊȫ
        b_zero_block = (num_nonzero <= LUMA_COEFF_COST && sum_dc_coeff <= MAX_COEFF_QUASI_ZERO);
    } else {

        num_nonzero += tu_recon_inter_luma(h, p_aec, p_cu, level, &p_cu->cu_info.i_cbp, 0, coeff_bak, 0, 0, cu_size, cu_size);

        // ǰCUб任ķϵ LUMA_COEFF_COST DCϵ£϶Ϊȫ
        sum_dc_coeff = XRD_ABS(p_cu->cu_info.p_coeff[0][0]);
        b_zero_block = (num_nonzero <= LUMA_COEFF_COST && sum_dc_coeff <= MAX_COEFF_QUASI_ZERO);

    }

    if (b_zero_block) {
        h->lcu.bypass_all_dmh |= (h->i_type == SLICE_TYPE_F && p_cu->cu_info.dmh_mode == 0);
        p_cu->cu_info.i_cbp = 0;
        g_funcs.pixf.copy_pp[PART_INDEX(cu_size, cu_size)](p_cu->cu_info.p_rec[0], FREC_STRIDE,
                                                           p_layer->buf_pred_inter, FREC_STRIDE);
    }

    /* set CBP */
    p_cu->cu_info.i_cbp += (int8_t)cbp_c;

    /* luma distortion */
    p_fenc = h->lcu.p_fenc[0] + pix_y * FENC_STRIDE + pix_x;
    p_fdec = p_cu->cu_info.p_rec[0];
    distortion[0] = g_funcs.pixf.ssd[PART_INDEX(cu_size, cu_size)](p_fenc, FENC_STRIDE, p_fdec, FREC_STRIDE);
    return distortion[0] + distortion[1] + distortion[2];
}

/* ---------------------------------------------------------------------------
 * R-D cost for a inter cu whether split or not
 * Return: rate-distortion cost of cu when TU is split or not
 */
static int tu_rdcost_inter(xrd_t *h, aec_t *p_aec, cu_t *p_cu,
                           dist_t distortion, int rate_chroma, rdcost_t *rdcost)
{
    int mode = p_cu->cu_info.i_mode;
    int level = p_cu->cu_info.i_level;
    int rate;
    int block_idx;
    cu_parallel_t *p_enc = cu_get_enc_context(h, level);
    int i_left_cmode = DM_PRED_C;
    /* check left */
    if (p_cu->p_left_cu != NULL) {
        i_left_cmode = p_cu->p_left_cu->i_intra_mode_c;
    }

    /* -------------------------------------------------------------
     * get rate */

    /* rate of cu header */
    rate = p_aec->binary.est_cu_header(p_aec, h, &p_cu->cu_info, i_left_cmode);

    /* rate of motion information */
    rate += p_aec->binary.est_cu_refs_mvds(p_aec, h, &p_cu->cu_info);

    /* tu information */
    if (mode != PRED_SKIP || p_cu->cu_info.i_cbp) {
        int bits_left = rdo_get_left_bits(h, *rdcost, distortion);
        int cur_bits_left;
        /* rate of cbp & dqp */
        rate += p_aec->binary.write_cu_cbp_dqp(p_aec, h, &p_cu->cu_info, h->i_slice_index, h->lcu.last_dequant);

        /* rate of luma coefficients */
        if (p_cu->cu_info.i_tu_split != TU_SPLIT_NON) {
            int use_wavelet = (level == B64X64_IN_BIT && p_cu->cu_info.i_tu_split != TU_SPLIT_CROSS);
            int i_tu_level = level - 1 - use_wavelet;
            for (block_idx = 0; block_idx < 4; block_idx++) {
                if (p_cu->cu_info.i_cbp & (1 << block_idx)) {
                    cb_t tb;
                    cur_bits_left = bits_left - rate;

                    cu_init_transform_block(p_cu->cu_info.i_level, p_cu->cu_info.i_tu_split, block_idx, &tb);

                    rate += p_aec->binary.est_luma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info,
                                                               p_cu->cu_info.p_coeff[0] + (block_idx << ((p_cu->cu_info.i_level - 1) << 1)),
                                                               i_tu_level, xrd_log2u(tb.w) - use_wavelet, 0, 0, cur_bits_left);
                }
            }
        } else {
            if (p_cu->cu_info.i_cbp & 15) {
                int i_tu_level = level - (level == B64X64_IN_BIT);
                cur_bits_left = bits_left - rate;
                rate += p_aec->binary.est_luma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info, p_cu->cu_info.p_coeff[0],
                                                           i_tu_level, i_tu_level, 0, 0, cur_bits_left);
            }
        }

        level--;
        if (p_cu->cu_info.i_cbp & (1 << 4)) {
            cur_bits_left = bits_left - rate;
            rate += p_aec->binary.est_chroma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info,
                p_cu->cu_info.p_coeff[1],
                level, cur_bits_left);
        }
        if (p_cu->cu_info.i_cbp & (1 << 5)) {
            cur_bits_left = bits_left - rate;
            rate += p_aec->binary.est_chroma_block_coeff(p_aec, &p_enc->runlevel, h, &p_cu->cu_info,
                p_cu->cu_info.p_coeff[2],
                level, cur_bits_left);
        }
    }

    /* -------------------------------------------------------------
     * get rate-distortion cost */
    *rdcost = distortion + h->f_lambda_mode * rate;

    return p_cu->cu_info.i_cbp;
}

/* ---------------------------------------------------------------------------
 * ȡȡɫȷԤֵMVǷЧΧ
 */
static ALWAYS_INLINE
int rdo_get_pred_inter(xrd_t *h, cu_t *p_cu, int cal_luma_chroma)
{
    cu_parallel_t *p_enc = cu_get_enc_context(h, p_cu->cu_info.i_level);
    cu_layer_t *p_layer = cu_get_layer(h, p_cu->cu_info.i_level);
    int blockidx;

    /* get prediction data */
    for (blockidx = 0; blockidx < p_cu->cu_info.num_pu; blockidx++) {
        cb_t cur_cb = p_cu->cu_info.cb[blockidx];
        int start_x = cur_cb.x;
        int start_y = cur_cb.y;
        int width   = cur_cb.w;
        int height  = cur_cb.h;
        int pix_x   = p_cu->i_pix_x + start_x;
        int pix_y   = p_cu->i_pix_y + start_y;

        mv_t mv_1st, mv_2nd;   // һǰB֡Ԥ⣩͵ڶ˶ʸ
        int ref_1st, ref_2nd;  // һǰB֡Ԥ⣩͵ڶ򣩲ο֡
        int num_mvs;
        int b_mv_valid;        // MVǷЧСȡֵǷڱ׼涨ЧΧ
        pel_t *p_temp = p_enc->buf_pixel_temp;
        pel_t *p_pred;
        xrd_frame_t *p_ref1 = NULL;
        xrd_frame_t *p_ref2 = NULL;

        /* MV1Ϊ˫ο֡/DMHԤ */
        num_mvs = cu_get_mvs_for_mc(h, p_cu, blockidx, &mv_1st, &mv_2nd, &ref_1st, &ref_2nd);
        b_mv_valid = check_mv_range(h, &mv_1st, ref_1st, pix_x, pix_y, width, height);
        if (num_mvs > 1) {
            b_mv_valid &= check_mv_range(h, &mv_2nd, ref_2nd, pix_x, pix_y, width, height);
            get_mv_for_mc(h, &mv_2nd, pix_x, pix_y, width, height);
            p_ref2 = h->fref[ref_2nd];
        }
        get_mv_for_mc(h, &mv_1st, pix_x, pix_y, width, height);
        p_ref1 = h->fref[ref_1st];

        if (!b_mv_valid && p_cu->cu_info.i_mode != PRED_SKIP) {
            return 0;
        }

        /* y component */
        if (cal_luma_chroma & 1) {
            p_pred = p_layer->buf_pred_inter + start_y * FREC_STRIDE + start_x;

            mc_luma(p_pred, FREC_STRIDE, mv_1st.x, mv_1st.y, width, height, p_ref1);
            if (num_mvs > 1) {
                mc_luma(p_temp, width, mv_2nd.x, mv_2nd.y, width, height, p_ref2);
                g_funcs.pixf.avg[PART_INDEX(width, height)](p_pred, FREC_STRIDE, p_pred, FREC_STRIDE, p_temp, width, 32);
            }
        }

        /* u and v component */
        if (h->param.chroma_format == CHROMA_420 && (cal_luma_chroma & 2)) {
            int uvoffset = (FREC_CSTRIDE >> 1);
            start_x >>= 1;
            width   >>= 1;
            pix_x   >>= 1;
            start_y >>= CHROMA_V_SHIFT;
            pix_y   >>= CHROMA_V_SHIFT;
            height  >>= CHROMA_V_SHIFT;

            p_pred = p_enc->buf_pred_inter_c + start_y * FREC_CSTRIDE + start_x;

            /* u component */
            mc_chroma(p_pred, p_pred + uvoffset, FREC_CSTRIDE, 
                      mv_1st.x, mv_1st.y, width, height, p_ref1);

            if (num_mvs > 1) {
                mc_chroma(p_temp, p_temp + uvoffset, FREC_CSTRIDE,
                          mv_2nd.x, mv_2nd.y, width, height, p_ref2);

                if (width != 2 && width != 6 && height != 2 && height != 6) {
                    pixel_avg_pp_t func_avg = g_funcs.pixf.avg[PART_INDEX(width, height)];
                    func_avg(p_pred           , FREC_CSTRIDE, p_pred           , FREC_CSTRIDE, p_temp           , FREC_CSTRIDE, 32);
                    func_avg(p_pred + uvoffset, FREC_CSTRIDE, p_pred + uvoffset, FREC_CSTRIDE, p_temp + uvoffset, FREC_CSTRIDE, 32);
                } else {
                    g_funcs.pixf.average(p_pred, FREC_CSTRIDE / 2, p_pred, FREC_CSTRIDE / 2, p_temp, FREC_CSTRIDE / 2, width, height * 2);
                }
            }
        }
    }

    return 1;
}

/* ---------------------------------------------------------------------------
 * compute rd-cost for inter cu
 * return 1, means it is the best mode
 *        0, means it is not the best mode
 */
static
int cu_rdcost_inter(xrd_t *h, aec_t *p_aec, cu_t *p_cu, rdcost_t *min_rdcost, cu_info_t *p_best)
{
    int mode = p_cu->cu_info.i_mode;
    int cu_size = p_cu->i_size;
    int tmp_cbp;                /* cbp for i_tu_split = 1*/
    int cbp_c = 0;
    int rate_chroma = 0;
    dist_t dist_chroma   = 0;
    dist_t distortion[3] = {0, 0, 0}; // dist for Y U V
    dist_t dist_split    = 0;
    dist_t dist_notsplit = 0;
    dist_t best_dist_cur = 0;
    rdcost_t rdcost = *min_rdcost;   // ʼΪRDCost
    rdcost_t rdcost_split = rdcost;
    pel_t *p_fenc = h->lcu.p_fenc[0] + p_cu->i_pos_y * FENC_STRIDE + p_cu->i_pos_x;
    cu_layer_t *p_layer  = cu_get_layer(h, p_cu->cu_info.i_level);
    cu_parallel_t *p_enc = cu_get_enc_context(h, p_cu->cu_info.i_level);

    /* -------------------------------------------------------------
     * 1, begin
     */
    p_cu->cu_info.i_cbp = 0;
    p_cu->cu_info.i_tu_split = TU_SPLIT_NON;  // cu_set_tu_split_type(h, &p_cu->cu_info, 0);

    /* set reference frame and block mode */
    cu_init_pu_inter(h, &p_cu->cu_info, p_cu->cu_info.i_level, mode);

    /* store coding state */
    h->copy_aec_state_rdo(&p_layer->cs_rdo, p_aec);

    /* -------------------------------------------------------------
    * 2. get prediction data
    */

    if (!rdo_get_pred_inter(h, p_cu, 3)) {
        return 0;
    }
    /* -------------------------------------------------------------
     * 3, tu decision
     */
    /* 3.1, check chroma residual coding */    
    if (h->param.chroma_format == CHROMA_420){
        cbp_c = cu_recon_chroma(h, p_aec, p_cu, distortion + 1);
        dist_chroma = distortion[1] + distortion[2];

    }

    /* 3.2, check luma CU tu-split type and CBP */
    /* 3.2.1, get luma residual */
    g_funcs.pixf.sub_ps[PART_INDEX(cu_size, cu_size)](p_enc->coeff_bak, cu_size, 
                                                      p_fenc, p_layer->buf_pred_inter, 
                                                      FENC_STRIDE, FREC_STRIDE);

    /* 3.2.2, check whether TU split is essential */
    //b_try_tu_split
    h->copy_aec_state_rdo(&p_enc->cs_tu, p_aec); /* store coding state for tu depth = 1 */

    dist_split = cu_recon_inter_luma(h, &p_enc->cs_tu, p_cu, 0, 1, cbp_c, distortion);
    tmp_cbp = tu_rdcost_inter(h, &p_enc->cs_tu, p_cu, dist_split, rate_chroma, &rdcost_split);

    /* store dct coefficients, rec data and coding state for tu depth = 1*/
    XRD_SWAP_PTR(p_cu->cu_info.p_rec[0], p_layer->p_rec_tmp[0]);
    XRD_SWAP_PTR(p_cu->cu_info.p_coeff[0], p_layer->p_coeff_tmp[0]);

    /* 3.2.4, get cost with tu depth equals 0 */
    //b_try_tu_no_split
    dist_notsplit = cu_recon_inter_luma(h, p_aec, p_cu, 0, 0, cbp_c, distortion);
    tu_rdcost_inter(h, p_aec, p_cu, dist_notsplit, rate_chroma, &rdcost);

    /* 3.2.5, choose the best tu depth (whether split or not) */
    if (rdcost > rdcost_split) {
        /* the best tu depth is 1 */
        rdcost = rdcost_split;
        best_dist_cur = dist_split;
        cu_set_tu_split_type(h, &p_cu->cu_info, 1);

        /* restore the cbp, dct coefficients, rec data and coding state for tu depth = 1*/
        p_cu->cu_info.i_cbp = (int8_t)tmp_cbp;
        XRD_SWAP_PTR(p_cu->cu_info.p_rec[0], p_layer->p_rec_tmp[0]);
        XRD_SWAP_PTR(p_cu->cu_info.p_coeff[0], p_layer->p_coeff_tmp[0]);

        h->copy_aec_state_rdo(p_aec, &p_enc->cs_tu); /* restore coding state */
    } else {
        best_dist_cur = dist_notsplit;
    }

    // }  // if (!skip_direct)

    /* 3.3, check skip mode for PRED_SKIP when CBP is nonzero */
    if (IS_SKIP_MODE(p_cu->cu_info.i_mode) && p_cu->cu_info.i_cbp != 0) {
        rdcost_t rdcost_skip = MAX_COST;
        dist_t dist_total_skip;
        int best_tu_split_type = p_cu->cu_info.i_tu_split;
    
        if (best_tu_split_type == TU_SPLIT_NON) {
            h->copy_aec_state_rdo(&p_enc->cs_tu, p_aec); /* store coding state for best Direct mode */
        }
    
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);/* restore coding state */
    
        tmp_cbp = p_cu->cu_info.i_cbp;
        /* backup reconstruction buffers, prepare for SKIP mode */
        XRD_SWAP_PTR(p_cu->cu_info.p_rec[0], p_layer->p_rec_tmp[0]);
        if (cbp_c != 0) {
            XRD_SWAP_PTR(p_cu->cu_info.p_rec[1], p_layer->p_rec_tmp[1]);
            XRD_SWAP_PTR(p_cu->cu_info.p_rec[2], p_layer->p_rec_tmp[2]);
        }
    
        /* check SKIP Mode */
        dist_total_skip = cu_recon_inter_luma(h, p_aec, p_cu, 1, 0, cbp_c, distortion);
        tu_rdcost_inter(h, p_aec, p_cu, dist_total_skip, rate_chroma, &rdcost_skip);

        if (rdcost_skip <= rdcost) {
            rdcost = rdcost_skip;    /* skip mode is the best */
            best_dist_cur = dist_total_skip;
            p_cu->cu_info.i_tu_split = TU_SPLIT_NON;
        } else {
            h->copy_aec_state_rdo(p_aec, &p_enc->cs_tu); /* restore coding state */
            /* revert buffers */
            XRD_SWAP_PTR(p_cu->cu_info.p_rec[0], p_layer->p_rec_tmp[0]);
            if (cbp_c != 0) {
                XRD_SWAP_PTR(p_cu->cu_info.p_rec[1], p_layer->p_rec_tmp[1]);
                XRD_SWAP_PTR(p_cu->cu_info.p_rec[2], p_layer->p_rec_tmp[2]);
            }
    
            p_cu->cu_info.i_cbp = (int8_t)(tmp_cbp);
            p_cu->cu_info.i_tu_split = (int8_t)(best_tu_split_type);
        }
    }

    /* -------------------------------------------------------------
     * 4, store the min cost for current cu mode
     */
    p_layer->mode_rdcost[mode] = XRD_MIN(rdcost, p_layer->mode_rdcost[mode]);

    /* -------------------------------------------------------------
     * 5, update the min cost, restore the coding state and return
     */
    if (rdcost >= *min_rdcost) {
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);
        return 0;  /* return code = 0, means it is not the best mode */
    } else {
        *min_rdcost = rdcost;
        p_cu->best_dist_total = best_dist_cur;
        /* store coding state for the best mode */
        h->copy_aec_state_rdo(&p_layer->cs_cu, p_aec);
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_rdo);
        /* update best CU information */
        cu_store_parameters(h, p_cu, p_best);

        return 1;  /* return code = 1, means it is the best mode */
    }
}

/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE void cu_set_mvs_noskip(cu_t *p_cu, int blockidx, int ref1, mv_t *pmv1, int ref2, mv_t *pmv2)
{
    const int mode = p_cu->cu_info.i_mode;

    if (mode == PRED_2Nx2N) {
        int k;
        for (k = 0; k < 4; k++) {
            p_cu->mc.mv[k][0].v = pmv1->v;
            p_cu->mc.mv[k][1].v = pmv2->v;
            p_cu->cu_info.ref_idx_1st[k] = (int8_t)ref1;
            p_cu->cu_info.ref_idx_2nd[k] = (int8_t)ref2;
        }
    } else {
        p_cu->mc.mv              [blockidx][0].v = pmv1->v;
        p_cu->cu_info.ref_idx_1st[blockidx] = (int8_t)ref1;
        
        p_cu->mc.mv              [blockidx][1].v = pmv2->v;
        p_cu->cu_info.ref_idx_2nd[blockidx] = (int8_t)ref2;
    }
}

/* ---------------------------------------------------------------------------
 */
static
rdcost_t cu_rdo_motion_estimation(xrd_t *h, cu_t *p_cu, xrd_me_t *p_me, int dualpred_enabled)
{
    const int mode = p_cu->cu_info.i_mode;
    const int block_num = mode == PRED_2Nx2N ? 1 : 2;
    int best_fwd_ref = 0;
    int best_pdir = PDIR_FWD;
    int dual_best_fst_ref = 0;
    int dual_best_snd_ref = 0;
    int pu_idx;
    dist_t fwd_cost = MAX_DISTORTION;
    dist_t bwd_cost = 0;
    rdcost_t total_cost = 0;
    int ref1, ref2;  // best references
    mv_t mv1, mv2;   // best mvs
    cb_t *p_cb;
    cu_mode_t *p_mode = cu_get_layer_mode(h, p_cu->cu_info.i_level);
    cu_mv_mode_t *p_mv_mode;

    p_cu->cu_info.b8pdir[0] = p_cu->cu_info.b8pdir[1] = p_cu->cu_info.b8pdir[2] = p_cu->cu_info.b8pdir[3] = (int8_t)PDIR_FWD;
    p_me->b_search_dmh = (dualpred_enabled == -1);

    // motion estimation for 2Nx2N, 2NxN, Nx2N, AMP blocks
    for (pu_idx = 0; pu_idx < block_num; pu_idx++) {
        mv1.v = mv2.v = 0;
        p_cb = &p_cu->cu_info.cb[pu_idx];
        cu_get_neighbors(h, p_cu, p_cb);

        /* һPUҪ½MEMVP䣩 */
        if (dualpred_enabled < 0 && pu_idx == 0) {
            best_fwd_ref = p_mode->ref_idx_single[0];
        } else {
            best_fwd_ref = pred_inter_search_single(h, p_cu, p_cb, pu_idx, p_me, &fwd_cost, &bwd_cost);
        }

        p_mv_mode = &p_mode->mvs[mode][pu_idx];

        if (dualpred_enabled < 0) {
            best_pdir = PDIR_FWD;
            mv1 = p_mv_mode->all_single_mv[best_fwd_ref];
            cu_set_mvs_noskip(p_cu, pu_idx, best_fwd_ref, &mv1, INVALID_REF, &mv2);
        } else if (h->i_type == SLICE_TYPE_F) {
            dist_t dual_mcost = MAX_DISTORTION;
            if (dualpred_enabled && (!(p_cu->cu_info.i_level == B8X8_IN_BIT && mode >= PRED_2NxN && mode <= PRED_nRx2N))) {
                pred_inter_search_dual(h, p_cu, p_cb, pu_idx, p_me, &dual_mcost, &dual_best_fst_ref, &dual_best_snd_ref);
            }

            if (fwd_cost <= dual_mcost) {
                best_pdir = PDIR_FWD;
                ref1 = best_fwd_ref;
                mv1 = p_mv_mode->all_single_mv[ref1];
                cu_set_mvs_noskip(p_cu, pu_idx, ref1, &mv1, INVALID_REF, &mv2);
                p_cu->mvcost[pu_idx] = fwd_cost;
            } else {
                fwd_cost = dual_mcost;
                best_pdir = PDIR_DUAL;
                ref1 = dual_best_fst_ref;
                ref2 = dual_best_snd_ref;
                mv1 = p_mv_mode->all_dual_mv_1st[ref1];
                mv2 = p_mv_mode->all_dual_mv_2nd[ref1];
                cu_set_mvs_noskip(p_cu, pu_idx, ref1, &mv1, ref2, &mv2);
                p_cu->mvcost[pu_idx] = dual_mcost;
            }
        } else if (h->i_type == SLICE_TYPE_B) {
            dist_t sym_mcost = MAX_DISTORTION;
            dist_t bid_mcost = MAX_DISTORTION;

            best_fwd_ref = 0;               // must reset
            if (!((p_cu->cu_info.i_level == B8X8_IN_BIT) && (mode >= PRED_2NxN && mode <= PRED_nRx2N))) {
                pred_inter_search_bi(h, p_cu, p_cb, pu_idx, p_me, &sym_mcost, &bid_mcost);
            }

            if (fwd_cost <= bwd_cost && fwd_cost <= sym_mcost && fwd_cost <= bid_mcost) {
                best_pdir = PDIR_FWD;
                ref1 = B_FWD;
                mv1 = p_mv_mode->all_single_mv[ref1];
                cu_set_mvs_noskip(p_cu, pu_idx, ref1, &mv1, INVALID_REF, &mv2);
                p_cu->mvcost[pu_idx] = fwd_cost;
            } else if (bwd_cost <= fwd_cost && bwd_cost <= sym_mcost && bwd_cost <= bid_mcost) {
                fwd_cost = bwd_cost;
                best_pdir = PDIR_BWD;
                ref2 = B_BWD;
                mv2 = p_mv_mode->all_single_mv[B_BWD];
                cu_set_mvs_noskip(p_cu, pu_idx, INVALID_REF, &mv1, ref2, &mv2);
                p_cu->mvcost[pu_idx] = bwd_cost;
            } else if (sym_mcost <= fwd_cost && sym_mcost <= bwd_cost && sym_mcost <= bid_mcost) {
                int dist_fwd = calculate_distance(h, B_FWD);  // fwd
                int dist_bwd = calculate_distance(h, B_BWD);  // bwd
                fwd_cost = sym_mcost;
                best_pdir = PDIR_SYM;
                ref1 = B_FWD;
                ref2 = B_BWD;
                mv1 = p_mv_mode->all_sym_mv[0];
                mv2.x = -scale_mv_skip  (   mv1.x, dist_bwd, dist_fwd);
                mv2.y = -scale_mv_skip_y(h, mv1.y, dist_bwd, dist_fwd);
                cu_set_mvs_noskip(p_cu, pu_idx, ref1, &mv1, ref2, &mv2);
                p_cu->mvcost[pu_idx] = sym_mcost;
            } else {
                fwd_cost = bid_mcost;
                best_pdir = PDIR_BID;
                ref1 = B_FWD;
                ref2 = B_BWD;
                mv1 = p_mv_mode->all_dual_mv_1st[0];
                mv2 = p_mv_mode->all_dual_mv_2nd[0];
                cu_set_mvs_noskip(p_cu, pu_idx, ref1, &mv1, ref2, &mv2);
                p_cu->mvcost[pu_idx] = bid_mcost;
            }
        } else {
            ref1 = best_fwd_ref;
            mv1 = p_mv_mode->all_single_mv[ref1];
            cu_set_mvs_noskip(p_cu, pu_idx, ref1, &mv1, INVALID_REF, &mv2);
            p_cu->mvcost[pu_idx] = p_me->mvcost[PDIR_FWD];
        }

        total_cost += fwd_cost;

        // store reference frame index and direction parameters
        p_mode->ref_idx_single[pu_idx] = (int8_t)best_fwd_ref;
        p_cu->cu_info.b8pdir[pu_idx] = (int8_t)best_pdir;
    }

    cu_get_mvds(h, p_cu);  // MVD

    return total_cost;  // СCost
}

/* ---------------------------------------------------------------------------
 * ͨ֡Ԥ黮ַʽӦCost
 */
static
void cu_check_inter_partition(xrd_t *h, aec_t *p_aec, cu_t *p_cu, int mode, int i_level, 
                              cu_info_t *best, rdcost_t *p_min_rdcost,
                              int b_dhp_enabled, int b_check_dmh)
{
    /* set reference frame and block mode */
    cu_init_pu_inter(h, &p_cu->cu_info, i_level, mode);

    /* ME */
    cu_rdo_motion_estimation(h, p_cu, &h->lcu.me_state, b_dhp_enabled);

    h->lcu.bypass_all_dmh = 0;

    /* һ֡仮ģʽRDCostȷűģʽ */
    p_cu->cu_info.directskip_wsm_idx = 0;
    p_cu->cu_info.directskip_mhp_idx = DS_NONE;
    p_cu->cu_info.dmh_mode = 0;
    cu_rdcost_inter(h, p_aec, p_cu, p_min_rdcost, best);

    /* DMHģʽ */
    if (h->i_type == SLICE_TYPE_F && h->param.enable_dmh && !h->lcu.bypass_all_dmh && b_check_dmh
        && !(i_level == B8X8_IN_BIT && mode != PRED_2Nx2N)) {  // disable 8x4 or 4x8 2MVs/PU mode
        int dmh_mode_candidate = 0;
        int max_dmh_mode;
        int best_dmh_mode = 0;
        int dmh_mode;

        if (p_cu->cu_info.b8pdir[0] == PDIR_FWD && p_cu->cu_info.b8pdir[1] == PDIR_FWD &&
            p_cu->cu_info.b8pdir[2] == PDIR_FWD && p_cu->cu_info.b8pdir[3] == PDIR_FWD) {
            /* MEȷŵPUԤⷽΪǰ򣬴ʱֻҪDMHģʽ */
            dmh_mode = 1;
        } else { // DHP Ҳο֡Ϊ2ʱп
            /* ŵPUа˫ǰ飬ʱҪPUΪǰʱRDCostsٱDMHģʽ */
            /* ʱMEͬʱһPUҪ */
            cu_rdo_motion_estimation(h, p_cu, &h->lcu.me_state, -1);
            dmh_mode = 0;
        }

        /* ܼ 2 * (DMH_MODE_NUM - 1) + 1 ģʽ */
        max_dmh_mode = DMH_MODE_NUM + DMH_MODE_NUM - 1;

        // ĳģʽµĲвΪȫʱкdmhģʽ
        for (; dmh_mode < max_dmh_mode && !h->lcu.bypass_all_dmh; dmh_mode++) {

            if (dmh_mode > (DMH_MODE_NUM - 1)) {
                if (best_dmh_mode != (dmh_mode - (DMH_MODE_NUM - 1))) { // ֻͬչ
                    continue;
                }
            }

            p_cu->cu_info.dmh_mode = (int8_t)dmh_mode;
            if (cu_rdcost_inter(h, p_aec, p_cu, p_min_rdcost, best)) {
                best_dmh_mode = dmh_mode;
            }
        }  // end loop of DMH modes
    }  // end of check DMH modes

}


/* ---------------------------------------------------------------------------
 */
static ALWAYS_INLINE
void cu_set_mvs_skip(xrd_t *h, cu_t *p_cu)
{
    int weighted_skip_mode = p_cu->cu_info.directskip_wsm_idx;
    int ds_mode = p_cu->cu_info.directskip_mhp_idx;
    const cu_mode_t *p_cu_mode = cu_get_layer_mode(h, p_cu->cu_info.i_level);
    int k;

    assert(p_cu->cu_info.i_mode == PRED_SKIP);
    assert(h->i_type != SLICE_TYPE_I);

    if (ds_mode != DS_NONE) {
        mv_t mv1 = p_cu_mode->skip_mv_1st[ds_mode];
        mv_t mv2 = p_cu_mode->skip_mv_2nd[ds_mode];
        int8_t ref1 = p_cu_mode->skip_ref_1st[ds_mode];
        int8_t ref2 = p_cu_mode->skip_ref_2nd[ds_mode];

        int i_dir_pred = tab_pdir_bskip[ds_mode];

        for (k = 0; k < 4; k++) {
            p_cu->cu_info.b8pdir[k] = (int8_t)i_dir_pred;
            p_cu->mc.mv[k][0] = mv1;
            p_cu->mc.mv[k][1] = mv2;
            p_cu->cu_info.ref_idx_1st[k] = ref1;
            p_cu->cu_info.ref_idx_2nd[k] = ref2;
        }
    } else if (weighted_skip_mode) {
        for (k = 0; k < 4; k++) {
            p_cu->cu_info.b8pdir[k] = PDIR_FWD;
            p_cu->mc.mv[k][0] = p_cu_mode->tskip_mv[k][0];
            p_cu->mc.mv[k][1] = p_cu_mode->tskip_mv[k][weighted_skip_mode];
            p_cu->cu_info.ref_idx_1st[k] = 0;
            p_cu->cu_info.ref_idx_2nd[k] = (int8_t)weighted_skip_mode;
        }
    } else if (h->i_type != SLICE_TYPE_B) {
        for (k = 0; k < 4; k++) {
            p_cu->cu_info.b8pdir[k] = PDIR_FWD;
            p_cu->mc.mv[k][0] = p_cu_mode->tskip_mv[k][0];
            p_cu->mc.mv[k][1].v = 0;
            p_cu->cu_info.ref_idx_1st[k] = 0;
            p_cu->cu_info.ref_idx_2nd[k] = INVALID_REF;
        }
    } else {
        for (k = 0; k < 4; k++) {
            p_cu->cu_info.b8pdir[k] = PDIR_SYM;
            p_cu->mc.mv[k][0] = p_cu_mode->tskip_mv[k][0];
            p_cu->mc.mv[k][1] = p_cu_mode->tskip_mv[k][1];
            p_cu->cu_info.ref_idx_1st[k] = B_FWD;
            p_cu->cu_info.ref_idx_2nd[k] = B_BWD;
        }
    }
}

/* ---------------------------------------------------------------------------
 */
typedef struct cu_skip_mc_t {
    mv_t         mv_1st[4];
    mv_t         mv_2nd[4];
    int8_t       ref_1st[4];
    int8_t       ref_2nd[4];
} cu_skip_mc_t;

/* ---------------------------------------------------------------------------
 * SkipMVԼ⵱ǰģʽMVǷ񱻱
 */
static ALWAYS_INLINE
int is_same_skip_mc_param(const cu_skip_mc_t *p_src1, const cu_skip_mc_t *p_src2)
{
    uint32_t *p1 = (uint32_t *)p_src1;
    uint32_t *p2 = (uint32_t *)p_src2;
    int num = sizeof(cu_skip_mc_t) >> 2;
    int i;

    for (i = 0; i < num; i++) {
        if (p1[i] != p2[i]) {
            return 0;
        }
    }

    return 1;
}

/* ---------------------------------------------------------------------------
 * SkipMVԼ⵱ǰģʽMVǷ񱻱
 */
static
int update_skip_mv_list(cu_skip_mc_t *p_skip_mvs, int i_num, cu_t *p_cu)
{
    cu_skip_mc_t cur_mc;
    int i;
    for (i = 0; i < 4; i++) {
        cur_mc.mv_1st[i].v = p_cu->mc.mv[i][0].v;
        cur_mc.mv_2nd[i].v = p_cu->mc.mv[i][1].v;
        cur_mc.ref_1st[i] = p_cu->cu_info.ref_idx_1st[i];
        cur_mc.ref_2nd[i] = p_cu->cu_info.ref_idx_2nd[i];
    }

    for (i = 0; i < i_num; i++) {
        if (is_same_skip_mc_param(p_skip_mvs + i, &cur_mc)) {
            break;
        }
    }

    if (i != i_num) {
        return 0;
    } else {
        memcpy(p_skip_mvs + i_num, &cur_mc, sizeof(cu_skip_mc_t));
        return 1;
    }
}


/* ---------------------------------------------------------------------------
 * Skip/DirectģʽıۣRDOѡȡŵSkipģʽ
 */
static
void cu_check_skip_direct_fullrdo(xrd_t *h, aec_t *p_aec, cu_info_t *p_best, cu_t *p_cu, rdcost_t *p_min_rdcost)
{
    int max_skip_mode_num, i;

    cb_t cur_cb = { { 0 } };
    cur_cb.w = cur_cb.h = (int8_t)p_cu->i_size;
    cu_get_neighbors(h, p_cu, &cur_cb);

    /* get Skip/Direct MVs and temporal SKIP mode number */
    max_skip_mode_num = h->lcu.get_skip_mvs(h, p_cu);

    /* 0, init cu data */
    p_cu->cu_info.dmh_mode = 0;

    /* 1, temporal skip mode, derive MV from temporal */
    p_cu->cu_info.directskip_mhp_idx = DS_NONE;
    p_cu->cu_info.directskip_wsm_idx = 0;

    /* ʱMVPԤֱRDCostٸŵRDCostȽϣ 3%ңʱ 20%~30% */
    cu_set_mvs_skip(h, p_cu);
    cu_rdcost_inter(h, p_aec, p_cu, p_min_rdcost, p_best);

    /* 2, Weighted skip mode, derive MV from temporal and scaling */
    for (i = 1; i < max_skip_mode_num; i++) {
        p_cu->cu_info.directskip_wsm_idx = (int8_t)i;
        cu_set_mvs_skip(h, p_cu);
        cu_rdcost_inter(h, p_aec, p_cu, p_min_rdcost, p_best);
    }

    /* 3, ĸspatial direct (single first, single second, dual first, dual second) */
    if (h->i_type == SLICE_TYPE_B || (h->i_type == SLICE_TYPE_F && h->param.enable_mhp_skip)) {
        p_cu->cu_info.directskip_wsm_idx = 0;
        for (i = 0; i < DS_MAX_NUM; i++) {
            p_cu->cu_info.directskip_mhp_idx = (int8_t)i;
            cu_init_pu_inter(h, &p_cu->cu_info, p_cu->cu_info.i_level, PRED_SKIP);
            cu_set_mvs_skip(h, p_cu);
            cu_rdcost_inter(h, p_aec, p_cu, p_min_rdcost, p_best);
        }
    }
}

/* ---------------------------------------------------------------------------
 * encode an intra cu (for I-picture)
 */
static
rdcost_t compress_cu_intra(xrd_t *h, aec_t *p_aec, cu_t *p_cu, cu_info_t *best, rdcost_t cost_limit)
{
    int i_level = p_cu->cu_info.i_level;
    cu_layer_t *p_layer = cu_get_layer(h, i_level);
    uint32_t intra_modes;   // valid intra modes
    rdcost_t split_flag_cost = 0;
    rdcost_t min_rdcost = MAX_COST;
    int mode;

    UNUSED_PARAMETER(cost_limit);
    if (i_level > MIN_CU_SIZE_IN_BIT) {
        split_flag_cost = h->f_lambda_mode * p_aec->binary.write_ctu_split_flag(p_aec, 0, i_level);
    }

    h->lcu.b_enable_rdoq     = (h->param.i_rdoq_level == RDOQ_ALL);
    h->lcu.b_2nd_rdcost_pass = 1;
    h->lcu.get_intra_dir_for_rdo_luma = h->get_intra_candidates_luma;

    //===== SET VALID MODES =====
    intra_modes = cu_get_valid_modes(h, h->i_type, i_level);

    // reset default parameters for chroma intra predictor
    p_cu->cu_info.i_intra_mode_c     = DC_PRED_C;
    p_cu->cu_info.directskip_wsm_idx = 0;
    p_cu->cu_info.directskip_mhp_idx = DS_NONE;

    //===== GET BEST MACROBLOCK MODE =====
    for (mode = PRED_I_2Nx2N; mode <= PRED_I_nx2N; mode++) {
        if (!(intra_modes & (1 << mode))) {
            continue;           // ֱģʽ
        }

        // init coding block(s)
        p_cu->cu_info.i_mode = (int8_t)mode;    // set cu type

        cu_check_intra(h, p_aec, p_cu, best, mode, &min_rdcost);
    }

    min_rdcost += split_flag_cost;
    return p_layer->best_rdcost = min_rdcost;
}

/* ---------------------------------------------------------------------------
 * encode an inter cu (for none I-picture)
 */
static
rdcost_t compress_cu_inter(xrd_t *h, aec_t *p_aec, cu_t *p_cu, cu_info_t *best,
                           uint32_t avail_modes,  // available prediction partition modes
                           rdcost_t min_rdcost,   // Cost of Skip/Direct mode
                           rdcost_t cost_limit)
{
    int b_dhp_enabled = h->param.enable_dhp && h->i_type == SLICE_TYPE_F && h->i_ref > 1;
    int i_level = p_cu->cu_info.i_level;
    int b_bypass_intra = 0;
    int b_check_dmh = 1;
    int mode;
    cu_layer_t *p_layer = cu_get_layer(h, p_cu->cu_info.i_level);

    /* -------------------------------------------------------------
     * 1, ʼ
     */
    UNUSED_PARAMETER(cost_limit);
    h->lcu.get_intra_dir_for_rdo_luma = h->get_intra_candidates_luma;

    h->lcu.b_enable_rdoq     = (h->param.i_rdoq_level == RDOQ_ALL);
    h->lcu.b_2nd_rdcost_pass = 0;

    for (mode = 0; mode < MAX_PRED_MODES; mode++) {
        p_layer->mode_rdcost[mode] = MAX_COST;
    }

    /* reset chroma intra predictor to default */
    p_cu->cu_info.i_intra_mode_c = DC_PRED_C;   // @luoflƳУᵼ²ƥ⣻20170304 19:52:32

    /* -------------------------------------------------------------
     * 2, SkipDirectģʽ
     */
    /* SKIP/Directģʽ */
    p_cu->cu_info.i_mode = PRED_SKIP;

    cu_check_skip_direct_fullrdo(h, p_aec, best, p_cu, &min_rdcost);

    p_layer->best_rdcost = min_rdcost;

    /* -------------------------------------------------------------
     * 3, Skip/Direct֡ģʽ
     */
    for (mode = 1; mode < MAX_INTER_MODES; mode++) {
        if (!(avail_modes & (1 << mode))) {
            continue;           // ֱģʽľ
        }

        /* -------------------------------------------------------------
         * Ա뵱ǰPUģʽ
         */
        p_cu->cu_info.i_mode = (int8_t)mode;

        cu_check_inter_partition(h, p_aec, p_cu, mode, i_level, best, &min_rdcost, b_dhp_enabled, b_check_dmh);
    }

    /* ͨ֡Ԥж֡Ҫ֡Ԥʱ֡ģʽ */
    if (!h->fenc->b_enable_intra) {
        b_bypass_intra = 1;
    }

    /* -------------------------------------------------------------
     * 4, get best intra mode
     */
    if (!b_bypass_intra) {
        for (mode = PRED_I_2Nx2N; mode <= PRED_I_nx2N; mode++) {
            if (!(avail_modes & (1 << mode))) {
                continue;           // ֱģʽľ
            }

            // init coding block(s)
            p_cu->cu_info.i_mode = (int8_t)mode;

            // cal rd-cost
            cu_check_intra(h, p_aec, p_cu, best, mode, &min_rdcost);

        }
    }   

    return p_layer->best_rdcost = min_rdcost;
}


/**
 * ===========================================================================
 * interface function defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * RDOPTʼʱòͬ֡CUСõģʽֱӲ
 */
void xrd_init_valid_mode_table(xrd_t *h)
{
    int frm_type;
    int level;

    for (frm_type = 0; frm_type < SLICE_TYPE_NUM; frm_type++) {
        int inter_frame = (frm_type != SLICE_TYPE_I);
        for (level = MIN_CU_SIZE_IN_BIT; level <= MAX_CU_SIZE_IN_BIT; level++) {
            uint32_t valid_pred_modes = 0;

            /* set validity of inter modes */
            if (inter_frame) {
                valid_pred_modes |= 1 << PRED_SKIP;
                valid_pred_modes |= 1 << PRED_2Nx2N;
                valid_pred_modes |= h->param.inter_2pu << PRED_2NxN;
                valid_pred_modes |= h->param.inter_2pu << PRED_Nx2N;
                if (h->param.enable_amp && level > MIN_CU_SIZE_IN_BIT) {
                    valid_pred_modes |= 1 << PRED_2NxnU;
                    valid_pred_modes |= 1 << PRED_2NxnD;
                    valid_pred_modes |= 1 << PRED_nLx2N;
                    valid_pred_modes |= 1 << PRED_nRx2N;
                }
            }

            /* set validity of intra modes */
            if (!inter_frame || h->param.enable_intra) {
                valid_pred_modes |= 1 << PRED_I_2Nx2N;
                valid_pred_modes |= (level == MIN_CU_SIZE_IN_BIT) << PRED_I_NxN;

                // only valid for 32x8,8x32, 16x4,4x16
                if (h->param.enable_sdip && (level == B16X16_IN_BIT || level == B32X32_IN_BIT)) {
                    valid_pred_modes |= 1 << PRED_I_2Nxn;
                    valid_pred_modes |= 1 << PRED_I_nx2N;
                }
            }

            // @luofl: SDIP is disabled here for speedup in inter frames
            if (inter_frame && level != MIN_CU_SIZE_IN_BIT) {
                valid_pred_modes &= ~((1 << PRED_I_2Nxn) | (1 << PRED_I_nx2N) | (1 << PRED_I_NxN));
            }
            h->valid_modes[frm_type][level - MIN_CU_SIZE_IN_BIT] = valid_pred_modes;
        }
    }
}

/* ---------------------------------------------------------------------------
 */
rdcost_t compress_ctu_intra_topdown(xrd_t *h, aec_t *p_aec, cu_t *p_cu, int i_level, int i_min_level, int i_max_level, rdcost_t cost_limit)
{
    aec_t cs_aec;
    cu_layer_t *p_layer    = cu_get_layer(h, i_level);
    cu_info_t *best        = &p_layer->cu_best;
    rdcost_t large_cu_cost = MAX_COST;
    rdcost_t split_cu_cost = MAX_COST;
    int b_inside_pic       = (p_cu->i_pix_x + p_cu->i_size <= h->i_width) && (p_cu->i_pix_y + p_cu->i_size <= h->i_height);
    int b_split_ctu        = (i_level > i_min_level || !b_inside_pic);
    int b_check_large_cu   = (b_inside_pic && i_level <= i_max_level);

    /* init current CU ---------------------------------------------
     */
    cu_init(h, p_cu, best, i_level);

    /* coding current CU -------------------------------------------
     */
    if (b_check_large_cu) {

        h->copy_aec_state_rdo(&cs_aec, p_aec);


        large_cu_cost = compress_cu_intra(h, &cs_aec, p_cu, best, cost_limit);


    }

    /* coding 4 sub-CUs --------------------------------------------
     */
    if (b_split_ctu) {
        int i;
        split_cu_cost = 0;

        // cal split cost
        if (b_inside_pic) {
            split_cu_cost += h->f_lambda_mode * p_aec->binary.write_ctu_split_flag(p_aec, 1, i_level);
        }

        for (i = 0; i < 4; i++) {
            cu_t *p_sub_cu = p_cu->sub_cu[i];

            if (p_sub_cu->i_pix_x >= h->i_width || p_sub_cu->i_pix_y >= h->i_height) {
                continue;       // current sub CU is outside the frame+*
            }

            split_cu_cost += compress_ctu_intra_topdown(h, p_aec, p_sub_cu, i_level - 1, i_min_level, i_max_level, large_cu_cost - split_cu_cost);

            if (split_cu_cost > large_cu_cost) {
                split_cu_cost = MAX_COST; // guide RDO to select large CU
                break;
            }
        }
    }

    /* decide split or not -----------------------------------------
     */
    if (large_cu_cost < split_cu_cost) {
        /* the larger cu is selected */
        cu_copy_stored_parameters(h, p_cu, best);
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_cu);
        split_cu_cost = large_cu_cost;
    }

    return split_cu_cost;
}

/* ---------------------------------------------------------------------------
 */
rdcost_t compress_ctu_intra_downtop(xrd_t *h, aec_t *p_aec, cu_t *p_cu, int i_level, int i_min_level, int i_max_level, rdcost_t cost_limit)
{
    aec_t cs_aec;
    cu_layer_t *p_layer    = cu_get_layer(h, i_level);
    cu_info_t *best        = &p_layer->cu_best;
    rdcost_t large_cu_cost = MAX_COST;
    rdcost_t split_cu_cost = MAX_COST;
    int b_inside_pic       = (p_cu->i_pix_x + p_cu->i_size <= h->i_width) && (p_cu->i_pix_y + p_cu->i_size <= h->i_height);
    int b_split_ctu        = (i_level > i_min_level || !b_inside_pic);
    int b_check_large_cu   = (b_inside_pic && i_level <= i_max_level);
    int num_subcu_split    = 0;

    /* init current CU ---------------------------------------------
     */
    cu_init(h, p_cu, best, i_level);

    if (b_check_large_cu) {
        h->copy_aec_state_rdo(&cs_aec, p_aec);
    }

    /* coding 4 sub-CUs --------------------------------------------
     */
    if (b_split_ctu) {
        int i;
        split_cu_cost = 0;

        // cal split cost
        if (b_inside_pic) {
            split_cu_cost += h->f_lambda_mode * p_aec->binary.write_ctu_split_flag(p_aec, 1, i_level);
        }

        for (i = 0; i < 4; i++) {
            cu_t *p_sub_cu = p_cu->sub_cu[i];

            if (p_sub_cu->i_pix_x >= h->i_width || p_sub_cu->i_pix_y >= h->i_height) {
                continue;       // current sub CU is outside the frame+*
            }

            split_cu_cost += compress_ctu_intra_downtop(h, p_aec, p_sub_cu, i_level - 1, i_min_level, i_max_level, large_cu_cost - split_cu_cost);
            if (p_sub_cu->is_ctu_split) {
                num_subcu_split++;
            }
        }
    }

    if (num_subcu_split > 0) {
        b_check_large_cu = FALSE;
    }

    /* coding current CU -------------------------------------------
     */
    if (b_check_large_cu) {
        large_cu_cost = compress_cu_intra(h, &cs_aec, p_cu, best, cost_limit);
    }

    /* decide split or not -----------------------------------------
     */
    if (large_cu_cost < split_cu_cost) {
        /* the larger cu is selected */
        cu_copy_stored_parameters(h, p_cu, best);
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_cu);
        split_cu_cost = large_cu_cost;
        p_cu->is_ctu_split = FALSE;
    } else {
        p_cu->is_ctu_split = TRUE;
    }

    return split_cu_cost;
}

/* ---------------------------------------------------------------------------
 */
rdcost_t compress_ctu_inter(xrd_t *h, aec_t *p_aec, cu_t *p_cu, int i_level, int i_min_level, int i_max_level, rdcost_t cost_limit)
{
    aec_t cs_aec;
    cu_layer_t *p_layer      = cu_get_layer(h, i_level);
    cu_info_t *best          = &p_layer->cu_best;
    rdcost_t large_cu_cost   = MAX_COST;
    rdcost_t split_cu_cost   = MAX_COST;
    rdcost_t split_flag_cost = 0;
    uint32_t avail_modes     = cu_get_valid_modes(h, h->i_type, i_level);
    int b_inside_pic         = (p_cu->i_pix_x + p_cu->i_size <= h->i_width) && (p_cu->i_pix_y + p_cu->i_size <= h->i_height);
    int b_split_ctu          = (i_level > i_min_level || !b_inside_pic);
    int b_check_large_cu     = (b_inside_pic && i_level <= i_max_level);

    /* init current CU ---------------------------------------------
     */
    cu_init(h, p_cu, best, i_level);

    /* coding current CU -------------------------------------------
     */
    if (b_check_large_cu) {
        h->copy_aec_state_rdo(&cs_aec, p_aec);
        if (i_level > MIN_CU_SIZE_IN_BIT) {
            split_flag_cost = h->f_lambda_mode * p_aec->binary.write_ctu_split_flag(&cs_aec, 0, i_level);
        }

        large_cu_cost = compress_cu_inter(h, &cs_aec, p_cu, best, avail_modes, large_cu_cost, cost_limit);
        large_cu_cost += split_flag_cost;

    }

    /* coding 4 sub-CUs --------------------------------------------
     */
    if (b_split_ctu) {
        int i;
        split_cu_cost = 0;

        // cal split cost
        if (b_inside_pic) {
            split_cu_cost += h->f_lambda_mode * p_aec->binary.write_ctu_split_flag(p_aec, 1, i_level);
        }

        for (i = 0; i < 4; i++) {
            cu_t *p_sub_cu = p_cu->sub_cu[i];
            

            if (p_sub_cu->i_pix_x >= h->i_width || p_sub_cu->i_pix_y >= h->i_height) {
                continue;       // current sub CU is outside the frame
            }

            split_cu_cost += compress_ctu_inter(h, p_aec, p_sub_cu, i_level - 1, i_min_level, i_max_level, large_cu_cost - split_cu_cost);

            if (split_cu_cost > large_cu_cost) {
                split_cu_cost = MAX_COST; // guide RDO to select large CU
                break;
            }

        }
    }

    /* decide split or not -----------------------------------------
     */
    if (large_cu_cost < split_cu_cost) {
        /* the larger cu is selected */
        cu_copy_stored_parameters(h, p_cu, best);
        h->copy_aec_state_rdo(p_aec, &p_layer->cs_cu);
        split_cu_cost = large_cu_cost;
        p_cu->is_ctu_split = FALSE;
    } else {
        p_cu->is_ctu_split = TRUE;
    }
    p_cu->rdcost = split_cu_cost;

    return split_cu_cost;
}

/* ---------------------------------------------------------------------------
 */
void xrd_rdo_init(intrinsic_func_t *pf)
{
    pf->compress_ctu[SLICE_TYPE_I] = compress_ctu_intra_topdown;
    pf->compress_ctu[SLICE_TYPE_P] = compress_ctu_inter;
    pf->compress_ctu[SLICE_TYPE_F] = compress_ctu_inter;
    pf->compress_ctu[SLICE_TYPE_B] = compress_ctu_inter;

    pf->get_skip_mv_predictors[SLICE_TYPE_I] = NULL;
    pf->get_skip_mv_predictors[SLICE_TYPE_P] = get_mv_predictors_pskip;
    pf->get_skip_mv_predictors[SLICE_TYPE_F] = get_mv_predictors_pskip;
    pf->get_skip_mv_predictors[SLICE_TYPE_B] = get_mv_predictors_bskip;
}
