/**
 * ===========================================================================
 * test.c
 * ---------------------------------------------------------------------------
 *
 * XRDENC, the Chinese AVS2 video encoder library.
 *
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * disable warning C4996: functions or variables may be unsafe. */
#if defined(_MSC_VER)
#define _CRT_SECURE_NO_WARNINGS
#endif

/* ---------------------------------------------------------------------------
 * alias of fseek/ftell */
#if defined(_MSC_VER) || defined(__INTEL_COMPILER)
#define file_seek       _fseeki64
#define file_tell       _ftelli64
#else
#define file_seek       fseek
#define file_tell       ftell
#endif

/* ---------------------------------------------------------------------------
 * include files */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdint.h>

#if defined(_MSC_VER)
#include <windows.h>
#include <process.h>
#else
#include <unistd.h>
#include <pthread.h>
#endif

#include "xrdenc.h"
#include "common.h"
#include "wrapper.h"
#include "encoder.h"


/* ---------------------------------------------------------------------------
 */
static FILE *g_infile  = NULL;
static FILE *g_outfile = NULL;

/* ---------------------------------------------------------------------------
 */
static void dump_encoded_data(void *coder, xrd_outpacket_t *packet)
{
    if (g_outfile) {
        if (packet->state == XRD_STATE_ENCODED) {
            fwrite(packet->stream, packet->len, 1, g_outfile);
        } else if (packet->state == XRD_STATE_FLUSH_END) {
            fwrite(packet->stream, packet->len, 1, g_outfile);
        }
        xrd_encoder_packet_unref(coder, packet);
    }
}

/* ---------------------------------------------------------------------------
 * read one frame data from file line by line
 */
static int read_one_frame(xrd_image_t *img, int shift_in)
{
    int k, j;
    if (img->in_sample_size != img->enc_sample_size) {
        static uint8_t p_buffer[16 * 1024];

        for (k = 0; k < img->i_plane; k++) {
            int i_width  = img->i_width[k];
            int i_stride = img->i_stride[k];

            if (img->in_sample_size == 1) {
                for (j = 0; j < img->i_lines[k]; j++) {
                    uint16_t *p_plane = (uint16_t *)&img->img_planes[k][j * i_stride];
                    int i;

                    if (fread(p_buffer, i_width, 1, g_infile) != 1) {
                        return -1;
                    }

                    memset(p_plane, 0, i_stride);
                    for (i = 0; i < i_width; i++) {
                        p_plane[i] = p_buffer[i] << shift_in;
                    }
                }
            } else {
                printf("Not supported high bit-depth for reading\n");
                return -1;
            }
        }
    } else {
        for (k = 0; k < img->i_plane; k++) {
            int size_line = img->i_width[k] * img->in_sample_size;
            for (j = 0; j < img->i_lines[k]; j++) {

                if (fread(img->img_planes[k] + img->i_stride[k] * j, size_line, 1, g_infile) != 1) {
                    return -1;
                }
            }
        }
    }
    return 0;
}

int start_encode(xrd_param_t *param)
{
    const char *in_file = xrd_encoder_opt_get(param, "input");
    const char *bs_file = xrd_encoder_opt_get(param, "output");
    const int shift_in  = atoi(xrd_encoder_opt_get(param, "SampleShift"));
    int num_frames      = atoi(xrd_encoder_opt_get(param, "frames"));
    xrd_picture_t pic;
    void *encoder = NULL;
    int k;
#if XRDENC_API_VERSION >= 2
    xrd_outpacket_t packet = {0};
#endif

    /* open input & output files */
    if ((g_infile = fopen(in_file, "rb")) == NULL) {
        fprintf(stderr, "error opening input file: \"%s\"\n", in_file);
        return -1;
    }

    if ((g_outfile = fopen(bs_file, "wb")) == NULL) {
        fprintf(stderr, "error opening output file: \"%s\"\n", bs_file);
        fclose(g_infile);

        return -1;
    }

    if (num_frames == 0) {
        num_frames = 1 << 30;
    }

    /* create the XRDENC video encoder */
#if XRDENC_API_VERSION >= 2
    encoder = xrd_encoder_create(param);
#else
    encoder = xrd_encoder_create(param, &dump_encoded_data, NULL);
#endif
    if (encoder == NULL) {
        fprintf(stderr, "Error: Can not create encoder. Null pointer returned.\n");
        fclose(g_infile);
        fclose(g_outfile);

        return -1;
    }

    /* read frame data and send to the XRDENC video encoder */
    for (k = 0; k < num_frames; k++) {
        if (xrd_encoder_get_buffer(encoder, &pic) < 0) {
            fprintf(stderr, "failed to get frame buffer [%3d,%3d].\n", k, num_frames);
            break;
        }

        if (read_one_frame(&pic.img, shift_in) < 0) {
            fprintf(stderr, "failed to read one YUV frame [%3d/%3d]\n", k, num_frames);
            /* return the buffer to the encoder */
            pic.i_state = XRD_STATE_NO_DATA;
#if XRDENC_API_VERSION >= 2
            xrd_encoder_encode(encoder, &pic, &packet);
            dump_encoded_data(encoder, &packet);
#else
            xrd_encoder_encode(encoder, &pic);
#endif
            break;
        }

        pic.i_state = 0;
        pic.i_type  = XRD_TYPE_AUTO;
        pic.i_pts   = k;

#if XRDENC_API_VERSION >= 2
        xrd_encoder_encode(encoder, &pic, &packet);
        dump_encoded_data(encoder, &packet);
#else
        xrd_encoder_encode(encoder, &pic);
#endif
    }

    /* flush delayed frames */
#if XRDENC_API_VERSION >= 2
    for (; packet.state != XRD_STATE_FLUSH_END;) {
        xrd_encoder_encode(encoder, NULL, &packet);
        dump_encoded_data(encoder, &packet);
    }
#else
    xrd_encoder_encode(encoder, NULL);
#endif

    /* destroy the encoder */
    xrd_encoder_destroy(encoder);

    return 0;
}

/* ---------------------------------------------------------------------------
 */
int main(int argc, char **argv)
{
    /* encoding parameters */
    xrd_param_t *param = NULL;
    int ret;

    if (argc < 2) {
        xrd_encoder_opt_help();            /* test.exe encoder.cfg */
        return -1;
    }

    /* parse parameters and modify the parameters */
    param = xrd_encoder_opt_alloc();
    if (xrd_encoder_opt_set(param, argc, argv) < 0) {
        fprintf(stderr, "parse contents error.");
        return -1;
    }

    /* test encoding */
    ret = start_encode(param);

    /* free spaces */
    xrd_encoder_opt_destroy(param);

    return ret;
}
