/**
 * ===========================================================================
 * xrdenc.h
 * - API interface
 * ---------------------------------------------------------------------------
 *
 * XRDENC, the Chinese AVS2 video encoder library.
 *
 * ===========================================================================
 */

#ifndef __XRDENC_H__
#define __XRDENC_H__

#ifdef __cplusplus
extern "C" {    // only need to export C interface if used by C++ source code
#endif

#if !defined(_STDINT_H) && !defined(_STDINT_H_) && !defined(_STDINT_H_INCLUDED) && !defined(_STDINT) &&\
    !defined(_SYS_STDINT_H_) && !defined(_INTTYPES_H) && !defined(_INTTYPES_H_) && !defined(_INTTYPES)
# ifdef _MSC_VER
#  pragma message("You must include stdint.h or inttypes.h before xrdenc.h")
# else
#  warning You must include stdint.h or inttypes.h before xrdenc.h
# endif
#endif

/* XRDEnc API version, version 2 the non-caller version */
#define XRDENC_API_VERSION          1

/**
 * ===========================================================================
 * const defines
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * state defines
 */
#define XRD_UNDEFINE        0
#define XRD_STATE_NO_DATA   1     /* no bitstream data */
#define XRD_STATE_ENCODED   2     /* one frame has been encoded */
#define XRD_STATE_FLUSH_END 9     /* flush end */
#define XRD_FLUSH           99    /* flush (fetch bitstream data only) */

/* ---------------------------------------------------------------------------
 * slice type
 */
#define XRD_TYPE_AUTO       0     /* Let XRDENC choose the right type */
#define XRD_TYPE_IDR        1
#define XRD_TYPE_I          2
#define XRD_TYPE_P          3
#define XRD_TYPE_F          4
#define XRD_TYPE_B          5
#define XRD_TYPE_KEYFRAME   6     /* IDR or I depending on b_open_gop option */
#define XRD_TYPE_G          7
#define XRD_TYPE_GB         8

/* ---------------------------------------------------------------------------
 * color space type
 */
#define XRD_CSP_MASK        0x00ff  /* mask */
#define XRD_CSP_NONE        0x0000  /* invalid mode     */
#define XRD_CSP_I420        0x0001  /* yuv 4:2:0 planar */
#define XRD_CSP_YV12        0x0002  /* yvu 4:2:0 planar */
#define XRD_CSP_NV12        0x0003  /* yuv 4:2:0, with one y plane and one packed u+v */
#define XRD_CSP_MAX         0x0004  /* end of list */
#define XRD_CSP_VFLIP       0x1000  /* the csp is vertically flipped */
#define XRD_CSP_HIGH_DEPTH  0x2000  /* the csp has a depth of 16 bits per pixel component */

/* ---------------------------------------------------------------------------
 * log level
 */
enum log_level_e {
    XRD_LOG_NONE     = -1,  /* none */
    XRD_LOG_ERROR    = 0,   /* level 0 */
    XRD_LOG_WARNING  = 1,   /* level 1 */
    XRD_LOG_INFO     = 2,   /* level 2 */
    XRD_LOG_DEBUG    = 3,   /* level 3 */
    XRD_LOG_NOPREFIX = 0x80 /* special using: without prefix */
};

/* ---------------------------------------------------------------------------
 * others
 */


/**
 * ===========================================================================
 * interface struct type defines
 * ===========================================================================
 */

/* -----------------------------
 * XRDENC input parameters
 *
 * For version safety you may use 
 *    xrd_encoder_opt_alloc(), xrd_encoder_opt_destroy()
 * to manage the allocation of xrd_param_t instances, and 
 *    xrd_encoder_opt_set(), xrd_encoder_opt_set2()
 * to assign values by name, and 
 *    xrd_encoder_opt_get(p, "param_name")
 * to get specific configuration value string.
 * 
 * Just treat xrd_param_t as an opaque data structure */
typedef struct xrd_param_t  xrd_param_t;


/* ---------------------------------------------------------------------------
 * xrd_image_t
 */
typedef struct xrd_image_t {
    int      i_csp;                    /* color space */
    int      in_sample_size;           /* input sample size in byte */
    int      enc_sample_size;          /* encoding sample size in byte */
    int      i_plane;                  /* number of image planes */
    int      i_width[3];               /* widths  for each plane */
    int      i_lines[3];               /* heights for each plane */
    int      i_stride[3];              /* strides for each plane */
    uint8_t *img_planes[4];            /* pointers to each plane (planes[3]: start buffer address) */
} xrd_image_t;


/* ---------------------------------------------------------------------------
 * xrd_picture_t
 */
typedef struct xrd_picture_t {
    /* [IN ]    flush or not
     * [OUT]    encoding state */
    int         i_state;
    /* [IN ]    force picture type (if not auto)
     *          if XRDENC encoding parameters are violated in the forcing of picture
     *          types, XRDENC will correct the input picture type and log a warning.
     *          the quality of frame type decisions may suffer if a great deal of
     *          fine-grained mixing of auto and forced frametypes is done
     * [OUT]    type of the picture encoded */
    int         i_type;
    /* [IN ]    force quantizer for != XRD_QP_AUTO */
    int         i_qpplus1;
    /* [OUT]    whether this frame is a keyframe. important when using modes that
     *          result in SEI recovery points being used instead of IDR frames */
    int         b_keyframe;
    /* [IN ]    user pts. [OUT]: pts of encoded picture (user) */
    int64_t     i_pts;
    /* [OUT]    frame dts. When the pts of the first frame is close to zero,
     *          initial frames may have a negative dts which must be dealt
     *          with by any muxer */
    int64_t     i_dts;
    /* [IN ]    raw data */
    xrd_image_t  img;
    /* [IN ]    private pointer, DO NOT change it */
    void       *priv;
} xrd_picture_t;

/* ---------------------------------------------------------------------------
 * xrd_outpacket_t
 */
typedef struct xrd_outpacket_t {
    void          *private_data;      /* private pointer, DONOT change it */
    const uint8_t *stream;            /* pointer to bitstream data buffer */
    int            len;               /* length  of bitstream data */
    int            state;             /* state of current frame encoded */
    int            type;              /* type  of current frame encoded */
    int64_t        pts;               /* pts   of current frame encoded */
    int64_t        dts;               /* dts   of current frame encoded */
    void           *opaque;           /* pointer to user data */
} xrd_outpacket_t;

#if XRDENC_API_VERSION < 2
/* ---------------------------------------------------------------------------
 * type define of dump function
 */
typedef void(*xrd_dump_func_t)(
    void *coder,                      /* handler of the encoder */
    xrd_outpacket_t *packet);       /* output bitstream */
#endif

#ifdef __cplusplus
}
#endif
#endif // __XRDENC_H__
